@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2009, 2013-2021 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta,
  *  GA 30332.
@@ -66,6 +66,7 @@
 #include "httpd-nodes.h"
 #include "httpd-io.h"
 #include "idle.h"
+#include "json.h"
 #include "kybd.h"
 #include "lazya.h"
 #include "login_macro.h"
@@ -77,6 +78,7 @@
 #include "opts.h"
 #include "popups.h"
 #include "pr3287_session.h"
+#include "prefer.h"
 #include "print_screen.h"
 #include "product.h"
 #include "proxy.h"
@@ -110,6 +112,7 @@
 #endif /*]*/
 
 #define STATS_POLL	(2 * 1000)
+#define NC_DELAY	50
 
 #if defined(_WIN32) /*[*/
 char *instdir = NULL;
@@ -125,6 +128,11 @@ static int bsent = 0;
 static int rsent = 0;
 static ioid_t stats_ioid = NULL_IOID;
 
+static bool b3270_toggle_yet = false;
+static char* crashptr;
+
+static ioid_t csdelay_ioid = NULL_IOID;
+
 static void b3270_toggle(toggle_index_t ix, enum toggle_type tt);
 static toggle_register_t toggles[] = {
     { MONOCASE,		b3270_toggle,	TOGGLE_NEED_INIT },
@@ -142,7 +150,8 @@ static toggle_register_t toggles[] = {
 };
 
 static void b3270_register(void);
-static void b3270_toggle_notify(const char *name, const char *value, ia_t ia);
+static void b3270_toggle_notify(const char *name, enum resource_type type, void **address, ia_t ia, unsigned flags);
+static void delayed_nc_report(ioid_t id);
 
 void
 usage(const char *msg)
@@ -150,7 +159,8 @@ usage(const char *msg)
     if (msg != NULL) {
 	fprintf(stderr, "%s\n", msg);
     }
-    fprintf(stderr, "Usage: %s [options] [profile-file.b3270]\n", programname);
+    fprintf(stderr, "Usage: %s [options] [<session-file>.b3270]\n",
+	    programname);
     fprintf(stderr, "Use " OptHelp1 " for the list of options\n");
     exit(1);
 }
@@ -159,11 +169,11 @@ usage(const char *msg)
 static void
 dump_stats(void)
 {
-    ui_vleaf(IndStats,
-	    AttrBytesReceived, lazyaf("%d", brcvd),
-	    AttrRecordsReceived, lazyaf("%d", rrcvd),
-	    AttrBytesSent, lazyaf("%d", bsent),
-	    AttrRecordsSent, lazyaf("%d", rsent),
+    ui_leaf(IndStats,
+	    AttrBytesReceived, AT_INT, (int64_t)brcvd,
+	    AttrRecordsReceived, AT_INT, (int64_t)rrcvd,
+	    AttrBytesSent, AT_INT, (int64_t)bsent,
+	    AttrRecordsSent, AT_INT, (int64_t)rsent,
 	    NULL);
 }
 
@@ -198,10 +208,32 @@ stats_poke(void)
  * Respond to a change in the connection, 3270 mode, or line mode.
  */
 static void
-b3270_connect(bool ignored)
+b3270_connect_common(bool direct_indication)
 {       
     static enum cstate old_cstate = NOT_CONNECTED;
 
+    /*
+     * If this is a direct call from the state change indication, and we've gone from something other than
+     * NOT_CONNECTED to NOT_CONNECTED, wait a moment before reporting that.  It's possible that NOT_CONNECTED
+     * was transient, and we don't want to report it.
+     *
+     * This is simpler than replumbing the retry/reconnect logic to skip the NOT_CONNECTED state, even if it is
+     * a bit awkward here.
+     */
+    if (direct_indication)
+    {
+	if (cstate == NOT_CONNECTED && old_cstate != NOT_CONNECTED && csdelay_ioid == NULL_IOID) {
+	    /* xxx => NOT_CONNECTED and no timer running. Start it. */
+	    csdelay_ioid = AddTimeOut(NC_DELAY, delayed_nc_report);
+	    return;
+	}
+	if (cstate != NOT_CONNECTED && csdelay_ioid != NULL_IOID) {
+	    /* Gone to something other than NOT_CONNECTED. Stop the timer, we want to report it. */
+	    RemoveTimeOut(csdelay_ioid);
+	    csdelay_ioid = NULL_IOID;
+	}
+    }
+
     if (cstate == old_cstate) {
 	return;
     }
@@ -224,8 +256,8 @@ b3270_connect(bool ignored)
 
     /* Tell the GUI about the new state. */
     if (cstate == NOT_CONNECTED) {
-	ui_vleaf(IndConnection,
-		AttrState, state_name[(int)cstate],
+	ui_leaf(IndConnection,
+		AttrState, AT_STRING, state_name[(int)cstate],
 		NULL);
     } else {
 	char *cause = NewString(ia_name[connect_ia]);
@@ -242,10 +274,10 @@ b3270_connect(bool ignored)
 	    }
 	    *s++ = c;
 	}
-	ui_vleaf(IndConnection,
-		AttrState, state_name[(int)cstate],
-		AttrHost, current_host,
-		AttrCause, cause,
+	ui_leaf(IndConnection,
+		AttrState, AT_STRING, state_name[(int)cstate],
+		AttrHost, AT_STRING, current_host,
+		AttrCause, AT_STRING, cause,
 		NULL);
 	Free(cause);
 
@@ -271,6 +303,23 @@ b3270_connect(bool ignored)
     old_cstate = cstate;
 }
 
+/* Report connection state after a short delay, it if transtioned to NOT_CONNECTED. */
+static void delayed_nc_report(ioid_t id)
+{
+    RemoveTimeOut(csdelay_ioid);
+    csdelay_ioid = NULL_IOID;
+    b3270_connect_common(false);	/* suppress adding another delay */
+}
+
+/**
+ * Respond to a change in the connection, 3270 mode, or line mode.
+ */
+static void
+b3270_connect(bool ignored)
+{
+    b3270_connect_common(true);		/* add a short delay if NOT_CONNECTED */
+}
+
 static void
 b3270_secure(bool ignored)
 {
@@ -281,13 +330,13 @@ b3270_secure(bool ignored)
     }
     is_secure = net_secure_connection();
 
-     ui_vleaf(IndTls,
-	     AttrSecure, ValTrueFalse(net_secure_connection()),
+     ui_leaf(IndTls,
+	     AttrSecure, AT_BOOLEAN, net_secure_connection(),
 	     AttrVerified,
-		 net_secure_connection()?
-		     ValTrueFalse(net_secure_unverified()): NULL,
-	     AttrSession, net_session_info(),
-	     AttrHostCert, net_server_cert_info(),
+		 net_secure_connection()? AT_BOOLEAN: AT_SKIP_BOOLEAN,
+		 !net_secure_unverified(),
+	     AttrSession, AT_STRING, net_session_info(),
+	     AttrHostCert, AT_STRING, net_server_cert_info(),
 	     NULL);
 }
 
@@ -302,9 +351,9 @@ report_terminal_name(void)
 	    || last_override != (appres.termname != NULL)) {
 	Replace(last_term_name, NewString(termtype));
 	last_override = appres.termname != NULL;
-	ui_vleaf(IndTerminalName,
-		AttrText, last_term_name,
-		AttrOverride, ValTrueFalse(last_override),
+	ui_leaf(IndTerminalName,
+		AttrText, AT_STRING, last_term_name,
+		AttrOverride, AT_BOOLEAN, last_override,
 		NULL);
     }
 }
@@ -325,28 +374,63 @@ static void
 dump_codepages(void)
 {
     cpname_t *cpnames = get_cpnames();
+    int i;
 
-    if (cpnames != NULL) {
-	int i;
+    if (cpnames == NULL) {
+	return;
+    }
 
-	ui_vpush(IndCodePages, NULL);
-	for (i = 0; cpnames[i].name != NULL; i++) {
-	    const char **params = Calloc(2 + (2 * cpnames[i].num_aliases) + 1,
-		    sizeof(char *));
+    if (JSON_MODE) {
+	uij_open_object(NULL);
+	uij_open_array(IndCodePages);
+    } else {
+	uix_push(IndCodePages, NULL);
+    }
+    for (i = 0; cpnames[i].name != NULL; i++) {
+	if (XML_MODE) {
+	    /*
+	     * In XML mode, it looks like:
+	     *  name="foo", alias1="bar", alias3="baz", ...
+	     */
 	    int j;
 
-	    params[0] = "name";
-	    params[1] = cpnames[i].name;
+	    uix_open_leaf(IndCodePage);
+	    ui_add_element("name", AT_STRING, cpnames[i].name);
 	    for (j = 0; j < cpnames[i].num_aliases; j++) {
-		params[2 + (j * 2)] = lazyaf("alias%d", j + 1);
-		params[2 + (j * 2) + 1] = cpnames[i].aliases[j];
+		ui_add_element(lazyaf("alias%d", j + 1), AT_STRING,
+			cpnames[i].aliases[j]);
 	    }
-	    ui_leaf(IndCodePage, params);
-	    Free((void *)params);
+	    uix_close_leaf();
+	} else {
+	    /*
+	     * In JSON mode, it looks like:
+	     *  {
+	     *    "name": "foo",
+	     *    "aliases": [ "bar", "baz", ... ]
+	     *  }
+	     */
+	    uij_open_object(NULL);
+	    ui_add_element("name", AT_STRING, cpnames[i].name);
+	    if (cpnames[i].num_aliases) {
+		int j;
+
+		uij_open_array("aliases");
+		for (j = 0; j < cpnames[i].num_aliases; j++) {
+		    ui_add_element(NULL, AT_STRING,
+			    cpnames[i].aliases[j]);
+		}
+		uij_close_array();
+	    }
+	    uij_close_object();
 	}
-	ui_pop();
-	free_cpnames(cpnames);
     }
+    if (JSON_MODE) {
+	uij_close_array();
+	uij_close_object();
+    } else {
+	uix_pop();
+    }
+    free_cpnames(cpnames);
 }
 
 /* Dump the model list. Called at initialization time. */
@@ -366,15 +450,34 @@ dump_models(void)
     };
     int i;
 
-    ui_vpush(IndModels, NULL);
+    if (XML_MODE) {
+	uix_push(IndModels, NULL);
+    } else {
+	uij_open_object(NULL);
+	uij_open_array(IndModels);
+    }
     for (i = 0; models[i].model != 0; i++) {
-	ui_vleaf(IndModel,
-		AttrModel, lazyaf("%d", models[i].model),
-		AttrRows, lazyaf("%d", models[i].rows),
-		AttrColumns, lazyaf("%d", models[i].columns),
-		NULL);
+	if (XML_MODE) {
+
+	    ui_leaf(IndModel,
+		    AttrModel, AT_INT, (int64_t)models[i].model,
+		    AttrRows, AT_INT, (int64_t)models[i].rows,
+		    AttrColumns, AT_INT, (int64_t)models[i].columns,
+		    NULL);
+	} else {
+	    uij_open_object(NULL);
+	    ui_add_element(AttrModel, AT_INT, (int64_t)models[i].model);
+	    ui_add_element(AttrRows, AT_INT, (int64_t)models[i].rows);
+	    ui_add_element(AttrColumns, AT_INT, (int64_t)models[i].columns);
+	    uij_close_object();
+	}
+    }
+    if (XML_MODE) {
+	uix_pop();
+    } else {
+	uij_close_array();
+	uij_close_object();
     }
-    ui_pop();
 }
 
 /* Dump the proxy list. */
@@ -383,24 +486,49 @@ dump_proxies(void)
 {
     proxytype_t type;
 
-    ui_vpush(IndProxies, NULL);
+    if (XML_MODE) {
+	uix_push(IndProxies, NULL);
+    } else {
+	uij_open_object(NULL);
+	uij_open_array(IndProxies);
+    }
+
     for (type = PT_FIRST; type < PT_MAX; type++) {
 	int default_port = proxy_default_port(type);
-	ui_vleaf(IndProxy,
-		AttrName, proxy_type_name(type),
-		AttrUsername, ValTrueFalse(proxy_takes_username(type)),
-		AttrPort, default_port? lazyaf("%d", default_port): NULL,
-		NULL);
+
+	if (XML_MODE) {
+	    ui_leaf(IndProxy,
+		    AttrName, AT_STRING, proxy_type_name(type),
+		    AttrUsername, AT_BOOLEAN, proxy_takes_username(type),
+		    AttrPort,
+			default_port? AT_INT: AT_SKIP_INT,
+			(int64_t)default_port,
+		    NULL);
+	} else {
+	    uij_open_object(NULL);
+	    ui_add_element(AttrName, AT_STRING, proxy_type_name(type));
+	    ui_add_element(AttrUsername, AT_BOOLEAN,
+		    proxy_takes_username(type));
+	    if (default_port) {
+		ui_add_element(AttrPort, AT_INT, (int64_t)default_port);
+	    }
+	    uij_close_object();
+	}
+    }
+    if (XML_MODE) {
+	uix_pop();
+    } else {
+	uij_close_array();
+	uij_close_object();
     }
-    ui_pop();
 }
 
 /* Dump the supported host prefix list. */
 static void
 dump_prefixes(void)
 {
-    ui_vleaf(IndPrefixes,
-	    AttrValue, host_prefixes(),
+    ui_leaf(IndPrefixes,
+	    AttrValue, AT_STRING, host_prefixes(),
 	    NULL);
 }
 
@@ -450,7 +578,9 @@ main(int argc, char *argv[])
     net_register();
     login_macro_register();
     vstatus_register();
+    prefer_register();
 
+    supports_cmdline_host = false;
     argc = parse_command_line(argc, (const char **)argv, &cl_hostname);
     if (cl_hostname != NULL) {
 	usage("Unrecognized option(s)");
@@ -459,11 +589,16 @@ main(int argc, char *argv[])
     check_min_version(appres.min_version);
 
     ui_io_init();
-    ui_vpush(IndInitialize, NULL);
-    ui_vleaf(IndHello,
-	    AttrVersion, lazyaf("%d.%d.%d", our_major, our_minor, our_iteration),
-	    AttrBuild, build,
-	    AttrCopyright,
+    if (XML_MODE) {
+	uix_push(IndInitialize, NULL);
+    } else {
+	uij_open_object(NULL);
+	uij_open_array(IndInitialize);
+    }
+    ui_leaf(IndHello,
+	    AttrVersion, AT_STRING, lazyaf("%d.%d.%d", our_major, our_minor, our_iteration),
+	    AttrBuild, AT_STRING, build,
+	    AttrCopyright, AT_STRING,
 lazyaf("\
 Copyright © 1993-%s, Paul Mattes.\n\
 Copyright © 1990, Jeff Sparkes.\n\
@@ -545,15 +680,42 @@ POSSIBILITY OF SUCH DAMAGE.", cyear),
     signal(SIGCHLD, sigchld_handler);
 #endif /*]*/
 
-    /* Handle initial toggle settings. */
+    /* Initialize the toggles. */
     initialize_toggles();
 
     /* Send TLS set-up. */
-    ui_vleaf(IndTlsHello,
-	    AttrSupported, ValTrueFalse(sio_supported()),
-	    AttrProvider, sio_provider(),
-	    AttrOptions, sio_option_names(),
-	    NULL);
+    if (sio_supported()) {
+	if (XML_MODE) {
+	    ui_leaf(IndTlsHello,
+		    AttrSupported, AT_BOOLEAN, sio_supported(),
+		    AttrProvider, AT_STRING, sio_provider(),
+		    AttrOptions, AT_STRING, sio_option_names(),
+		    NULL);
+	} else {
+	    char *option_names = NewString(sio_option_names());
+	    char *str = option_names;
+	    char *token;
+	    char *saveptr = NULL;
+
+	    uij_open_object(NULL);
+	    uij_open_object(IndTlsHello);
+	    ui_add_element(AttrSupported, AT_BOOLEAN, sio_supported());
+	    ui_add_element(AttrProvider, AT_STRING, sio_provider());
+	    uij_open_array(AttrOptions);
+	    while ((token = strtok_r(str, " ", &saveptr)) != NULL) {
+		str = NULL;
+		ui_add_element(NULL, AT_STRING, token);
+	    }
+	    uij_close_array();
+	    Free(option_names);
+	    uij_close_object();
+	    uij_close_object();
+	}
+    } else {
+	ui_leaf(IndTlsHello,
+		AttrSupported, AT_BOOLEAN, false,
+		NULL);
+    }
 
     /*
      * Register for extended toggle notifies, which will cause a dump of the
@@ -562,6 +724,7 @@ POSSIBILITY OF SUCH DAMAGE.", cyear),
      * Then dump the traditional toggles.
      */
     register_extended_toggle_notify(b3270_toggle_notify);
+    b3270_toggle_yet = true;
     for (ix = MONOCASE; ix < N_TOGGLES; ix++) {
 	if (toggle_supported(ix)) {
 	    b3270_toggle(ix, TT_INITIAL);
@@ -572,7 +735,12 @@ POSSIBILITY OF SUCH DAMAGE.", cyear),
     peer_script_init();
 
     /* Done with initialization.*/
-    ui_pop();
+    if (XML_MODE) {
+	uix_pop();
+    } else {
+	uij_close_array();
+	uij_close_object();
+    }
     popups_dump();
 
     /* Process events forever. */
@@ -673,10 +841,8 @@ Crash_action(ia_t ia, unsigned argc, const char **argv)
 	exit(999);
 	popup_an_error(AnCrash "(): Exit did not work");
     } else if (!strcasecmp(argv[0], KwNull)) {
-	char *s = NULL;
 	char c;
-
-	printf("%c\n", c = *s);
+	printf("%c\n", c = *crashptr);
 	popup_an_error(AnCrash "(): Null did not work");
     } else {
 	popup_an_error(AnCrash "(): Must specify " KwAssert ", " KwExit " or "
@@ -737,9 +903,10 @@ ForceStatus_action(ia_t ia, unsigned argc, const char **argv)
 		    reasons[reason], argv[1]);
 	    return false;
 	}
-	ui_vleaf(IndOia,
-		AttrField, OiaLock,
-		AttrValue, lazyaf("%s %s", reasons[reason], oerrs[oerr]),
+	ui_leaf(IndOia,
+		AttrField, AT_STRING, OiaLock,
+		AttrValue, AT_STRING,
+		    lazyaf("%s %s", reasons[reason], oerrs[oerr]),
 		NULL);
     } else if (!strcmp(argv[0], OiaLockScrolled)) {
 	int n;
@@ -755,9 +922,9 @@ ForceStatus_action(ia_t ia, unsigned argc, const char **argv)
 	    return false;
 	}
 	
-	ui_vleaf(IndOia,
-		AttrField, OiaLock,
-		AttrValue, lazyaf("%s %d", reasons[reason], n),
+	ui_leaf(IndOia,
+		AttrField, AT_STRING, OiaLock,
+		AttrValue, AT_STRING, lazyaf("%s %d", reasons[reason], n),
 		NULL);
     } else if (argc > 1) {
 	popup_an_error("ForceStatus: Reason '%s' does not take an argument",
@@ -765,25 +932,25 @@ ForceStatus_action(ia_t ia, unsigned argc, const char **argv)
 	return false;
     } else {
 	if (!strcmp(reasons[reason], STATUS_RECONNECTING)) {
-	    ui_vleaf(IndOia,
-		    AttrField, OiaLock,
-		    AttrValue, OiaLockNotConnected,
+	    ui_leaf(IndOia,
+		    AttrField, AT_STRING, OiaLock,
+		    AttrValue, AT_STRING, OiaLockNotConnected,
 		    NULL);
-	    ui_vleaf(IndConnection,
-		    AttrState, state_name[(int)RECONNECTING],
+	    ui_leaf(IndConnection,
+		    AttrState, AT_STRING, state_name[(int)RECONNECTING],
 		    NULL);
 	} else if (!strcmp(reasons[reason], STATUS_RESOLVING)) {
-	    ui_vleaf(IndOia,
-		    AttrField, OiaLock,
-		    AttrValue, OiaLockNotConnected,
+	    ui_leaf(IndOia,
+		    AttrField, AT_STRING, OiaLock,
+		    AttrValue, AT_STRING, OiaLockNotConnected,
 		    NULL);
-	    ui_vleaf(IndConnection,
-		    AttrState, state_name[(int)RESOLVING],
+	    ui_leaf(IndConnection,
+		    AttrState, AT_STRING, state_name[(int)RESOLVING],
 		    NULL);
 	} else {
-	    ui_vleaf(IndOia,
-		    AttrField, OiaLock,
-		    AttrValue, reasons[reason],
+	    ui_leaf(IndOia,
+		    AttrField, AT_STRING, OiaLock,
+		    AttrValue, AT_STRING, reasons[reason],
 		    NULL);
 	}
     }
@@ -794,26 +961,26 @@ ForceStatus_action(ia_t ia, unsigned argc, const char **argv)
 /**
  * xterm text escape
  *
- * @param[in] opcode    Operation to perform
- * @param[in] text      Associated text
+ * @param[in] code	Operation to perform
+ * @param[in] text	Associated text
  */
 void
 xterm_text_gui(int code, const char *text)
 {
     if (code == 0 || code == 1) {
-	ui_vleaf(IndIconName,
-		AttrText, text,
+	ui_leaf(IndIconName,
+		AttrText, AT_STRING, text,
 		NULL);
     }
     if (code == 0 || code == 2) {
-	ui_vleaf(IndWindowTitle,
-		AttrText, text,
+	ui_leaf(IndWindowTitle,
+		AttrText, AT_STRING, text,
 		NULL);
     }
 
     if (code == 50) {
-	ui_vleaf(IndFont,
-		AttrText, text,
+	ui_leaf(IndFont,
+		AttrText, AT_STRING, text,
 		NULL);
     }
 }
@@ -830,6 +997,12 @@ product_set_appres_defaults(void)
      */
     appres.oerr_lock = true;
     appres.interactive.save_lines = 4096;
+
+    /* XML is wrapped by default. */
+    appres.b3270.wrapper_doc = true;
+
+    /* UTF-8 by default. */
+    appres.utf8 = true;
 }
 
 /**
@@ -842,7 +1015,7 @@ b3270_toggle(toggle_index_t ix, enum toggle_type tt)
 {
     int i;
 
-    if (tt != TT_INITIAL) {
+    if (!b3270_toggle_yet || tt != TT_INITIAL) {
 	return;
     }
 
@@ -855,14 +1028,14 @@ b3270_toggle(toggle_index_t ix, enum toggle_type tt)
 	return;
     }
 
-    ui_vleaf(IndSetting,
-	    AttrName, toggle_names[i].name,
-	    AttrValue, toggled(ix)? ValTrue: ValFalse,
+    ui_leaf(IndSetting,
+	    AttrName, AT_STRING, toggle_names[i].name,
+	    AttrValue, AT_BOOLEAN, toggled(ix),
 	    NULL);
 
     if (ix == TRACING) {
-	ui_vleaf(IndTraceFile,
-		AttrName, (toggled(ix) && tracefile_name != NULL)?
+	ui_leaf(IndTraceFile,
+		AttrName, AT_STRING, (toggled(ix) && tracefile_name != NULL)?
 		    tracefile_name: NULL,
 		NULL);
     }
@@ -886,13 +1059,35 @@ menubar_as_set(bool sensitive _is_unused)
  * Handle a generic toggle change.
  */
 static void
-b3270_toggle_notify(const char *name, const char *value, ia_t cause)
+b3270_toggle_notify(const char *name, enum resource_type type, void **address, ia_t cause, unsigned flags)
 {
-    ui_vleaf(IndSetting,
-	    AttrName, name,
-	    AttrValue, value,
-	    AttrCause, ia_name[cause],
-	    NULL);
+    if (address == NULL) {
+	return;
+    }
+
+    switch (type) {
+    case XRM_STRING:
+	ui_leaf(IndSetting,
+		AttrName, AT_STRING, name,
+		AttrValue, AT_STRING, *(char **)address,
+		AttrCause, AT_STRING, ia_name[cause],
+		NULL);
+	break;
+    case XRM_INT:
+	ui_leaf(IndSetting,
+		AttrName, AT_STRING, name,
+		AttrValue, AT_INT, (int64_t)*(int *)address,
+		AttrCause, AT_STRING, ia_name[cause],
+		NULL);
+	break;
+    case XRM_BOOLEAN:
+	ui_leaf(IndSetting,
+		AttrName, AT_STRING, name,
+		AttrValue, AT_BOOLEAN, *(bool *)address,
+		AttrCause, AT_STRING, ia_name[cause],
+		NULL);
+	break;
+    }
 }
 
 /**
@@ -918,10 +1113,10 @@ tls_passwd_gui_callback(char *buf, int size, bool again)
 static void
 b3270_printer(bool on)
 {
-    ui_vleaf(IndOia,
-	    AttrField, "printer-session",
-	    AttrValue, ValTrueFalse(on),
-	    AttrLu, on? pr3287_session_lu(): NULL,
+    ui_leaf(IndOia,
+	    AttrField, AT_STRING, "printer-session",
+	    AttrValue, AT_BOOLEAN, on,
+	    AttrLu, AT_STRING, on? pr3287_session_lu(): NULL,
 	    NULL);
 }
 
@@ -953,15 +1148,26 @@ b3270_register(void)
     static opt_t b3270_opts[] = {
 	{ OptCallback, OPT_STRING,  false, ResCallback,
 	    aoffset(scripting.callback), NULL, "Callback address and port" },
+	{ OptIndent,   OPT_BOOLEAN, true,  ResIndent,    aoffset(b3270.indent),
+	    NULL, "Indent ouput" },
+	{ OptJson,     OPT_BOOLEAN, true,  ResJson,      aoffset(b3270.json),
+	    NULL, "Use JSON format" },
+	{ OptNoWrapperDoc,OPT_BOOLEAN,false,ResWrapperDoc,aoffset(b3270.wrapper_doc),
+	    NULL, "Do not use an XML wrapper document" },
 	{ OptUtf8,     OPT_BOOLEAN, true,  ResUtf8,      aoffset(utf8),
 	    NULL, "Force local codeset to be UTF-8" },
+	{ OptXml,      OPT_BOOLEAN, false, ResJson,      aoffset(b3270.json),
+	    NULL, "Use XML format" },
     };
     static res_t b3270_resources[] = {
 	{ ResCallback,		aoffset(scripting.callback), XRM_STRING },
 	{ ResIdleCommand,aoffset(idle_command),     XRM_STRING },
 	{ ResIdleCommandEnabled,aoffset(idle_command_enabled),XRM_BOOLEAN },
 	{ ResIdleTimeout,aoffset(idle_timeout),     XRM_STRING },
+	{ ResIndent,		aoffset(b3270.indent),XRM_BOOLEAN },
+	{ ResJson,		aoffset(b3270.json),XRM_BOOLEAN },
 	{ ResUtf8,		aoffset(utf8),      XRM_BOOLEAN },
+	{ ResWrapperDoc,	aoffset(b3270.wrapper_doc), XRM_BOOLEAN },
     };
     static xres_t b3270_xresources[] = {
 	{ ResPrintTextScreensPerPage,	V_FLAT },
