@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2021 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta, GA
  *  30332.
@@ -81,6 +81,7 @@
 #include "opts.h"
 #include "popups.h"
 #include "pr3287_session.h"
+#include "prefer.h"
 #include "print_screen.h"
 #include "product.h"
 #include "proxy_toggle.h"
@@ -126,8 +127,6 @@
 # define DELENV		"WC3DEL"
 #endif /*]*/
 
-#define PR3287_NAME	"pr3287"
-
 #if !defined(_WIN32) /*[*/
 # if defined(HAVE_LIBREADLINE) /*[*/
 #  define PROMPT_PRE	xs_buffer("%c%s%c", '\001', \
@@ -152,7 +151,7 @@ static bool merge_profile(void);
 #endif /*]*/
 
 #if defined(HAVE_LIBREADLINE) /*[*/
-static char **attempted_completion();
+static char **attempted_completion(const char *text, int start, int end);
 static char *completion_entry(const char *, int);
 static char *readline_command;
 static bool readline_done = false;
@@ -213,6 +212,8 @@ static bool exit_pending = false;
 static bool aux_input = false;
 static bool aux_pwinput = false;
 
+static bool escape_single = false;
+
 static void c3270_input(iosrc_t fd, ioid_t id);
 static ioid_t c3270_input_id = NULL_IOID;
 static task_cb_ir_state_t command_ir_state;
@@ -286,6 +287,8 @@ usage(const char *msg)
     }
     fprintf(stderr, "Usage: %s [options] [prefix:][LUname@]hostname[:port]\n",
 	    programname);
+    fprintf(stderr, "       %s [options] [<session-file>.s3270]\n",
+	    programname);
     fprintf(stderr, "Use " OptHelp1 " for the list of options\n");
     exit(1);
 }
@@ -349,8 +352,8 @@ c3270_3270_mode(bool ignored)
  * This handles asynchronous errors, such as file transfers that do not start
  * or abort.
  *
- * @param[in] s		Text to display
- * @param[in] any_error	true to set any_error_output
+ * @param[in] s			Text to display
+ * @param[in] set_any_error	true to set any_error_output
  * @returns true
  */
 static bool
@@ -446,10 +449,10 @@ c3270_connect(bool ignored)
     /* Not connected. */
     if (!appres.secure &&
 	    !PCONNECTED &&
-	    !appres.interactive.reconnect) {
+	    !host_retry_mode) {
 	glue_gui_error_cond("Disconnected.", false);
     }
-    if (connect_once) {
+    if (connect_once && !host_retry_mode) {
 	/* Exit after the connection is broken. */
 	if (command_running || PAGER_RUNNING) {
 	    /* Exit when the command and pager are complete. */
@@ -549,6 +552,17 @@ exit_pause(bool mode _is_unused)
 }
 #endif /*]*/
 
+/* Initialize the prompt. */
+static void
+prompt_init(void)
+{
+    prompt.default_string = appres.secure? "[Press <Enter>] ":
+	(color_prompt?
+	 xs_buffer("%s%s> %s", PROMPT_PRE, app, PROMPT_POST):
+	 xs_buffer("%s> ", app));
+    prompt.string = prompt.default_string;
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -624,6 +638,7 @@ main(int argc, char *argv[])
     net_register();
     login_macro_register();
     vstatus_register();
+    prefer_register();
 
 #if !defined(_WIN32) /*[*/
     register_merge_profile(merge_profile);
@@ -700,15 +715,7 @@ Type 'help' for help information.\n\n",
     color_prompt = screen_has_ansi_color();
 #endif /*]*/
 
-    if (appres.secure) {
-	/* We don't allow a command line in secure mode. */
-	prompt.default_string = prompt.string = "[Press <Enter>] ";
-    } else {
-	prompt.default_string = prompt.string =
-	    color_prompt?
-		xs_buffer("%s%s> %s", PROMPT_PRE, app, PROMPT_POST):
-		xs_buffer("%s> ", app);
-    }
+    prompt_init();
 
 #if defined(_WIN32) /*[*/
     /* Create a thread to read data from stdin. */
@@ -765,7 +772,7 @@ Type 'help' for help information.\n\n",
     if (cl_hostname != NULL) {
 	pause_for_errors();
 	/* Connect to the host. */
-	if (!appres.interactive.reconnect) {
+	if (!appres.reconnect) {
 	    connect_once = true;
 	}
 	c3270_push_command(lazyaf(AnConnect "(\"%s\")", cl_hostname));
@@ -804,7 +811,12 @@ Type 'help' for help information.\n\n",
 		!prompt.displayed &&
 		!command_running &&
 		!PAGER_RUNNING) {
-	    interact();
+	    if (escape_single && !aux_input) {
+		escape_single = false;
+		screen_resume();
+	    } else {
+		interact();
+	    }
 	}
     }
 }
@@ -827,6 +839,16 @@ echo_mode(bool echo)
 #endif /*]*/
 }
 
+/* Reset the prompt. */
+static void
+reset_prompt(void)
+{
+    if (prompt.string != prompt.default_string) {
+	Free(prompt.string);
+    }
+    prompt.string = prompt.default_string;
+}
+
 #if !defined(_WIN32) /*[*/
 /* Synchronous signal handler. */
 static void
@@ -879,7 +901,7 @@ synchronous_signal(iosrc_t fd, ioid_t id)
 	    aux_pwinput = false;
 	    echo_mode(true);
 	    c3270_push_command(RESUME_INPUT "(" RESUME_INPUT_ABORT ")");
-	    Replace(prompt.string, prompt.default_string);
+	    reset_prompt();
 	    RemoveInput(c3270_input_id);
 	    c3270_input_id = NULL_IOID;
 	    prompt.displayed = false;
@@ -1133,7 +1155,7 @@ c3270_input(iosrc_t fd, ioid_t id)
 	    aux_pwinput = false;
 	    echo_mode(true);
 	    c3270_push_command(RESUME_INPUT "(" RESUME_INPUT_ABORT ")");
-	    Replace(prompt.string, prompt.default_string);
+	    reset_prompt();
 	} else {
 	    /* Get more input. */
 	    printf("\n");
@@ -1193,7 +1215,7 @@ c3270_input(iosrc_t fd, ioid_t id)
 	echo_mode(true);
 	c3270_push_command(lazyaf(RESUME_INPUT "(%s)",
 		    *s? lazya(base64_encode(s)): "\"\""));
-	Replace(prompt.string, prompt.default_string);
+	reset_prompt();
     } else {
 	c3270_push_command(s);
     }
@@ -1202,7 +1224,7 @@ done:
 #if defined(HAVE_LIBREADLINE) /*[*/
     Free(command);
 #endif /*]*/
-    s = s; /* to make the compiler happy */
+    return;
 }
 
 #if defined(_WIN32) /*[*/
@@ -1339,6 +1361,9 @@ start_pager(void)
 
 	s = Malloc(strlen(pager_cmd) + strlen(or_cat) + 1);
 	sprintf(s, "%s%s", pager_cmd, or_cat);
+	if (appres.secure) {
+	    putenv("LESSSECURE=1");
+	}
 	pager.fp = xpopen(s, "w", &pager.pid);
 	Free(s);
 	if (pager.fp == NULL) {
@@ -1491,7 +1516,7 @@ attempted_completion(const char *text, int start, int end)
 
     /* If this is not the first word, fail. */
     s = rl_line_buffer;
-    while (*s && isspace(*s)) {
+    while (*s && isspace((int)*s)) {
 	s++;
     }
     if (s - rl_line_buffer < start) {
@@ -1505,10 +1530,10 @@ attempted_completion(const char *text, int start, int end)
 	 */
 
 	/* See if we're in the second word. */
-	while (*t && !isspace(*t)) {
+	while (*t && !isspace((int)*t)) {
 	    t++;
 	}
-	while (*t && isspace(*t)) {
+	while (*t && isspace((int)*t)) {
 	    t++;
 	}
 	if (t - rl_line_buffer < start) {
@@ -1520,8 +1545,8 @@ attempted_completion(const char *text, int start, int end)
 	 * we might do other expansions, and this code would need to
 	 * be generalized.
 	 */
-	if (!((!strncasecmp(s, AnOpen, 4) && isspace(*(s + 4))) ||
-	      (!strncasecmp(s, AnConnect, 7) && isspace(*(s + 7))))) {
+	if (!((!strncasecmp(s, AnOpen, 4) && isspace((int)*(s + 4))) ||
+	      (!strncasecmp(s, AnConnect, 7) && isspace((int)*(s + 7))))) {
 	    return NULL;
 	}
 
@@ -1636,8 +1661,26 @@ completion_entry(const char *text, int state)
 static bool
 Escape_action(ia_t ia, unsigned argc, const char **argv)
 {
+    bool no_prompt_after = false;
+
     action_debug(AnEscape, ia, argc, argv);
-    if (check_argc(AnEscape, argc, 0, 1) < 0) {
+    if (check_argc(AnEscape, argc, 0, 2) < 0) {
+	return false;
+    }
+
+    if (argc > 0 && !strcasecmp(argv[0], KwDashNoPromptAfter)) {
+	no_prompt_after = true;
+	argc--;
+	argv++;
+	if (argc == 0) {
+	    popup_an_error(AnEscape "(): Must specify an action with "
+		    KwDashNoPromptAfter);
+	    return false;
+	}
+    }
+    if (argc > 1) {
+	popup_an_error(AnEscape "(): Too many arguments, or unrecognized "
+		"option");
 	return false;
     }
 
@@ -1653,6 +1696,7 @@ Escape_action(ia_t ia, unsigned argc, const char **argv)
 	}
 	c3270_screen_suspend();
 	if (argc > 0) {
+	    escape_single = no_prompt_after;
 	    c3270_push_command(argv[0]);
 	}
     }
@@ -1728,6 +1772,9 @@ command_reqinput(task_cbh handle, const char *buf, size_t len, bool echo)
 
     p = lazya(base64_decode(buf));
 
+    if (prompt.string != prompt.default_string) {
+	Free(prompt.string);
+    }
     prompt.string =
 	color_prompt?
 	    xs_buffer("%s%s%s", PROMPT_PRE, p, PROMPT_POST):
@@ -1818,7 +1865,7 @@ command_getflags(task_cbh handle)
      * CONNECT_NONBLOCK: We do not want Connect()/Open() to block.
      * PWINPUT: We understand [pwinput] responses.
      */
-    return CBF_INTERACTIVE | CBF_CONNECT_NONBLOCK | CBF_PWINPUT;
+    return CBF_INTERACTIVE | CBF_CONNECT_FT_NONBLOCK | CBF_PWINPUT;
 }
 
 /**
@@ -2136,7 +2183,7 @@ product_set_appres_defaults(void)
     appres.trace_monitor = true;
     set_toggle(UNDERSCORE, true);
 #else /*][*/
-    appres.c3270.meta_escape = "auto";
+    appres.c3270.meta_escape = NewString("auto");
     appres.c3270.curses_keypad = true;
     appres.c3270.mouse = true;
 #endif /*]*/
@@ -2204,15 +2251,41 @@ glue_gui_output(const char *s)
  * This handles asynchronous errors, such as file transfers that do not start
  * or abort.
  *
+ * @param[in] type	Error type
  * @param[in] s		Text to display
- * @returns true
+ * @returns true (the error has been handled)
  */
 bool
-glue_gui_error(const char *s)
+glue_gui_error(pae_t type, const char *s)
 {
+    if (type == ET_CONNECT && host_retry_mode) {
+	return true;
+    }
     return glue_gui_error_cond(s, true);
 }
 
+/**
+ * Determine if it is safe to process an Open()/Connect().
+ *
+ * @returns true if safe.
+ */
+bool
+glue_gui_open_safe(void)
+{
+    return screen_initted || task_running_cb_contains(&command_cb);
+}
+
+/**
+ * Determine if it is valid to run an interactive Script() action.
+ *
+ * @returns true.
+ */
+bool
+glue_gui_script_interactive(void)
+{
+    return true;
+}
+
 /**
  * c3270 main module registration.
  */
@@ -2232,12 +2305,8 @@ c3270_register(void)
 	{ OptNoPrompt, OPT_BOOLEAN, true,  ResNoPrompt,
 	    aoffset(secure),
 	    NULL, "Alias for -secure" },
-	{ OptPrinterLu,OPT_STRING,  false, ResPrinterLu,
-	    aoffset(interactive.printer_lu),
-	    "<luname>",
-	    "Automatically start a "PR3287_NAME" printer session to <luname>" },
 	{ OptReconnect,OPT_BOOLEAN, true,  ResReconnect,
-	    aoffset(interactive.reconnect),
+	    aoffset(reconnect),
 	    NULL, "Reconnect to host as soon as it disconnects" },
 	{ OptSaveLines, OPT_INT,    false, ResSaveLines,
 	    aoffset(interactive.save_lines),
@@ -2298,9 +2367,6 @@ c3270_register(void)
 	{ ResKeymap,	aoffset(interactive.key_map),	XRM_STRING },
 	{ ResMenuBar,	aoffset(interactive.menubar),	XRM_BOOLEAN },
 	{ ResNoPrompt,	aoffset(secure),		XRM_BOOLEAN },
-	{ ResPrinterLu,	aoffset(interactive.printer_lu),XRM_STRING },
-	{ ResPrinterOptions,aoffset(interactive.printer_opts),XRM_STRING },
-	{ ResReconnect,	aoffset(interactive.reconnect),XRM_BOOLEAN },
 	{ ResSaveLines,	aoffset(interactive.save_lines),XRM_INT },
 #if !defined(_WIN32) /*[*/
 	{ ResCbreak,	aoffset(c3270.cbreak_mode),	XRM_BOOLEAN },
@@ -2330,14 +2396,9 @@ c3270_register(void)
     };
     static xres_t c3270_xresources[] = {
 	{ ResKeymap,			V_WILD },
-	{ ResAssocCommand,		V_FLAT },
-	{ ResLuCommandLine,		V_FLAT },
 	{ ResPrintTextScreensPerPage,	V_FLAT },
 	{ ResMessage,			V_WILD },
 #if defined(_WIN32) /*[*/
-	{ ResPrinterCodepage,		V_FLAT },
-	{ ResPrinterCommand,		V_FLAT },
-	{ ResPrinterName, 		V_FLAT },
 	{ ResPrintTextFont, 		V_FLAT },
 	{ ResPrintTextHorizontalMargin,	V_FLAT },
 	{ ResPrintTextOrientation,	V_FLAT },
@@ -2349,7 +2410,6 @@ c3270_register(void)
 	{ ResHostColorForProtectedIntensified,V_FLAT },
 	{ ResConsoleColorForHostColor,	V_COLOR },
 #else /*][*/
-	{ ResPrinterCommand,		V_FLAT },
 	{ ResPrintTextCommand,		V_FLAT },
 	{ ResCursesColorForDefault,	V_FLAT },
 	{ ResCursesColorForIntensified,	V_FLAT },
