@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2010-2021 Paul Mattes.
+ * Copyright (c) 2010-2023 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -521,7 +521,7 @@ menu_key(menu_key_t k, ucs4_t u)
 	if (getmouse(&m) != OK) {
 	    return;
 	}
-	if (!(m.bstate & (BUTTON1_PRESSED || BUTTON1_RELEASED))) {
+	if (!(m.bstate & (BUTTON1_PRESSED | BUTTON1_RELEASED))) {
 	    return;
 	}
 
@@ -902,6 +902,25 @@ menu_callback file_menu_actions[FM_COUNT] = {
     fm_quit
 };
 
+static file_menu_enum fm_insecure[] = {
+    FM_PROMPT,
+    FM_PRINT,
+    FM_XFER,
+    FM_TRACE,
+    FM_SCREENTRACE,
+    FM_SCREENTRACE_PRINTER,
+    FM_SAVE_INPUT,
+    FM_RESTORE_INPUT,
+#if defined(HAVE_START) /*[*/
+    FM_HELP,
+#endif /*]*/
+#if defined(_WIN32) /*[*/
+    FM_WIZARD,
+    FM_WIZARD_SESS,
+#endif /*]*/
+};
+#define NUM_FM_INSECURE	((int)(sizeof(fm_insecure) / sizeof(file_menu_enum)))
+
 /* Options menu. */
 typedef enum {
     OM_MONOCASE,
@@ -917,6 +936,7 @@ typedef enum {
     OM_VISIBLE_CONTROL,
     OM_TYPEAHEAD,
     OM_ALWAYS_INSERT,
+    OM_UNDERSCORE_BLANK_FILL,
     OM_COUNT
 } options_menu_enum;
 cmenu_item_t *options_menu_items[OM_COUNT];
@@ -933,7 +953,8 @@ toggle_index_t option_index[OM_COUNT] = {
 #endif /*]*/
     VISIBLE_CONTROL,
     TYPEAHEAD,
-    ALWAYS_INSERT
+    ALWAYS_INSERT,
+    UNDERSCORE_BLANK_FILL,
 };
 char *option_names[OM_COUNT] = {
     "Monocase",
@@ -949,6 +970,7 @@ char *option_names[OM_COUNT] = {
     "Visible Control",
     "Typeahead",
     "Default Insert Mode",
+    "Underscore Blank Fill",
 };
 
 cmenu_t *file_menu;
@@ -982,7 +1004,7 @@ popup_keypad(void *ignored _is_unused)
 
 /* Run an item from the Macros menu. */
 static void
-run_macro(void *param)
+menu_run_macro(void *param)
 {
     struct macro_def *m = (struct macro_def *)param;
 
@@ -1017,43 +1039,54 @@ draw_topline(void)
 void
 menu_init(void)
 {
-    int j;
+    file_menu_enum f;
+    options_menu_enum o;
 
     basic_menu_init();
 
     file_menu = add_menu("File");
-    for (j = 0; j < FM_COUNT; j++) {
-	if (appres.secure &&
-		(j == FM_PROMPT || j == FM_XFER || j == FM_TRACE)) {
-	    continue;
+    for (f = 0; f < FM_COUNT; f++) {
+	if (appres.secure) {
+	    int fi;
+	    bool secure = true;
+
+	    for (fi = 0; fi < NUM_FM_INSECURE; fi++) {
+		if (f == fm_insecure[fi]) {
+		    secure = false;
+		    break;
+		}
+	    }
+	    if (!secure) {
+		continue;
+	    }
 	}
 #if defined(WC3270) /*[*/
-	if (j == FM_WIZARD_SESS && profile_path == NULL) {
+	if (f == FM_WIZARD_SESS && profile_path == NULL) {
 	    continue;
 	}
-	if (j == FM_WIZARD_SESS) {
+	if (f == FM_WIZARD_SESS) {
 	    char *text;
 
 	    text = xs_buffer("Edit Session %s", profile_name);
 
-	    file_menu_items[j] = add_item(file_menu, text,
-		    file_menu_actions[j], profile_path);
+	    file_menu_items[f] = add_item(file_menu, text,
+		    file_menu_actions[f], profile_path);
 	} else
 #endif /*]*/
 	{
-	    file_menu_items[j] = add_item(file_menu, file_menu_names[j],
-		    file_menu_actions[j], NULL);
+	    file_menu_items[f] = add_item(file_menu, file_menu_names[f],
+		    file_menu_actions[f], NULL);
 	}
     }
     options_menu = add_menu("Options");
-    for (j = 0; j < OM_COUNT; j++) {
+    for (o = 0; o < OM_COUNT; o++) {
 	char *name;
 
 	name = xs_buffer("%s %s",
-		toggled(option_index[j])? "Disable": "Enable",
-		option_names[j]);
-	options_menu_items[j] = add_item(options_menu, name, toggle_option,
-		&option_index[j]);
+		toggled(option_index[o])? "Disable": "Enable",
+		option_names[o]);
+	options_menu_items[o] = add_item(options_menu, name, toggle_option,
+		&option_index[o]);
 	Free(name);
     }
     keypad_menu = add_menu("Keypad");
@@ -1089,7 +1122,7 @@ menubar_connect(bool connected)
 		strcpy(mm->name, m->name);
 		mm->action = mm->name + strlen(mm->name) + 1;
 		strcpy(mm->action, m->action);
-		add_item(macros_menu, m->name, run_macro, mm);
+		add_item(macros_menu, m->name, menu_run_macro, mm);
 
 		macro_save = (struct macro_def **)Realloc(macro_save,
 			(n_ms + 1) * sizeof(struct macro_def *));
@@ -1145,7 +1178,7 @@ menubar_retoggle(toggle_index_t ix)
 	Free(s);
 	return;
     }
-    if (ix == TRACING) {
+    if (ix == TRACING && !appres.secure) {
 	s = xs_buffer("%sable Tracing", (toggled(TRACING))? "Dis": "En");
 	rename_item(file_menu_items[FM_TRACE], s);
 	Free(s);
