@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2016, 2018-2021 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -36,7 +36,6 @@
 #include <fcntl.h>
 #if !defined(_WIN32) /*[*/
 # include <signal.h>
-# include <sys/signal.h>
 # include <sys/wait.h>
 # include <sys/types.h>
 # include <sys/socket.h>
@@ -52,6 +51,7 @@
 #include "child_popups.h"
 #include "childscript.h"
 #include "find_console.h"
+#include "glue_gui.h"
 #include "httpd-core.h"
 #include "httpd-io.h"
 #include "lazya.h"
@@ -484,7 +484,7 @@ child_stdout(iosrc_t fd _is_unused, ioid_t id)
 #endif /*]*/
 
 /**
- * Callback for data returned to child script command via a pipe.
+ * Callback for data returned to child script command via a pipe (POSIX only).
  *
  * @param[in] handle    Callback handle
  * @param[in] buf       Buffer
@@ -496,18 +496,31 @@ child_data(task_cbh handle, const char *buf, size_t len, bool success)
 {
 #if !defined(_WIN32) /*[*/
     child_t *c = (child_t *)handle;
-    char *s = lazyaf(DATA_PREFIX "%.*s\n", (int)len, buf);
+    char *b;
+    char *newline;
+    char *s;
     ssize_t nw;
 
+    while (len > 0 && buf[len - 1] == '\n') {
+	len--;
+    }
+    b = xs_buffer("%.*s", (int)len, buf);
+    while ((newline = strchr(b, '\n')) != NULL) {
+	*newline = ' ';
+    }
+
+    s = xs_buffer("%s%s\n", DATA_PREFIX, b);
+    Free(b);
     nw = write(c->outfd, s, strlen(s));
     if (nw != (ssize_t)strlen(s)) {
 	vtrace("child_data: short write\n");
     }
+    Free(s);
 #endif /*]*/
 }
 
 /**
- * Callback for input request
+ * Callback for input request (POSIX only).
  *
  * @param[in] handle    Callback handle
  * @param[in] buf       Buffer
@@ -1070,6 +1083,7 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
     unsigned short httpd_port;
     unsigned short script_port;
     struct sockaddr_in *sin;
+    bool interactive = false;
 #if !defined(_WIN32) /*[*/
     pid_t pid;
     int inpipe[2] = { -1, -1 };
@@ -1107,17 +1121,35 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
 	    stdout_redirect = false;
 	    argc--;
 	    argv++;
+	} else if (glue_gui_script_interactive() &&
+		!strcasecmp(argv[0], KwDashInteractive)) {
+	    interactive = true;
+	    stdout_redirect = false;
+#if defined(_WIN32) /*[*/
+	    share_console = true;
+#endif /*]*/
+	    argc--;
+	    argv++;
 #if defined(_WIN32) /*[*/
 	} else if (!strcasecmp(argv[0], KwDashShareConsole)) {
 	    share_console = true;
 	    argc--;
 	    argv++;
 #endif /*]*/
+	} else if (argv[0][0] == '-') {
+	    popup_an_error(AnScript "() unknown option %s", argv[0]);
+	    return false;
 	} else {
 	    break;
 	}
     }
 
+    if (async && interactive) {
+	popup_an_error(AnScript "(): cannot specify both " KwDashAsync " and "
+		KwDashInteractive);
+	return false;
+    }
+
     listeners.peer = NULL;
     listeners.httpd = NULL;
 
@@ -1147,9 +1179,6 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
     script_port = ntohs(sin->sin_port);
     Free(sin);
 
-    putenv(lazyaf(URL_ENV "=http://127.0.0.1:%u/3270/rest/", httpd_port));
-    putenv(lazyaf(PORT_ENV "=%d", script_port));
-
 #if !defined(_WIN32) /*[*/
     /*
      * Create pipes and stdout stream for the script process.
@@ -1170,13 +1199,15 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
     }
 
     /* Create a pipe to capture child stdout. */
-    if (pipe(stdoutpipe) < 0) {
-	popup_an_error("pipe() failed");
-	close(outpipe[0]);
-	close(outpipe[1]);
-	close(inpipe[0]);
-	close(inpipe[1]);
-	close_listeners(&listeners);
+    if (!interactive) {
+	if (pipe(stdoutpipe) < 0) {
+	    popup_an_error("pipe() failed");
+	    close(outpipe[0]);
+	    close(outpipe[1]);
+	    close(inpipe[0]);
+	    close(inpipe[1]);
+	    close_listeners(&listeners);
+	}
     }
 
     /* Fork and exec the script process. */
@@ -1186,8 +1217,10 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
 	close(inpipe[1]);
 	close(outpipe[0]);
 	close(outpipe[1]);
-	close(stdoutpipe[0]);
-	close(stdoutpipe[1]);
+	if (!interactive) {
+	    close(stdoutpipe[0]);
+	    close(stdoutpipe[1]);
+	}
 	close_listeners(&listeners);
 	return false;
     }
@@ -1203,19 +1236,25 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
 	/* Clean up the pipes. */
 	close(outpipe[1]);
 	close(inpipe[0]);
-	close(stdoutpipe[0]);
+	if (!interactive) {
+	    close(stdoutpipe[0]);
+	}
 
 	/* Redirect output. */
-	if (stdout_redirect) {
-	    dup2(stdoutpipe[1], 1);
-	} else {
-	    dup2(open("/dev/null", O_WRONLY), 1);
+	if (!interactive) {
+	    if (stdout_redirect) {
+		dup2(stdoutpipe[1], 1);
+	    } else {
+		dup2(open("/dev/null", O_WRONLY), 1);
+	    }
+	    dup2(stdoutpipe[1], 2);
 	}
-	dup2(stdoutpipe[1], 2);
 
-	/* Export the names of the pipes into the environment. */
+	/* Export the names of the pipes, URL and port into the environment. */
 	putenv(xs_buffer(OUTPUT_ENV "=%d", outpipe[0]));
 	putenv(xs_buffer(INPUT_ENV "=%d", inpipe[1]));
+	putenv(xs_buffer(URL_ENV "=http://127.0.0.1:%u/3270/rest/", httpd_port));
+	putenv(xs_buffer(PORT_ENV "=%d", script_port));
 
 	/* Set up arguments. */
 	child_argv = (char **)Malloc((argc + 1) * sizeof(char *));
@@ -1240,7 +1279,7 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
     c->pid = pid;
     c->exit_id = AddChild(pid, child_exited);
     c->enabled = true;
-    c->stdoutpipe = stdoutpipe[0];
+    c->stdoutpipe = interactive? -1: stdoutpipe[0];
     task_cb_init_ir_state(&c->ir_state);
 
     /* Clean up our ends of the pipes. */
@@ -1248,7 +1287,9 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
     close(inpipe[1]);
     c->outfd = outpipe[1];
     close(outpipe[0]);
-    close(stdoutpipe[1]);
+    if (!interactive) {
+	close(stdoutpipe[1]);
+    }
 
     /* Link the listeners. */
     c->listeners = listeners; /* struct copy */
@@ -1257,15 +1298,18 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
     c->id = AddInput(c->infd, child_input);
 
     /* Capture child output. */
-    c->stdout_id = AddInput(c->stdoutpipe, child_stdout);
+    c->stdout_id = interactive? NULL_IOID:
+	AddInput(c->stdoutpipe, child_stdout);
 
 #else /*]*/
 
     /* Set up the stdout/stderr output pipes. */
     c = (child_t *)Calloc(1, sizeof(child_t));
-    if (!setup_cr(c)) {
+    if (stdout_redirect && !setup_cr(c)) {
 	Free(c);
 	return false;
+    } else if (!stdout_redirect) {
+	c->cr.collected_eof = true;
     }
     task_cb_init_ir_state(&c->ir_state);
     cr = &c->cr;
@@ -1275,9 +1319,9 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
     startupinfo.cb = sizeof(STARTUPINFO);
     if (stdout_redirect) {
 	startupinfo.hStdOutput = cr->pipe_wr_handle;
+	startupinfo.hStdError = cr->pipe_wr_handle;
+	startupinfo.dwFlags |= STARTF_USESTDHANDLES;
     }
-    startupinfo.hStdError = cr->pipe_wr_handle;
-    startupinfo.dwFlags |= STARTF_USESTDHANDLES;
     memset(&process_information, '\0', sizeof(PROCESS_INFORMATION));
     args = NewString(argv[0]);
     for (i = 1; i < argc; i++) {
@@ -1293,6 +1337,11 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
 	Free(args);
 	args = t;
     }
+
+    /* Export the names of the URL and port into the environment. */
+    SetEnvironmentVariable(URL_ENV, lazyaf("http://127.0.0.1:%u/3270/rest/", httpd_port));
+    SetEnvironmentVariable(PORT_ENV, lazyaf("%d", script_port));
+
     if (CreateProcess(NULL, args, NULL, NULL, TRUE,
 		(stdout_redirect && !share_console)? DETACHED_PROCESS: 0,
 		NULL, NULL, &startupinfo, &process_information) == 0) {
@@ -1301,22 +1350,25 @@ Script_action(ia_t ia, unsigned argc, const char **argv)
 	close_listeners(&listeners);
 
 	/* Let the read thread complete. */
-	CloseHandle(cr->pipe_wr_handle);
-	cr->pipe_wr_handle = INVALID_HANDLE_VALUE;
-	SetEvent(cr->enable_event);
-	WaitForSingleObject(cr->done_event, INFINITE);
-
-	cr_teardown(cr);
+	if (stdout_redirect) {
+	    CloseHandle(cr->pipe_wr_handle);
+	    cr->pipe_wr_handle = INVALID_HANDLE_VALUE;
+	    SetEvent(cr->enable_event);
+	    WaitForSingleObject(cr->done_event, INFINITE);
+	    cr_teardown(cr);
+	}
 	Free(c);
 	Free(args);
 	return false;
     }
 
     Free(args);
-    CloseHandle(process_information.hThread);
-    CloseHandle(cr->pipe_wr_handle);
-    cr->pipe_wr_handle = INVALID_HANDLE_VALUE;
-    SetEvent(cr->enable_event);
+    if (stdout_redirect) {
+	CloseHandle(process_information.hThread);
+	CloseHandle(cr->pipe_wr_handle);
+	cr->pipe_wr_handle = INVALID_HANDLE_VALUE;
+	SetEvent(cr->enable_event);
+    }
 
     /* Create a new script description. */
     llist_init(&c->llist);
