@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2009, 2013-2021 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta, GA
  *  30332.
@@ -38,7 +38,10 @@
  */
 
 #include "globals.h"
+
 #include <errno.h>
+#include <assert.h>
+
 #include "3270ds.h"
 #include "appres.h"
 #include "ctlr.h"
@@ -86,6 +89,9 @@ struct ea *ea_buf;		/* 3270 device buffer */
 				/* ea_buf[-1] is the dummy default field
 				   attribute */
 struct ea *aea_buf;	/* alternate 3270 extended attribute buffer */
+#if defined(CHECK_AEA_BUF) /*[*/
+unsigned long ea_sum, aea_sum;
+#endif /*]*/
 bool formatted = false;	/* set in screen_disp */
 bool screen_changed = false;
 int first_changed = -1;
@@ -110,6 +116,7 @@ static void ctlr_negotiating(bool ignored);
 static void ctlr_connect(bool ignored);
 static int sscp_start;
 static void ctlr_add_ic(int baddr, unsigned char ic);
+static bool ctlr_initted = false;
 
 static void ticking_stop(struct timeval *tp);
 
@@ -185,6 +192,7 @@ ctlr_reinit(unsigned cmask)
     static struct ea *real_ea_buf = NULL;
     static struct ea *real_aea_buf = NULL;
 
+    ctlr_initted = true;
     if (cmask & MODEL_CHANGE) {
 	/* Allocate buffers */
 	if (real_ea_buf) {
@@ -199,16 +207,75 @@ ctlr_reinit(unsigned cmask)
 	real_aea_buf = (struct ea *)Calloc(sizeof(struct ea),
 		(maxROWS * maxCOLS) + 1);
 	aea_buf = real_aea_buf + 1;
+#if defined(CHECK_AEA_BUF) /*[*/
+	ea_sum = 0;
+	aea_sum = 0;
+#endif /*]*/
 	Replace(zero_buf, (unsigned char *)Calloc(sizeof(struct ea),
 		    maxROWS * maxCOLS));
 	cursor_addr = 0;
 	buffer_addr = 0;
 
 	ea_buf[-1].fa  = FA_PRINTABLE | FA_MODIFY;
+	ea_buf[-1].ic  = 1;
 	aea_buf[-1].fa = FA_PRINTABLE | FA_MODIFY;
+	aea_buf[-1].ic = 1;
     }
 }
 
+/*
+ * Checks a model number and oversize rows and columns.
+ * Ideally this should be called by set_rows_cols() below.
+ */
+bool
+check_rows_cols(int mn, unsigned ovc, unsigned ovr)
+{
+    unsigned mxc, mxr; /* Maximum rows, columns */
+
+    switch (mn) {
+    case 2:
+	mxc = MODEL_2_COLS;
+	mxr = MODEL_2_ROWS; 
+	break;
+    case 3:
+	mxc = MODEL_3_COLS;
+	mxr = MODEL_3_ROWS; 
+	break;
+    case 4:
+	mxc = MODEL_4_COLS;
+	mxr = MODEL_4_ROWS; 
+	break;
+    case 5:
+	mxc = MODEL_5_COLS;
+	mxr = MODEL_5_ROWS; 
+	break;
+    default:
+	popup_an_error("Unknown model: %d", mn);
+	return false;
+    }
+
+    /* Check oversize. */
+    if (ovc > 0 || ovr > 0) {
+	if (ovc == 0) {
+	    popup_an_error("Invalid %s %dx%d columns:\nzero", ResOversize, ovc, ovr);
+	    return false;
+	} else if (ovr == 0) {
+	    popup_an_error("Invalid %s %dx%d rows:\nzero", ResOversize, ovc, ovr);
+	    return false;
+	} else if (ovc > MAX_ROWS_COLS || ovr > MAX_ROWS_COLS || ovc * ovr > MAX_ROWS_COLS) {
+	    popup_an_error("Invalid %s %dx%d:\nExceeds protocol limit", ResOversize, ovc, ovr);
+	    return false;
+	} else if (ovc > 0 && ovc < mxc) {
+	    popup_an_error("Invalid %s columns (%d):\nLess than model %d columns (%d)", ResOversize, ovc, mn, mxc);
+	    return false;
+	} else if (ovr > 0 && ovr < mxr) {
+	    popup_an_error("Invalid %s rows (%d):\nLess than model %d rows (%d)", ResOversize, ovr, mn, mxr);
+	    return false;
+	}
+    }
+    return true;
+}
+
 /*
  * Deal with the relationships between model numbers and rows/cols.
  */
@@ -256,17 +323,13 @@ set_rows_cols(int mn, int ovc, int ovr)
     ov_rows = 0;
     if (ovc != 0 || ovr != 0) {
 	if (ovc <= 0 || ovr <= 0) {
-	    popup_an_error("Invalid %s %dx%d:\nNegative or zero", ResOversize,
-		    ovc, ovr);
-	} else if (ovc * ovr >= 0x4000) {
-	    popup_an_error("Invalid %s %dx%d:\nExceeds protocol limit",
-		    ResOversize, ovc, ovr);
+	    popup_an_error("Invalid %s %dx%d:\nNegative or zero", ResOversize, ovc, ovr);
+	} else if (ovc > MAX_ROWS_COLS || ovr > MAX_ROWS_COLS || ovc * ovr > MAX_ROWS_COLS) {
+	    popup_an_error("Invalid %s %dx%d:\nExceeds protocol limit", ResOversize, ovc, ovr);
 	} else if (ovc > 0 && ovc < maxCOLS) {
-	    popup_an_error("Invalid %s cols (%d):\nLess than model %d cols "
-		    "(%d)", ResOversize, ovc, model_num, maxCOLS);
+	    popup_an_error("Invalid %s cols (%d):\nLess than model %d cols (%d)", ResOversize, ovc, model_num, maxCOLS);
 	} else if (ovr > 0 && ovr < maxROWS) {
-	    popup_an_error("Invalid %s rows (%d):\nLess than model %d rows "
-		    "(%d)", ResOversize, ovr, model_num, maxROWS);
+	    popup_an_error("Invalid %s rows (%d):\nLess than model %d rows (%d)", ResOversize, ovr, model_num, maxROWS);
 	} else {
 	    ov_cols = maxCOLS = ovc;
 	    ov_rows = maxROWS = ovr;
@@ -275,7 +338,7 @@ set_rows_cols(int mn, int ovc, int ovr)
 
     /* Update the model name. */
     sprintf(model_name, "327%c-%d%s",
-	mode.m3279 ? '9' : '8',
+	(mode.m3279 && (appres.wrong_terminal_name || model_num < 4)) ? '9' : '8',
 	model_num,
 	mode.extended ? "-E" : "");
 
@@ -290,6 +353,9 @@ set_rows_cols(int mn, int ovc, int ovr)
 
     /* The model changed. */
     st_changed(ST_REMODEL, true);
+    if (ctlr_initted) {
+	ctlr_reinit(MODEL_CHANGE);
+    }
 }
 
 /*
@@ -526,76 +592,87 @@ ctlr_erase(bool alt)
     screen_alt = alt;
 }
 
+/* Restore the keyboard. */
+static void
+restore_keyboard(void)
+{
+    aid = AID_NO;
+    do_reset(false);
+    ticking_stop(&net_last_recv_ts);
+}
+
 /*
  * Interpret an incoming 3270 command.
  */
 enum pds
-process_ds(unsigned char *buf, size_t buflen)
+process_ds(unsigned char *buf, size_t buflen, bool kybd_restore)
 {
-    enum pds rv;
+    enum pds rv = PDS_OKAY_NO_OUTPUT;
 
-    if (!buflen) {
-	return PDS_OKAY_NO_OUTPUT;
+    if (buflen) {
+	scroll_to_bottom();
+
+	switch (buf[0]) {	/* 3270 command */
+	case CMD_EAU:	/* erase all unprotected */
+	case SNA_CMD_EAU:
+	    ctlr_erase_all_unprotected();
+	    trace_ds("< EraseAllUnprotected\n");
+	    break;
+	case CMD_EWA:	/* erase/write alternate */
+	case SNA_CMD_EWA:
+	    ctlr_erase(true);
+	    trace_ds("< EraseWriteAlternate");
+	    rv = ctlr_write(buf, buflen, true);
+	    break;
+	case CMD_EW:	/* erase/write */
+	case SNA_CMD_EW:
+	    ctlr_erase(false);
+	    trace_ds("< EraseWrite");
+	    rv = ctlr_write(buf, buflen, true);
+	    break;
+	case CMD_W:	/* write */
+	case SNA_CMD_W:
+	    trace_ds("< Write");
+	    rv = ctlr_write(buf, buflen, false);
+	    break;
+	case CMD_RB:	/* read buffer */
+	case SNA_CMD_RB:
+	    trace_ds("< ReadBuffer\n");
+	    ctlr_read_buffer(aid);
+	    rv = PDS_OKAY_OUTPUT;
+	    break;
+	case CMD_RM:	/* read modifed */
+	case SNA_CMD_RM:
+	    trace_ds("< ReadModified\n");
+	    ctlr_read_modified(aid, false);
+	    rv = PDS_OKAY_OUTPUT;
+	    break;
+	case CMD_RMA:	/* read modifed all */
+	case SNA_CMD_RMA:
+	    trace_ds("< ReadModifiedAll\n");
+	    ctlr_read_modified(aid, true);
+	    rv = PDS_OKAY_OUTPUT;
+	    break;
+	case CMD_WSF:	/* write structured field */
+	case SNA_CMD_WSF:
+	    trace_ds("< WriteStructuredField");
+	    rv = write_structured_field(buf, buflen);
+	    break;
+	case CMD_NOP:	/* no-op */
+	    trace_ds("< NoOp\n");
+	    break;
+	default:
+	    /* unknown 3270 command */
+	    popup_an_error("Unknown 3270 Data Stream command: X'%X'\n", buf[0]);
+	    rv = PDS_BAD_CMD;
+	    break;
+	}
     }
 
-    scroll_to_bottom();
-
-    switch (buf[0]) {	/* 3270 command */
-    case CMD_EAU:	/* erase all unprotected */
-    case SNA_CMD_EAU:
-	ctlr_erase_all_unprotected();
-	trace_ds("< EraseAllUnprotected\n");
-	return PDS_OKAY_NO_OUTPUT;
-    case CMD_EWA:	/* erase/write alternate */
-    case SNA_CMD_EWA:
-	ctlr_erase(true);
-	trace_ds("< EraseWriteAlternate");
-	if ((rv = ctlr_write(buf, buflen, true)) < 0) {
-	    return rv;
-	}
-	return PDS_OKAY_NO_OUTPUT;
-    case CMD_EW:	/* erase/write */
-    case SNA_CMD_EW:
-	ctlr_erase(false);
-	trace_ds("< EraseWrite");
-	if ((rv = ctlr_write(buf, buflen, true)) < 0) {
-	    return rv;
-	}
-	return PDS_OKAY_NO_OUTPUT;
-    case CMD_W:	/* write */
-    case SNA_CMD_W:
-	trace_ds("< Write");
-	if ((rv = ctlr_write(buf, buflen, false)) < 0) {
-	    return rv;
-	}
-	return PDS_OKAY_NO_OUTPUT;
-    case CMD_RB:	/* read buffer */
-    case SNA_CMD_RB:
-	trace_ds("< ReadBuffer\n");
-	ctlr_read_buffer(aid);
-	return PDS_OKAY_OUTPUT;
-    case CMD_RM:	/* read modifed */
-    case SNA_CMD_RM:
-	trace_ds("< ReadModified\n");
-	ctlr_read_modified(aid, false);
-	return PDS_OKAY_OUTPUT;
-    case CMD_RMA:	/* read modifed all */
-    case SNA_CMD_RMA:
-	trace_ds("< ReadModifiedAll\n");
-	ctlr_read_modified(aid, true);
-	return PDS_OKAY_OUTPUT;
-    case CMD_WSF:	/* write structured field */
-    case SNA_CMD_WSF:
-	trace_ds("< WriteStructuredField");
-	return write_structured_field(buf, buflen);
-    case CMD_NOP:	/* no-op */
-	trace_ds("< NoOp\n");
-	return PDS_OKAY_NO_OUTPUT;
-    default:
-	/* unknown 3270 command */
-	popup_an_error("Unknown 3270 Data Stream command: 0x%X\n", buf[0]);
-	return PDS_BAD_CMD;
+    if (kybd_restore) {
+	restore_keyboard();
     }
+    return rv;
 }
 
 /*
@@ -1284,10 +1361,15 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 			formatted = true; \
 		}
 
+#define WRITE_ERROR "Host write error:\n"
+#define TOO_SHORT "Record too short, "
+
     kybd_inhibit(false);
 
     if (buflen < 2) {
-	return PDS_BAD_CMD;
+	/* Need flags at minimum. */
+	popup_an_error(WRITE_ERROR TOO_SHORT "missing write flags");
+	return PDS_BAD_ADDR;
     }
 
     default_fg = 0;
@@ -1337,15 +1419,17 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
     last_zpt = false;
     current_fa = get_field_attribute(buffer_addr);
 
+#define END_WRITE END_TEXT("\n")
 #define ABORT_WRITEx { \
     rv = PDS_BAD_ADDR; \
     aborted = true; \
     break; \
 }
 #define ABORT_WRITE(s) { \
-    trace_ds(" [" s "; write aborted]\n"); \
+    END_WRITE; \
+    popup_an_error(WRITE_ERROR s); \
     ABORT_WRITEx; \
-} \
+}
 
     for (cp = &buf[2]; !aborted && cp < (buf + buflen); cp++) {
 	switch (*cp) {
@@ -1356,6 +1440,9 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    }
 	    previous = ORDER;
 	    cp++;		/* skip field attribute */
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing SF attributes");
+	    }
 	    START_FIELD(*cp);
 	    ctlr_add_fg(buffer_addr, 0);
 	    ctlr_add_bg(buffer_addr, 0);
@@ -1364,14 +1451,19 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    last_zpt = false;
 	    break;
 	case ORDER_SBA:	/* set buffer address */
-	    cp += 2;	/* skip buffer address */
-	    buffer_addr = DECODE_BADDR(*(cp-1), *cp);
 	    END_TEXT("SetBufferAddress");
+	    cp += 2;	/* skip buffer address */
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing SBA address");
+	    }
+	    buffer_addr = DECODE_BADDR(*(cp-1), *cp);
 	    previous = SBA;
 	    trace_ds("%s", rcba(buffer_addr));
 	    if (buffer_addr >= COLS * ROWS) {
-		trace_ds("COLS %d ROWS %d\n", COLS, ROWS);
-		ABORT_WRITE("invalid SBA address");
+		END_WRITE;
+		popup_an_error(WRITE_ERROR "SBA address %d > maximum %d",
+			buffer_addr, (COLS * ROWS) - 1);
+		ABORT_WRITEx;
 	    }
 	    current_fa = get_field_attribute(buffer_addr);
 	    last_cmd = true;
@@ -1441,9 +1533,21 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	case ORDER_RA:	/* repeat to address */
 	    END_TEXT("RepeatToAddress");
 	    cp += 2;	/* skip buffer address */
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing RA address");
+	    }
 	    baddr = DECODE_BADDR(*(cp-1), *cp);
 	    trace_ds("%s", rcba(baddr));
+	    if (baddr >= COLS * ROWS) {
+		END_WRITE;
+		popup_an_error(WRITE_ERROR "RA address %d > maximum %d",
+			baddr, (COLS * ROWS) - 1);
+		ABORT_WRITEx;
+	    }
 	    cp++;		/* skip char to repeat */
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing RA character");
+	    }
 	    add_dbcs = false;
 	    ra_ge = false;
 	    previous = ORDER;
@@ -1463,7 +1567,8 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		add_c1 = *cp;
 		cp++;
 		if (cp >= buf + buflen) {
-		    ABORT_WRITE("missing second half of DBCS character");
+		    ABORT_WRITE(TOO_SHORT "missing second half of RA DBCS "
+			    "character");
 		}
 		add_c2 = *cp;
 		if (add_c1 == EBC_null) {
@@ -1477,15 +1582,18 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		    case EBC_fm:
 			break;
 		    default:
-			trace_ds(" [invalid DBCS RA control character "
-				"X'%02x%02x'; write aborted]", add_c1, add_c2);
+			END_WRITE;
+			popup_an_error(WRITE_ERROR "Invalid DBCS RA "
+				"control character X'%02X%02X'", add_c1,
+				add_c2);
 			ABORT_WRITEx;
 		    }
 		} else if (add_c1 < 0x40 || add_c1 > 0xfe || add_c2 < 0x40 ||
 			add_c2 > 0xfe) {
-		    trace_ds(" [invalid DBCS RA character X'%02x%02x'; write "
-			    "aborted]", add_c1, add_c2);
-			ABORT_WRITEx;
+		    END_WRITE;
+		    popup_an_error(WRITE_ERROR "Invalid DBCS RA "
+			    "character X'%02X%02X'", add_c1, add_c2);
+		    ABORT_WRITEx;
 		}
 		ebcdic_to_multibyte((add_c1 << 8) | add_c2, mb, sizeof(mb));
 		trace_ds("'%s'", mb);
@@ -1494,6 +1602,9 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		    ra_ge = true;
 		    trace_ds("GraphicEscape");
 		    cp++;
+		    if (cp >= buf + buflen) {
+			ABORT_WRITE(TOO_SHORT "missing RA GE character");
+		    }
 		}
 		add_c1 = *cp;
 		if (add_c1) {
@@ -1504,9 +1615,6 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		    trace_ds("'");
 		}
 	    }
-	    if (baddr >= COLS * ROWS) {
-		ABORT_WRITE("invalid RA address");
-	    }
 	    do {
 		if (add_dbcs) {
 		    ctlr_add(buffer_addr, add_c1, default_cs);
@@ -1520,6 +1628,7 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		    }
 		}
 		ctlr_add_fg(buffer_addr, default_fg);
+		ctlr_add_bg(buffer_addr, default_bg);
 		ctlr_add_gr(buffer_addr, default_gr);
 		ctlr_add_ic(buffer_addr, default_ic);
 		INC_BA(buffer_addr);
@@ -1537,15 +1646,19 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    last_zpt = false;
 	    break;
 	case ORDER_EUA:	/* erase unprotected to address */
-	    cp += 2;	/* skip buffer address */
-	    baddr = DECODE_BADDR(*(cp-1), *cp);
 	    END_TEXT("EraseUnprotectedAll");
-	    if (previous != SBA) {
-		trace_ds("%s", rcba(baddr));
+	    cp += 2;	/* skip buffer address */
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing EUA address");
 	    }
+	    baddr = DECODE_BADDR(*(cp-1), *cp);
+	    trace_ds("%s", rcba(baddr));
 	    previous = ORDER;
 	    if (baddr >= COLS * ROWS) {
-		ABORT_WRITE("invalid EUA address");
+		END_WRITE;
+		popup_an_error(WRITE_ERROR "EUA address %d > maximum %d",
+			baddr, (COLS * ROWS) - 1);
+		ABORT_WRITEx;
 	    }
 	    d = ctlr_lookleft_state(buffer_addr, &why);
 	    if (d == DBCS_RIGHT) {
@@ -1571,6 +1684,9 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    /* XXX: DBCS? */
 	    END_TEXT("GraphicEscape ");
 	    cp++;		/* skip char */
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing GE character");
+	    }
 	    previous = ORDER;
 	    if (*cp) {
 		trace_ds("'");
@@ -1597,10 +1713,16 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    }
 	    previous = ORDER;
 	    cp++;
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing MF count");
+	    }
 	    na = *cp;
 	    if (ea_buf[buffer_addr].fa) {
 		for (i = 0; i < (int)na; i++) {
 		    cp++;
+		    if (cp + 1 >= buf + buflen) {
+			ABORT_WRITE(TOO_SHORT "missing MF attribute");
+		    }
 		    if (*cp == XA_3270) {
 			trace_ds(" 3270");
 			cp++;
@@ -1659,6 +1781,9 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    }
 	    previous = ORDER;
 	    cp++;	/* skip order */
+	    if (cp >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing SFE count");
+	    }
 	    na = *cp;
 	    any_fa = 0;
 	    efa_fg = 0;
@@ -1668,6 +1793,9 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    efa_ic = 0;
 	    for (i = 0; i < (int)na; i++) {
 		cp++;
+		if (cp + 1 >= buf + buflen) {
+		    ABORT_WRITE(TOO_SHORT "missing SFE attribute");
+		}
 		if (*cp == XA_3270) {
 		    trace_ds(" 3270");
 		    cp++;
@@ -1729,6 +1857,9 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    END_TEXT("SetAttribute");
 	    previous = ORDER;
 	    cp++;
+	    if (cp + 1 >= buf + buflen) {
+		ABORT_WRITE(TOO_SHORT "missing SA attribute");
+	    }
 	    if (*cp == XA_FOREGROUND)  {
 		trace_ds("%s", see_efa(*cp, *(cp + 1)));
 		if (mode.m3279) {
@@ -1789,7 +1920,7 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    previous = ORDER;
 	    d = ctlr_lookleft_state(buffer_addr, &why);
 	    if (default_cs == CS_DBCS || d != DBCS_NONE) {
-		ABORT_WRITE("invalid format control order in DBCS field");
+		ABORT_WRITE("Invalid format control order in DBCS field");
 	    }
 	    ctlr_add(buffer_addr, *cp, default_cs);
 	    ctlr_add_fg(buffer_addr, default_fg);
@@ -1811,7 +1942,7 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		ABORT_WRITE("SO in DBCS field");
 	    }
 	    if (d != DBCS_NONE && why == DBCS_SUBFIELD) {
-		ABORT_WRITE("double SO");
+		ABORT_WRITE("Double SO");
 	    }
 	    /* All is well. */
 	    previous = ORDER;
@@ -1841,7 +1972,7 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		   ((fa_addr >= 0 && baddr != fa_addr) ||
 		    (fa_addr < 0 && baddr != ROWS*COLS - 1))) {
 		if (ea_buf[baddr].ec == FCORDER_SI) {
-		    ABORT_WRITE("double SI");
+		    ABORT_WRITE("Double SI");
 		}
 		if (ea_buf[baddr].ec == FCORDER_SO) {
 		    break;
@@ -1875,7 +2006,7 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		add_c1 = EBC_null;
 		cp++;
 		if (cp >= buf + buflen) {
-		    ABORT_WRITE("missing second half of DBCS character");
+		    ABORT_WRITE("Missing second half of DBCS character");
 		}
 		add_c2 = *cp;
 		switch (add_c2) {
@@ -1897,8 +2028,9 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 		    cp--;
 		    break;
 		default:
-		    trace_ds(" [invalid DBCS control character X'%02x%02x'; "
-			    "write aborted]", add_c1, add_c2);
+		    END_WRITE;
+		    popup_an_error(WRITE_ERROR "Invalid DBCS control "
+			    "character X'%02X%02X'", add_c1, add_c2);
 		    ABORT_WRITEx;
 		    break;
 		}
@@ -1943,19 +2075,20 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 	    add_dbcs = false;
 	    d = ctlr_lookleft_state(buffer_addr, &why);
 	    if (d == DBCS_RIGHT) {
-		ABORT_WRITE("overwriting right half of DBCS character");
+		ABORT_WRITE("Overwriting right half of DBCS character");
 	    }
 	    if (d != DBCS_NONE || default_cs == CS_DBCS) {
 		add_c1 = *cp;
 		cp++;
 		if (cp >= buf + buflen) {
-		    ABORT_WRITE("missing second half of DBCS character");
+		    ABORT_WRITE("Missing second half of DBCS character");
 		}
 		add_c2 = *cp;
 		if (add_c1 < 0x40 || add_c1 > 0xfe ||
 		    add_c2 < 0x40 || add_c2 > 0xfe) {
-		    trace_ds(" [invalid DBCS character X'%02x%02x'; write "
-			    "aborted]", add_c1, add_c2);
+		    END_WRITE;
+		    popup_an_error(WRITE_ERROR "Invalid DBCS character "
+			    "X'%02X%02X'", add_c1, add_c2);
 		    ABORT_WRITEx;
 		}
 		add_dbcs = true;
@@ -2026,6 +2159,7 @@ ctlr_write(unsigned char buf[], size_t buflen, bool erase)
 #undef START_FIELD
 #undef END_TEXT0
 #undef END_TEXT
+#undef WRITE_ERROR
 #undef ABORT_WRITEx
 #undef ABORT_WRITE
 
@@ -2050,7 +2184,7 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
      * we display other control codes as spaces.
      */
 
-    trace_ds("SSCP-LU data\n< ");
+    trace_ds("SSCP-LU data\n<");
     for (i = 0; i < buflen; cp++, i++) {
 	switch (*cp) {
 	case FCORDER_NL:
@@ -2072,6 +2206,10 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
 		ctlr_add_ic(buffer_addr, default_ic);
 		INC_BA(buffer_addr);
 	    }
+	    if (buffer_addr == 0) {
+		ctlr_scroll(0, 0);
+		buffer_addr = (ROWS - 1) * COLS;
+	    }
 	    break;
 
 	case ORDER_SF:
@@ -2082,14 +2220,22 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
 		trace_ds("'");
 		text = false;
 	    }
-	    trace_ds(" SF%s %s [translated to space]\n", rcba(buffer_addr),
-		    see_attr(*cp));
+	    if (cp >= buf + buflen) {
+		trace_ds(" SF%s [translated to space]\n", rcba(buffer_addr));
+	    } else {
+		trace_ds(" SF%s %s [translated to space]\n", rcba(buffer_addr),
+			see_attr(*cp));
+	    }
 	    ctlr_add(buffer_addr, EBC_space, default_cs);
 	    ctlr_add_fg(buffer_addr, default_fg);
 	    ctlr_add_bg(buffer_addr, default_bg);
 	    ctlr_add_gr(buffer_addr, default_gr);
 	    ctlr_add_ic(buffer_addr, default_ic);
 	    INC_BA(buffer_addr);
+	    if (buffer_addr == 0) {
+		ctlr_scroll(0, 0);
+		buffer_addr = (ROWS - 1) * COLS;
+	    }
 	    break;
 	case ORDER_IC:
 	    if (text) {
@@ -2099,8 +2245,12 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
 	    trace_ds(" IC%s [ignored]\n", rcba(buffer_addr));
 	    break;
 	case ORDER_SBA:
-	    baddr = DECODE_BADDR(*(cp+1), *(cp+2));
-	    trace_ds(" SBA%s [ignored]\n", rcba(baddr));
+	    if (cp + 2 >= buf + buflen) {
+		trace_ds(" SBA [ignored]\n");
+	    } else {
+		baddr = DECODE_BADDR(*(cp+1), *(cp+2));
+		trace_ds(" SBA%s [ignored]\n", rcba(baddr));
+	    }
 	    cp += 2;
 	    i += 2;
 	    break;
@@ -2108,6 +2258,11 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
 	case ORDER_GE:
 	    cp++;
 	    if (++i >= buflen) {
+		if (text) {
+		    trace_ds("'");
+		    text = false;
+		}
+		trace_ds(" GE");
 		break;
 	    }
 	    if (*cp <= 0x40) {
@@ -2126,6 +2281,10 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
 	    ctlr_add_gr(buffer_addr, default_gr);
 	    ctlr_add_ic(buffer_addr, default_ic);
 	    INC_BA(buffer_addr);
+	    if (buffer_addr == 0) {
+		ctlr_scroll(0, 0);
+		buffer_addr = (ROWS - 1) * COLS;
+		}
 	    break;
 
 	default:
@@ -2140,6 +2299,10 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
 	    ctlr_add_gr(buffer_addr, default_gr);
 	    ctlr_add_ic(buffer_addr, default_ic);
 	    INC_BA(buffer_addr);
+	    if (buffer_addr == 0) {
+		ctlr_scroll(0, 0);
+		buffer_addr = (ROWS - 1) * COLS;
+	    }
 	    break;
 	}
     }
@@ -2158,6 +2321,14 @@ ctlr_write_sscp_lu(unsigned char buf[], size_t buflen)
     task_host_output();
 }
 
+void
+ctlr_sscp_up(void)
+{
+    if (sscp_start > COLS) {
+	sscp_start -= COLS;
+    }
+}
+
 /*
  * Determine the DBCS state of a buffer location strictly by looking left.
  * Used only to validate write operations.
@@ -2445,7 +2616,7 @@ ps_process(void)
     if (ft_state != FT_NONE &&      /* transfer in progress */
 	    formatted &&                /* screen is formatted */
 	    !screen_alt &&              /* 24x80 screen */
-	    !kybdlock &&                /* keyboard not locked */
+	    !(kybdlock & ~KL_FT) &&     /* keyboard not locked */
 	    /* magic field */
 	    ea_buf[1919].fa && FA_IS_SKIP(ea_buf[1919].fa)) {
 	ft_cut_data();
@@ -2820,20 +2991,52 @@ ctlr_changed(int bstart, int bend)
     REGION_CHANGED(bstart, bend);
 }
 
+
+#if defined(CHECK_AEA_BUF) /*[*/
+/*
+ * Compute a simple checksum.
+ */
+static unsigned long
+csum(struct ea *ea)
+{
+    size_t i;
+    unsigned long sum = 0;
+    unsigned char *c = (unsigned char *)(void *)ea;
+
+    for (i = 0; i < (ROWS * COLS) * sizeof(struct ea); i++) {
+	sum += c[i];
+    }
+    return sum;
+}
+#endif /*]*/
+
 /*
  * Swap the regular and alternate screen buffers
  */
 void
 ctlr_altbuffer(bool alt)
 {
-    struct ea *etmp;
-
     if (alt != is_altbuffer) {
+	struct ea *etmp;
+#if defined(CHECK_AEA_BUF) /*[*/
+	unsigned long stmp;
+#endif /*]*/
 
 	etmp = ea_buf;
 	ea_buf = aea_buf;
 	aea_buf = etmp;
 
+#if defined(CHECK_AEA_BUF) /*[*/
+	stmp = ea_sum;
+	ea_sum = aea_sum;
+	aea_sum = stmp;
+
+	if (ea_sum != 0) {
+	    assert(csum(ea_buf) == ea_sum);
+	}
+	aea_sum = csum(aea_buf);
+#endif /*]*/
+
 	is_altbuffer = alt;
 	ALL_CHANGED;
 	unselect(0, ROWS*COLS);
