@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1994-2015, 2018-2021 Paul Mattes.
+ * Copyright (c) 1994-2022 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -310,7 +310,7 @@ fprint_screen_start(FILE *f, ptype_t ptype, unsigned opts, const char *caption,
 		fprintf(f, "<html>\n"
 		   "<head>\n"
 		   " <meta http-equiv=\"Content-Type\" "
-		     "content=\"text/html; charset=UTF-8\">\n"
+		     "content=\"text/html; charset=utf-8\">\n"
 		   "</head>\n"
 		   " <body>\n") < 0) {
 	    rv = FPS_STATUS_ERROR;
@@ -376,6 +376,16 @@ fprint_screen_start(FILE *f, ptype_t ptype, unsigned opts, const char *caption,
     return rv;
 }
 
+/* Get the DBCS state for part of the screen, real or imagined. */
+static enum dbcs_state
+protected_dbcs_state(int baddr)
+{
+    if (baddr < ROWS * COLS) {
+	return ctlr_dbcs_state(baddr);
+    }
+    return DBCS_NONE;
+}
+
 #define FAIL do { \
     rv = -1; \
     goto done; \
@@ -416,11 +426,15 @@ fprint_screen_body(fps_t ofps)
     }
 
     if (fps->opts & FPS_OIA) {
+	int fa;
+
 	xea = (struct ea *)Calloc(1 + ((ROWS + 2) * COLS), sizeof(struct ea));
 	lazya(xea);
 	memcpy(xea, ea_buf - 1, (1 + (ROWS * COLS)) * sizeof(struct ea));
 	xea++;
 	vstatus_line(xea + (ROWS * COLS));
+	fa = find_field_attribute((ROWS * COLS) - 1);
+	xea[((ROWS + 2) * COLS) - 1] = ea_buf[fa]; /* struct copy */
 	xrows = ROWS + 2;
     } else {
 	xea = ea_buf;
@@ -456,7 +470,7 @@ fprint_screen_body(fps_t ofps)
     fa_underline = xea[fa_addr].gr & GR_UNDERLINE;
     current_underline = fa_underline;
     fa_reverse = xea[fa_addr].gr & GR_REVERSE;
-    current_reverse = fa_underline;
+    current_reverse = fa_reverse;
 
     switch (fps->ptype) {
     case P_RTF:
@@ -589,19 +603,19 @@ fprint_screen_body(fps_t ofps)
 	if (FA_IS_ZERO(fa) &&
 		(FA_IS_PROTECTED(fa) ||
 		 !(fps->opts & FPS_INCLUDE_ZERO_INPUT))) {
-	    if (ctlr_dbcs_state(i) == DBCS_LEFT) {
+	    if (protected_dbcs_state(i) == DBCS_LEFT) {
 		uc = 0x3000;
 	    } else {
 		uc = ' ';
 	    }
 	} else if (is_nvt(&xea[i], false, &uc)) {
 	    /* NVT-mode text. */
-	    if (ctlr_dbcs_state(i) == DBCS_RIGHT) {
+	    if (protected_dbcs_state(i) == DBCS_RIGHT) {
 		continue;
 	    }
 	} else {
 	    /* Convert EBCDIC to Unicode. */
-	    switch (ctlr_dbcs_state(i)) {
+	    switch (protected_dbcs_state(i)) {
 	    case DBCS_NONE:
 	    case DBCS_SB:
 		uc = ebcdic_to_unicode(xea[i].ec, xea[i].cs, EUO_NONE);
@@ -641,7 +655,9 @@ fprint_screen_body(fps_t ofps)
 	    bool underline;
 	    bool reverse;
 
-	    if (xea[i].gr & GR_INTENSIFY) {
+	    if (xea[i].fa) {
+		high = false;
+	    } else if (xea[i].gr & GR_INTENSIFY) {
 		high = true;
 	    } else {
 		high = fa_high;
@@ -658,7 +674,9 @@ fprint_screen_body(fps_t ofps)
 		}
 		current_high = high;
 	    }
-	    if (xea[i].gr & GR_UNDERLINE) {
+	    if (xea[i].fa) {
+		underline = false;
+	    } else if (xea[i].gr & GR_UNDERLINE) {
 		underline = true;
 	    } else {
 		underline = fa_underline;
@@ -675,7 +693,9 @@ fprint_screen_body(fps_t ofps)
 		}
 		current_underline = underline;
 	    }
-	    if (xea[i].gr & GR_REVERSE) {
+	    if (xea[i].fa) {
+		reverse = false;
+	    } else if (xea[i].gr & GR_REVERSE) {
 		reverse = true;
 	    } else {
 		reverse = fa_reverse;
@@ -700,39 +720,51 @@ fprint_screen_body(fps_t ofps)
 	    int fg_color, bg_color;
 	    bool high;
 	    bool underline;
+	    bool reverse;
 
 	    if (xea[i].fg) {
 		fg_color = xea[i].fg & 0x0f;
 	    } else {
 		fg_color = fa_fg;
 	    }
-	    if (xea[i].bg) {
+	    if (xea[i].fa) {
+		bg_color = HOST_COLOR_BLACK;
+	    } else if (xea[i].bg) {
 		bg_color = xea[i].bg & 0x0f;
 	    } else {
 		bg_color = fa_bg;
 	    }
-	    if (xea[i].gr & GR_REVERSE) {
-		int tmp;
-
-		tmp = fg_color;
-		fg_color = bg_color;
-		bg_color = tmp;
+	    if (xea[i].fa) {
+		reverse = false;
+	    } else if (xea[i].gr & GR_REVERSE) {
+		reverse = true;
+	    } else {
+		reverse = fa_reverse;
+	    }
+	    if (xea[i].fa) {
+		high = false;
+	    } else if (xea[i].gr & GR_INTENSIFY) {
+		high = true;
+	    } else {
+		high = fa_high;
+	    }
+	    if (xea[i].fa) {
+		underline = false;
+	    } else if (xea[i].gr & GR_UNDERLINE) {
+		underline = true;
+	    } else {
+		underline = fa_underline;
 	    }
 
 	    if (i == cursor_addr) {
 		fg_color = (bg_color == HOST_COLOR_RED)?
 		    HOST_COLOR_BLACK: bg_color;
 		bg_color = HOST_COLOR_RED;
-	    }
-	    if (xea[i].gr & GR_INTENSIFY) {
-		high = true;
-	    } else {
-		high = fa_high;
-	    }
-	    if (xea[i].gr & GR_UNDERLINE) {
-		underline = true;
-	    } else {
-		underline = fa_underline;
+	    } else if (reverse) {
+		int tmp = fg_color;
+
+		fg_color = bg_color;
+		bg_color = tmp;
 	    }
 
 	    if (fg_color != current_fg ||
