@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1993-2020 Paul Mattes.
- * Copyright (c) 1990, Jeff Sparkes.
- * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta, GA
+ * Copyright (c) 1993-2023 Paul Mattes.
+ * Copyright (c) 1990 Jeff Sparkes.
+ * Copyright (c) 1989 Georgia Tech Research Corporation (GTRC), Atlanta, GA
  *  30332.
  * All rights reserved.
  *
@@ -59,6 +59,7 @@
 #include "opts.h"
 #include "product.h"
 #include "readres.h"
+#include "resolver.h"
 #include "screen.h"
 #include "selectc.h"
 #include "sio.h"
@@ -87,9 +88,10 @@ static void parse_local_process(int *argcp, const char **argv,
 #endif /*]*/
 static void set_appres_defaults(void);
 static void parse_options(int *argcp, const char **argv, bool warn);
-static void parse_set_clear(int *argcp, const char **argv, bool warn);
 static int parse_model_number(char *m);
 static void xparse_xrm(const char *arg, const char *where, bool warn);
+static void parse_set(const char *arg, const char *where, bool warn);
+static void parse_clear(const char *arg, const char *where, bool warn);
 static merge_profile_t *merge_profilep = NULL;
 static char *session_suffix[4];
 static size_t session_suffix_len[4];
@@ -99,6 +101,7 @@ unsigned sorted_help_count = 0;
 
 /* Globals */
 const char     *programname;
+bool		supports_cmdline_host = true;
 char		full_model_name[13] = "IBM-";
 char	       *model_name = &full_model_name[4];
 AppRes          appres;
@@ -152,6 +155,7 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
     const char **xargv;
     bool read_session_or_profile = false;
     int suffix_match = -1;
+    char *s;
 
     /* Figure out who we are */
 #if defined(_WIN32) /*[*/
@@ -167,7 +171,7 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
 	char *path = getenv("PATH");
 
 	/* Add our path to $PATH so we can find x3270if. */
-	putenv(xs_buffer("PATH=%.*s%s%s",
+	(void) putenv(xs_buffer("PATH=%.*s%s%s",
 		    (int)(programname - argv[0]), argv[0],
 		    path? ":": "",
 		    path? path: ""));
@@ -219,9 +223,6 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
     /* Parse command-line options. */
     parse_options(&argc, argv, true);
 
-    /* Pick out the remaining -set and -clear toggle options. */
-    parse_set_clear(&argc, argv, true);
-
     /* Now figure out if there's a hostname. */
     for (hn_argc = 1; hn_argc < argc; hn_argc++) {
 	if (!strcmp(argv[hn_argc], LAST_ARG)) {
@@ -238,12 +239,18 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
 	*cl_hostname = argv[1];
 	break;
     case 3:
-	no_minus(argv[1]);
-	no_minus(argv[2]);
-	*cl_hostname = xs_buffer("%s:%s", argv[1], argv[2]);
-	break;
+	if (supports_cmdline_host) {
+	    no_minus(argv[1]);
+	    no_minus(argv[2]);
+	    *cl_hostname = xs_buffer("%s:%s", argv[1], argv[2]);
+	    break;
+	}
+	/* else fall through... */
     default:
-	usage("Too many command-line arguments");
+	for (i = 0; i < hn_argc; i++) {
+	    no_minus(argv[i]);
+	}
+	usage("Too many command-line options");
 	break;
     }
 
@@ -261,8 +268,13 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
     if (n_session_suffixes == 0) {
 	add_session_suffix(xs_buffer(".%s", app));
 #if defined(_WIN32) /*[*/
-	add_session_suffix(xs_buffer(".%s", app + 1));
-	add_session_suffix(xs_buffer(".%.3s", app));
+	if (app[0] == 'w') {
+	    add_session_suffix(xs_buffer(".%s", app + 1));
+	    add_session_suffix(xs_buffer(".%.3s", app));
+	} else {
+	    add_session_suffix(xs_buffer(".w%s", app));
+	    add_session_suffix(xs_buffer(".w%.2s", app));
+	}
 #endif /*]*/
 	if (appres.alias != NULL) {
 	    add_session_suffix(xs_buffer(".%s", appres.alias));
@@ -289,9 +301,12 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
 	Replace(profile_path, NewString(*cl_hostname));
 	profile_name[strlen(profile_name) - session_suffix_len[suffix_match]]
 	    = '\0';
-	*cl_hostname = appres.hostname; /* might be NULL */
+	*cl_hostname = supports_cmdline_host? appres.hostname: NULL;
     } else {
 	/* There is no session file. */
+	if (*cl_hostname && !supports_cmdline_host) {
+	    usage("Unknown command-line argument");
+	}
 
 	/* For c3270 only, read in the c3270 profile (~/.c3270pro). */
 	if (merge_profilep != NULL) {
@@ -314,9 +329,8 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
      */
     if (read_session_or_profile) {
 	parse_options(&xargc, xargv, false);
-	parse_set_clear(&xargc, xargv, false);
     }
-    /* Can't free xcmd, parts of it are still in use. */
+    Free(xcmd);
     Free((char *)xargv);
 
     /*
@@ -342,13 +356,17 @@ parse_command_line(int argc, const char **argv, const char **cl_hostname)
      * It's an alias for 'codePage', but it doesn't override it.
      */
     if (appres.codepage == NULL) {
-	appres.codepage = appres.charset;
+	appres.codepage = NewString(appres.charset);
     }
     if (appres.codepage == NULL) {
 	appres.codepage = NewString("bracket");
     }
 
-    appres.termname = clean_termname(appres.termname);
+    s = clean_termname(appres.termname);
+    Replace(appres.termname, s);
+
+    /* Set up the resolver. */
+    set_46(appres.prefer_ipv4, appres.prefer_ipv6);
 
     return argc;
 }
@@ -410,7 +428,7 @@ static void
 no_minus(const char *arg)
 {
     if (arg[0] == '-') {
-	usage(xs_buffer("Unknown or incomplete option: %s", arg));
+	usage(xs_buffer("Unknown or incomplete option: '%s'", arg));
     }
 }
 
@@ -458,11 +476,11 @@ set_appres_defaults(void)
     mode.extended = true;
     mode.m3279 = true;
     appres.debug_tracing = true;
-    appres.conf_dir = LIBX3270DIR;
+    appres.conf_dir = NewString(LIBX3270DIR);
 
     appres.model = NewString("3279-4-E");
     appres.hostsfile = NULL;
-    appres.port = "23";
+    appres.port = NewString("23");
     /* Do this when we finally deprecate 'charset'. */
     /* appres.codepage = NewString("bracket"); */
     appres.termname = NULL;
@@ -472,6 +490,7 @@ set_appres_defaults(void)
 #endif /*]*/
     appres.oversize = NULL;
     appres.bind_limit = true;
+    appres.contention_resolution = true;
     appres.new_environ = true;
     appres.max_recent = 5;
 
@@ -495,6 +514,7 @@ set_appres_defaults(void)
     set_toggle(TYPEAHEAD, true);
     set_toggle(BLANK_FILL, true);
     set_toggle(OVERLAY_PASTE, true);
+    set_toggle(UNDERSCORE_BLANK_FILL, true);
 
 #if defined(_WIN32) /*[*/
     appres.local_cp = GetACP();
@@ -516,11 +536,11 @@ set_appres_defaults(void)
 
 static opt_t base_opts[] = {
 { OptAlias,    OPT_STRING,  false, ResAlias,   aoffset(alias),
-    "<name>", "Define application alias for -xrm and session file suffix" },
+    "<name>", "Define application alias for " OptXrm " and session file suffix" },
 { OptCharset,  OPT_STRING,  false, ResCodePage, aoffset(codepage),
     NULL, NULL },
-{ OptClear,    OPT_SKIP2,   false, NULL,         NULL,
-    "<toggle>", "Turn on <toggle>" },
+{ OptClear,    OPT_CLEAR,   false, NULL,         NULL,
+    "<resource>", "Set <resource> to " ResFalse },
 { OptCodePage,  OPT_STRING,  false, ResCodePage, aoffset(codepage),
     "<name>", "Use host ECBDIC code page <name>"},
 { OptConnectTimeout, OPT_INT,false,ResConnectTimeout,aoffset(connect_timeout),
@@ -554,14 +574,18 @@ static opt_t base_opts[] = {
     "<cols>x<rows>", "Larger screen dimensions" },
 { OptPort,     OPT_STRING,  false, ResPort,      aoffset(port),
     "<port>", "Default TELNET port" },
+{ OptPreferIpv4, OPT_BOOLEAN, true, ResPreferIpv4, aoffset(prefer_ipv4),
+    NULL,	"Prefer IPv4 host addresses" },
+{ OptPreferIpv6, OPT_BOOLEAN, true, ResPreferIpv6, aoffset(prefer_ipv6),
+    NULL,	"Prefer IPv6 host addresses" },
 { OptProxy,    OPT_STRING,  false, ResProxy,     aoffset(proxy),
     "<type>:<host>[:<port>]", "Proxy type and server" },
 { OptScriptPort,OPT_STRING, false, ResScriptPort, aoffset(script_port),
     "[<addr>:]<port>", "TCP port to listen on for script commands" },
 { OptScriptPortOnce,OPT_BOOLEAN,true,ResScriptPortOnce,aoffset(script_port_once),
     NULL, "Accept one script connection, then exit" },
-{ OptSet,      OPT_SKIP2,   false, NULL,         NULL,
-    "<toggle>", "Turn on <toggle>" },
+{ OptSet,      OPT_SET,     false, NULL,         NULL,
+    "<resource>[=<value>]", "Set <resource> to " ResTrue " or <value>" },
 { OptSocket,   OPT_BOOLEAN, true,  ResSocket,    aoffset(socket),
     NULL, "Create socket for script control" },
 { OptTermName, OPT_STRING,  false, ResTermName,  aoffset(termname),
@@ -574,21 +598,20 @@ static opt_t base_opts[] = {
     "<n>[KM]", "Limit trace file to <n> bytes" },
 { OptUser,     OPT_STRING,  false, ResUser,      aoffset(user),
     "<name>", "User name for RFC 4777" },
-{ OptV,        OPT_V,	false, NULL,	     NULL,
+{ OptV,        OPT_V,       false, NULL,	     NULL,
     NULL, "Display build options and character sets" },
-{ OptVersion,  OPT_V,	false, NULL,	     NULL,
+{ OptVersion,  OPT_V,       false, NULL,	     NULL,
     NULL, "Display build options and character sets" },
-{ OptHelp1,     OPT_HELP, false, NULL,            NULL,
+{ OptHelp1,    OPT_HELP,    false, NULL,            NULL,
     NULL, "Display command-line help" },
-{ OptHelp2,     OPT_HELP, false, NULL,            NULL,
+{ OptHelp2,    OPT_HELP,    false, NULL,            NULL,
     NULL, "Display command-line help" },
 #if defined(_WIN32) /*[*/
-{ OptHelp3,     OPT_HELP, false, NULL,            NULL,
+{ OptHelp3,    OPT_HELP,    false, NULL,            NULL,
     NULL, "Display command-line help" },
 #endif /*]*/
-{ "-xrm",      OPT_XRM,     false, NULL,         NULL,
-    "'*.<resource>: <value>'", "Set <resource> to <value>"
-},
+{ OptXrm,      OPT_XRM,     false, NULL,         NULL,
+    "'*.<resource>: <value>'", "Set <resource> to <value>" },
 { LAST_ARG,    OPT_DONE,    false, NULL,         NULL,
     NULL, "Terminate argument list" }
 };
@@ -658,26 +681,40 @@ parse_options(int *argcp, const char **argv, bool warn)
 	case OPT_BOOLEAN:
 	    *(bool *)opts[j].aoff = opts[j].flag;
 	    if (opts[j].res_name != NULL) {
-		add_resource(NewString(opts[j].name),
-			opts[j].flag? "true": "false");
+		add_resource(opts[j].name, opts[j].flag? "true": "false");
 	    }
 	    break;
 	case OPT_STRING:
 	    if (i == *argcp - 1) {	/* missing arg */
-		popup_an_error("Missing value for '%s'", argv[i]);
+		usage(xs_buffer("Missing value for '%s'", argv[i]));
 		continue;
 	    }
+	    Free(*(char **)opts[j].aoff);
 	    *(const char **)opts[j].aoff = NewString(argv[++i]);
 	    if (opts[j].res_name != NULL) {
-		add_resource(NewString(opts[j].res_name), NewString(argv[i]));
+		add_resource(opts[j].res_name, argv[i]);
 	    }
 	    break;
 	case OPT_XRM:
 	    if (i == *argcp - 1) {	/* missing arg */
-		popup_an_error("Missing value for '%s'", argv[i]);
+		usage(xs_buffer("Missing value for '%s'", argv[i]));
 		continue;
 	    }
-	    xparse_xrm(argv[++i], "-xrm", warn);
+	    xparse_xrm(argv[++i], OptXrm, warn);
+	    break;
+	case OPT_SET:
+	    if (i == *argcp - 1) {	/* missing arg */
+		usage(xs_buffer("Missing value for '%s'", argv[i]));
+		continue;
+	    }
+	    parse_set(argv[++i], OptSet, warn);
+	    break;
+	case OPT_CLEAR:
+	    if (i == *argcp - 1) {	/* missing arg */
+		usage(xs_buffer("Missing value for '%s'", argv[i]));
+		continue;
+	    }
+	    parse_clear(argv[++i], OptClear, warn);
 	    break;
 	case OPT_SKIP2:
 	    argv_out[argc_out++] = argv[i++];
@@ -689,12 +726,12 @@ parse_options(int *argcp, const char **argv, bool warn)
 	    break;
 	case OPT_INT:
 	    if (i == *argcp - 1) {	/* missing arg */
-		popup_an_error("Missing value for '%s'", argv[i]);
+		usage(xs_buffer("Missing value for '%s'", argv[i]));
 		continue;
 	    }
 	    *(int *)opts[j].aoff = atoi(argv[++i]);
 	    if (opts[j].res_name != NULL) {
-		add_resource(NewString(opts[j].name), NewString(argv[i]));
+		add_resource(opts[j].name, argv[i]);
 	    }
 	    break;
 	case OPT_V:
@@ -794,10 +831,14 @@ cmdline_help(bool as_action)
     unsigned i;
 
     if (!as_action) {
-	 fprintf(stderr, "Usage: "
-		 "%s [options] [[prefix:][LUname@]hostname[:port]]\n",
-		 programname);
-	 fprintf(stderr, "Options:\n");
+	if (supports_cmdline_host) {
+	    fprintf(stderr, "Usage: %s [options] "
+		    "[[prefix:][LUname@]hostname[:port]]\n", programname);
+	}
+	fprintf(stderr, "%s %s [options] [<session-file>].%s\n",
+		supports_cmdline_host? "      ": "Usage:",
+		programname, app);
+	fprintf(stderr, "Options:\n");
     }
     sort_help();
     for (i = 0; i < sorted_help_count; i++) {
@@ -833,132 +874,6 @@ cmdline_help(bool as_action)
     }
 }
 
-/* Comparison function for toggle name qsort. */
-static int
-name_cmp(const void *p1, const void *p2)
-{
-    const char *s1 = *(char *const *)p1;
-    const char *s2 = *(char *const *)p2;
-
-    return strcmp(s1, s2);
-}
-
-/*
- * Pick out -set and -clear toggle options.
- */
-static void
-parse_set_clear(int *argcp, const char **argv, bool warn)
-{
-    int i, j;
-    int argc_out = 0;
-    const char **argv_out =
-	(const char **) Malloc((*argcp + 1) * sizeof(char *));
-
-    argv_out[argc_out++] = argv[0];
-
-    for (i = 1; i < *argcp; i++) {
-	bool is_set = false;
-	const char *eq = NULL;
-	size_t nlen;
-
-	if (!strcmp(argv[i], OptSet)) {
-	    is_set = true;
-	} else if (strcmp(argv[i], OptClear)) {
-	    argv_out[argc_out++] = argv[i];
-	    continue;
-	}
-
-	if (i == *argcp - 1) {	/* missing arg */
-	    continue;
-	}
-
-	/* Delete the argument. */
-	i++;
-
-	if (argv[i][0] != '=' &&
-		(eq = strchr(argv[i], '=')) != NULL) {
-	    /* -set foo=bar */
-	    if (!is_set) {
-		fprintf(stderr, "Error: " OptClear
-			" parameter cannot include a value\n");
-		exit(1);
-	    }
-	    nlen = eq - argv[i];
-	} else {
-	    nlen = strlen(argv[i]);
-	}
-
-	for (j = 0; toggle_names[j].name != NULL; j++) {
-	    if (!toggle_supported(toggle_names[j].index)) {
-		continue;
-	    }
-	    if (!strcasecmp(argv[i], toggle_names[j].name) &&
-		    toggle_names[j].name[nlen] == '\0') {
-		bool value;
-
-		if (eq == NULL) {
-		    value = is_set;
-		} else {
-		    const char *err = boolstr(eq + 1, &value);
-
-		    if (err != NULL) {
-			fprintf(stderr, "Error: " OptSet " %s: %s\n", argv[i],
-				err);
-			exit(1);
-		    }
-		}
-		appres.toggle[toggle_names[j].index] = value;
-		break;
-	    }
-	}
-	if (toggle_names[j].name == NULL) {
-	    bool bool_only = !is_set || !eq;
-	    int xt = init_extended_toggle(argv[i], nlen, bool_only,
-		    eq? eq + 1: (is_set? ResTrue: ResFalse), NULL);
-
-	    if (xt < 0) {
-		fprintf(stderr, "Error: " OptSet " %s: invalid value\n",
-			argv[i]);
-		exit(1);
-	    }
-
-	    if (xt == 0) {
-		ccp_t *tn;
-		int ntn = 0;
-		int nx;
-		char **nxnames;
-
-		nxnames = extended_toggle_names(&nx, bool_only);
-		tn = (ccp_t *)Calloc(N_TOGGLES + nx, sizeof(ccp_t));
-		for (j = 0; toggle_names[j].name != NULL; j++) {
-		    if (!toggle_supported(toggle_names[j].index)) {
-			continue;
-		    }
-		    if (!toggle_names[j].is_alias) {
-			tn[ntn++] = toggle_names[j].name;
-		    }
-		}
-		memcpy((void *)(tn + ntn), nxnames, nx * sizeof(ccp_t));
-		ntn += nx;
-		qsort((void *)tn, ntn, sizeof(const char *), name_cmp);
-		fprintf(stderr, "Unknown %stoggle '%.*s'. Toggle names are:\n",
-			bool_only? "Boolean ": "", (int)nlen, argv[i]);
-		for (j = 0; j < ntn; j++) {
-		    fprintf(stderr, " %s", tn[j]);
-		}
-		fprintf(stderr, "\n");
-		Free((void *)tn);
-		exit(1);
-	    }
-	}
-
-    }
-    *argcp = argc_out;
-    argv_out[argc_out] = NULL;
-    memcpy((char *)argv, (char *)argv_out, (argc_out + 1) * sizeof(char *));
-    Free((char *)argv_out);
-}
-
 /*
  * Parse the model number.
  * Returns -1 (error), 0 (default), or the specified number.
@@ -1043,6 +958,7 @@ static res_t base_resources[] = {
     { ResCodePage,	aoffset(codepage),	XRM_STRING },
     { ResConfDir,	aoffset(conf_dir),	XRM_STRING },
     { ResConnectTimeout,aoffset(connect_timeout),XRM_INT },
+    { ResContentionResolution, aoffset(contention_resolution), XRM_BOOLEAN },
     { ResCrosshairColor,aoffset(interactive.crosshair_color),	XRM_STRING },
     { ResConsole,aoffset(interactive.console),	XRM_STRING },
     { ResDbcsCgcsgid, aoffset(dbcs_cgcsgid),	XRM_STRING },
@@ -1064,6 +980,7 @@ static res_t base_resources[] = {
     { ResFtLocalFile,	aoffset(ft.local_file),	XRM_STRING },
     { ResFtLrecl,	aoffset(ft.lrecl),	XRM_INT },
     { ResFtMode,	aoffset(ft.mode),	XRM_STRING },
+    { ResFtOtherOptions,aoffset(ft.other_options),XRM_STRING },
     { ResFtPrimarySpace,aoffset(ft.primary_space),XRM_INT },
     { ResFtRecfm,	aoffset(ft.recfm),	XRM_STRING },
     { ResFtRemap,	aoffset(ft.remap),	XRM_STRING },
@@ -1098,6 +1015,8 @@ static res_t base_resources[] = {
     { ResProxy,		aoffset(proxy),		XRM_STRING },
     { ResQrBgColor,	aoffset(qr_bg_color),	XRM_BOOLEAN },
     { ResQuit,		aoffset(linemode.quit),	XRM_STRING },
+    { ResReconnect,	aoffset(reconnect),	XRM_BOOLEAN },
+    { ResRetry,		aoffset(retry),		XRM_BOOLEAN },
     { ResRprnt,		aoffset(linemode.rprnt),	XRM_STRING },
     { ResScreenTraceFile,aoffset(screentrace.file),XRM_STRING },
     { ResScreenTraceTarget,aoffset(screentrace.target),XRM_STRING },
@@ -1114,7 +1033,8 @@ static res_t base_resources[] = {
     { ResTraceMonitor,aoffset(trace_monitor),	XRM_BOOLEAN },
     { ResUnlockDelay,aoffset(unlock_delay),	XRM_BOOLEAN },
     { ResUnlockDelayMs,aoffset(unlock_delay_ms),	XRM_INT },
-    { ResWerase,	aoffset(linemode.werase),XRM_STRING }
+    { ResWerase,	aoffset(linemode.werase),XRM_STRING },
+    { ResWrongTerminalName,aoffset(wrong_terminal_name),XRM_BOOLEAN }
 };
 
 typedef struct reslist {
@@ -1284,7 +1204,7 @@ xparse_xrm(const char *arg, const char *where, bool warn)
     bool quoted;
     char c;
     reslist_t *r;
-    char *hide;
+    char *hide = NULL;
     bool arbitrary = false;
     const char *errmsg;
 
@@ -1312,7 +1232,7 @@ xparse_xrm(const char *arg, const char *where, bool warn)
 	    if (!toggle_supported(toggle_names[i].index)) {
 		continue;
 	    }
-	    if (!strncapcmp(toggle_names[i].name, name, rnlen)) {
+	    if (!strncasecmp(toggle_names[i].name, name, rnlen)) {
 		address = &appres.toggle[toggle_names[i].index];
 		type = XRM_BOOLEAN;
 		break;
@@ -1343,6 +1263,7 @@ xparse_xrm(const char *arg, const char *where, bool warn)
 	break;
     case XRM_STRING:
 	t = Malloc(strlen(s) + 1);
+	Free(*(char **)address);
 	*(char **)address = t;
 	quoted = false;
 #if defined(_WIN32) /*[*/
@@ -1422,15 +1343,45 @@ xparse_xrm(const char *arg, const char *where, bool warn)
 	strncpy(rsname, name, rnlen);
 	rsname[rnlen] = '\0';
 	add_resource(rsname, hide);
+	Free(rsname);
+	Free(hide);
     }
 }
 
+
 void
 parse_xrm(const char *arg, const char *where)
 {
     xparse_xrm(arg, where, true);
 }
 
+/* Parse a '-set' option. */
+static void
+parse_set(const char *arg, const char *where, bool warn)
+{
+    const char *eq = strchr(arg, '=');
+    char *xrm_arg;
+
+    if (eq != NULL) {
+	xrm_arg = xs_buffer("%s.%.*s: %s", app, (int)(eq - arg), arg,
+		eq + 1);
+    } else {
+	xrm_arg = xs_buffer("%s.%s: %s", app, arg, ResTrue);
+    }
+    xparse_xrm(xrm_arg, where, warn);
+    Free(xrm_arg);
+}
+
+/* Parse a '-clear' option. */
+static void
+parse_clear(const char *arg, const char *where, bool warn)
+{
+    char *xrm_arg = xs_buffer("%s.%s: %s", app, arg, ResFalse);
+
+    xparse_xrm(xrm_arg, where, warn);
+    Free(xrm_arg);
+}
+
 /*
  * Clean up a string for display (undo what parse_xrm does).
  */
