@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2020 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -37,10 +37,13 @@
 #include "resources.h"
 
 #include <assert.h>
+#include <limits.h>
 #include "actions.h"
 #include "boolstr.h"
+#include "glue_gui.h"
 #include "host.h"
 #include "host_gui.h"
+#include "lazya.h"
 #include "login_macro.h"
 #include "names.h"
 #include "popups.h"
@@ -76,6 +79,7 @@ unsigned short  current_port;
 char	       *reconnect_host = NULL;
 char	       *qualified_host = NULL;
 enum iaction	connect_ia = IA_NONE;
+bool		host_retry_mode = false;
 
 struct host *hosts = NULL;
 static struct host *last_host = NULL;
@@ -91,7 +95,6 @@ static void try_reconnect(ioid_t id);
 static action_t Connect_action;
 static action_t Disconnect_action;
 static action_t Reconnect_action;
-static action_t SubjectNames_action;
 
 static char *
 stoken(char **s)
@@ -162,7 +165,7 @@ read_hosts_file(void)
 		continue;
 	    }
 	    h = (struct host *)Malloc(sizeof(*h));
-	    if (!split_hier(NewString(name), &h->name, &h->parents)) {
+	    if (!split_hier(name, &h->name, &h->parents)) {
 		Free(h);
 		continue;
 	    }
@@ -246,30 +249,115 @@ host_cancel_reconnect(void)
     }
 }
 
+/**
+ * Common logic for when the reconnect or retry options change.
+ */
+static void
+reconnect_retry_touched(void)
+{
+    /*
+     * Turning off reconnect/retry is the way to stop a reconnect in progress.
+     */
+    if (!appres.reconnect && !appres.retry) {
+	host_cancel_reconnect();
+    }
+
+    /* They have changed one of the flags. Reset host_retry_mode. */
+    host_retry_mode = appres.reconnect || appres.retry;
+}
+
+/**
+ * Canonicalize the configuration directory.
+ *
+ * @param[in] dir	Configuration directory.
+ *
+ * @return Canonicalized form
+ */
+static const char *
+canon_conf_dir(const char *dir)
+{
+#if !defined(_WIN32) /*[*/
+    static char resolved_path[PATH_MAX];
+    char *result = realpath(dir, resolved_path);
+
+    return result? result: dir;
+#else /*][*/
+    static char resolved_path[MAX_PATH];
+    DWORD len = GetFullPathName(dir, MAX_PATH, resolved_path, NULL);
+
+    return len? resolved_path: dir;
+#endif /*]*/
+}
+
+/**
+ * Toggle the configuration directory.
+ *
+ * @param[in] name	Toggle name.
+ * @param[in] value	New value.
+ * @param[in] flags	Set() flags.
+ * @param[in] ia	Cause.
+ *
+ * @return toggle_upcall_ret_t
+ */
+static toggle_upcall_ret_t
+set_conf_dir(const char *name _is_unused, const char *value, unsigned flags, ia_t ia)
+{
+    popup_an_error("Cannot set " ResConfDir);
+    return TU_FAILURE;
+}
+
 /**
  * Toggle the reconnect flag.
  *
  * @param[in] name	Toggle name.
  * @param[in] value	New value.
+ * @param[in] flags	Set() flags.
+ * @param[in] ia	Cause.
  *
- * @return true if sucessful
+ * @return toggle_upcall_ret_t
  */
-static bool
-set_reconnect(const char *name _is_unused, const char *value)
+static toggle_upcall_ret_t
+set_reconnect(const char *name _is_unused, const char *value, unsigned flags, ia_t ia)
 {
-    bool previous = appres.interactive.reconnect;
+    bool previous = appres.reconnect;
     const char *errmsg;
 
-    if ((errmsg = boolstr(value, &appres.interactive.reconnect)) != NULL) {
+    if ((errmsg = boolstr(value, &appres.reconnect)) != NULL) {
 	popup_an_error("%s", errmsg);
-	return false;
+	return TU_FAILURE;
     }
 
-    if (appres.interactive.reconnect != previous &&
-	    !appres.interactive.reconnect) {
-	host_cancel_reconnect();
+    if (appres.reconnect != previous) {
+	reconnect_retry_touched();
     }
-    return true;
+    return TU_SUCCESS;
+}
+
+/**
+ * Toggle the retry flag.
+ *
+ * @param[in] name	Toggle name.
+ * @param[in] value	New value.
+ * @param[in] flags	Set() flags.
+ * @param[in] ia	Cause.
+ *
+ * @return toggle_upcall_ret_t
+ */
+static toggle_upcall_ret_t
+set_retry(const char *name _is_unused, const char *value, unsigned flags, ia_t ia)
+{
+    bool previous = appres.retry;
+    const char *errmsg;
+
+    if ((errmsg = boolstr(value, &appres.retry)) != NULL) {
+	popup_an_error("%s", errmsg);
+	return TU_FAILURE;
+    }
+
+    if (appres.reconnect != previous) {
+	reconnect_retry_touched();
+    }
+    return TU_SUCCESS;
 }
 
 /**
@@ -284,15 +372,18 @@ host_register(void)
 	{ AnDisconnect,	Disconnect_action,	ACTION_KE },
 	{ AnOpen,	Connect_action,		ACTION_KE },
 	{ AnReconnect,	Reconnect_action,	ACTION_KE },
-	{ AnSubjectNames,SubjectNames_action,	ACTION_KE }
     };
 
     /* Register for events. */
     register_schange(ST_EXITING, host_exiting);
 
     /* Register our toggles. */
+    register_extended_toggle(ResConfDir, set_conf_dir, NULL, canon_conf_dir,
+	    (void **)&appres.conf_dir, XRM_STRING);
     register_extended_toggle(ResReconnect, set_reconnect, NULL, NULL,
-	    (void **)&appres.interactive.reconnect, XRM_BOOLEAN);
+	    (void **)&appres.reconnect, XRM_BOOLEAN);
+    register_extended_toggle(ResRetry, set_retry, NULL, NULL,
+	    (void **)&appres.retry, XRM_BOOLEAN);
 
     /* Register our actions. */
     register_actions(host_actions, array_count(host_actions));
@@ -312,6 +403,8 @@ hostfile_init(void)
 
     read_hosts_file();
 
+    host_retry_mode = appres.reconnect || appres.retry;
+
     hostfile_initted = true;
 }
 
@@ -329,7 +422,7 @@ hostfile_lookup(const char *name, char **hostname, char **loginstring)
 	if (h->entry_type == RECENT) {
 	    continue;
 	}
-	if (!strcmp(name, h->name)) {
+	if (!strcasecmp(name, h->name)) {
 	    *hostname = h->hostname;
 	    if (h->loginstring != NULL) {
 		*loginstring = h->loginstring;
@@ -423,6 +516,11 @@ host_connect(const char *n, enum iaction ia)
     bool has_colons = false;
     net_connect_t nc;
 
+    if (!glue_gui_open_safe()) {
+	popup_an_error("User interface in the wrong state");
+	return false;
+    }
+
     if (cstate == RECONNECTING) {
 	popup_an_error("Reconnect in progress");
 	return false;
@@ -451,7 +549,9 @@ host_connect(const char *n, enum iaction ia)
     }
 
     /* Remember this hostname, as the last hostname we connected to. */
-    Replace(reconnect_host, NewString(nb));
+    if (reconnect_host == NULL || strcmp(reconnect_host, nb)) {
+	Replace(reconnect_host, NewString(nb));
+    }
 
     /* Remember this hostname in the recent connection list and file. */
     save_recent(nb);
@@ -471,16 +571,16 @@ host_connect(const char *n, enum iaction ia)
 	    goto failure;
 	}
 
-	/* Look up the name in the hosts file. */
+	/* If the hostname is naked, look it up in the hosts file. */
 	if (!needed && hostfile_lookup(s, &target_name, &ps)) {
 	    /*
-	     * Rescan for qualifiers.
-	     * Qualifiers, LU names, ports and accept names  are all
-	     * overridden by the hosts file.
+	     * Split out all of the other decorations from the entry in the
+	     * hosts file.
 	     */
 	    Free(s);
-	    if (!(s = split_host(target_name, &host_flags, luname, &port,
-			    &accept, &needed))) {
+	    s = split_host(target_name, &host_flags, luname, &port, &accept,
+		    &needed);
+	    if (s == NULL) {
 		goto failure;
 	    }
 	}
@@ -498,7 +598,7 @@ host_connect(const char *n, enum iaction ia)
      *   and port number
      *  full_current_host is the entire string, for use in reconnecting
      */
-    if (n != full_current_host) {
+    if (full_current_host == NULL || strcmp(full_current_host, n)) {
 	Replace(full_current_host, NewString(n));
     }
     Replace(current_host, NULL);
@@ -525,17 +625,26 @@ host_connect(const char *n, enum iaction ia)
 		(accept != NULL)? accept: ""));
 
     /* Attempt contact. */
+    host_retry_mode = appres.reconnect || appres.retry;
     ever_3270 = false;
     nc = net_connect(chost, port, accept, localprocess_cmd != NULL, &net_sock);
+    if (port != appres.port) {
+	Replace(port, NULL);
+    }
+    Replace(accept, NULL);
     if (nc == NC_FAILED) {
 	if (!host_gui_connect()) {
-	    if (appres.interactive.reconnect) {
+	    if (host_retry_mode) {
 		reconnect_id = AddTimeOut(RECONNECT_ERR_MS, try_reconnect);
 		change_cstate(RECONNECTING, "host_connect");
 	    }
 	}
 	/* Redundantly signal a disconnect. */
 	change_cstate(NOT_CONNECTED, "host_connect");
+	if (reconnect_id != NULL_IOID) {
+	    /* Change it back so we are in the right state when we retry. */
+	    change_cstate(RECONNECTING, "host_connect");
+	}
 	goto failure;
     }
 
@@ -575,8 +684,8 @@ host_connect(const char *n, enum iaction ia)
     /* Set state and tell the world. */
     if (nc == NC_CONNECT_PENDING) {
 	change_cstate(TCP_PENDING, "host_connect");
-    } else {
-	/* cstate == NC_CONNECTED */
+    } else if (cstate != TLS_PENDING) {
+	/* nc == NC_CONNECTED and TLS not pending */
 	if (appres.nvt_mode || HOST_FLAG(ANSI_HOST)) {
 	    change_cstate(CONNECTED_NVT, "host_connect");
 	} else {
@@ -686,7 +795,7 @@ host_disconnect(bool failed)
     net_disconnect(true);
     net_sock = INVALID_IOSRC;
     if (!host_gui_disconnect()) {
-	if (appres.interactive.reconnect && reconnect_id == NULL_IOID) {
+	if (host_retry_mode && reconnect_id == NULL_IOID) {
 	    /* Schedule an automatic reconnection. */
 	    reconnect_id = AddTimeOut(failed? RECONNECT_ERR_MS:
 					      RECONNECT_MS,
@@ -727,6 +836,7 @@ void
 host_connected(void)
 {
     change_cstate(TELNET_PENDING, "host_connected");
+    host_retry_mode = appres.reconnect;
     host_gui_connected();
 }
 
@@ -1041,36 +1151,6 @@ Disconnect_action(ia_t ia, unsigned argc, const char **argv)
     return true;
 }
 
-/*
- * Display the subject names for a host.
- */
-static bool
-SubjectNames_action(ia_t ia, unsigned argc, const char **argv)
-{
-    char *s;
-
-    action_debug(AnSubjectNames, ia, argc, argv);
-    if (check_argc(AnSubjectNames, argc, 1, 1) < 0) {
-	return false;
-    }
-
-    /*
-     * Run the following commands in a macro:
-     *  Connect(Y:<hostname>)
-     *  Wait(InputField)	# needed for the c3270> prompt, which
-     *  			# does non-blocking connects
-     *  Show(TlsSubjectNames)
-     *  Disconnect()
-     */
-    s = xs_buffer(AnConnect "(Y:%s) "
-	    AnWait "(" KwInputField ") "
-	    AnShow "(" KwTlsSubjectNames ") "
-	    AnDisconnect "()", argv[0]);
-    push_stack_macro(s);
-    Free(s);
-    return true;
-}
-
 bool
 host_reconnecting(void)
 {
