@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2016, 2018-2021 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -47,8 +47,11 @@
 #include <fcntl.h>
 
 #include "actions.h"
+#include "json.h"
+#include "json_run.h"
 #include "kybd.h"
 #include "lazya.h"
+#include "names.h"
 #include "peerscript.h"
 #include "popups.h"
 #include "s3270_proto.h"
@@ -126,11 +129,14 @@ typedef struct {
     ioid_t id;		/* I/O identifier */
     char *buf;		/* pending command */
     size_t buf_len;	/* length of pending command */
+    size_t pj_offset;	/* partial JSON offset */
     bool enabled;	/* is this peer enabled? */
     char *name;		/* task name */
     unsigned capabilities; /* self-reported capabilities */
     void *irhandle;	/* input request handle */
     task_cb_ir_state_t ir_state; /* named input request state */
+    bool json_mode;	/* true if in JSON mode */
+    json_t *json_result; /* pending JSON result */
 } peer_t;
 static llist_t peer_scripts = LLIST_INIT(peer_scripts);
 
@@ -179,41 +185,109 @@ close_peer(peer_t *p)
     }
     task_cb_abort_ir_state(&p->ir_state);
 
+    json_free(p->json_result);
+
     Free(p);
 }
 
+/**
+ * Pushes a command, with possible JSON parsing.
+ *
+ * @param[in] p		Peer state
+ * @param[in] buf	Command buffer
+ * @param[in] len	Buffer length
+ *
+ * @return true if command complete, false if partial JSON found
+ */
+static bool
+do_push(peer_t *p, const char *buf, size_t len)
+{
+    const char *s = buf;
+    char *name;
+    tcb_t *tcb = (p->capabilities & CBF_INTERACTIVE)?
+	&interactive_cb : &peer_cb;
+
+    while (len && isspace((int)*s)) {
+	s++;
+	len--;
+    }
+
+    /* Try JSON parsing. */
+    if (!(p->capabilities & CBF_INTERACTIVE) &&
+	    (*s == '{' || *s == '[' || *s == '"')) {
+	cmd_t **cmds;
+	char *single;
+	char *errmsg;
+	hjparse_ret_t ret;
+
+	ret = hjson_parse(s, len, &cmds, &single, &errmsg);
+	if (ret == HJ_OK) {
+	    /* Good JSON. */
+	    p->json_mode = true;
+	    if (cmds != NULL) {
+		name = push_cb_split(cmds, tcb, (task_cbh)p);
+	    } else {
+		name = push_cb(single, strlen(single), tcb, (task_cbh)p);
+		Free(single);
+	    }
+	} else if (ret == HJ_INCOMPLETE) {
+	    Free(errmsg);
+	    return false;
+	} else {
+	    /* Bad JSON. */
+	    char *fail = xs_buffer(AnFail "(\"%s\")", errmsg);
+
+	    /* Answer in JSON only if successfully parsed. */
+	    p->json_mode = (ret != HJ_BAD_SYNTAX);
+	    Free(errmsg);
+	    name = push_cb(fail, strlen(fail), tcb, (task_cbh)p);
+	    Free(fail);
+	}
+    } else {
+	p->json_mode = false;
+	name = push_cb(s, len, tcb, (task_cbh)p);
+    }
+    Replace(p->name, NewString(name));
+    return true;
+}
+
 /**
  * Run the next command in the peer buffer.
  *
  * @param[in,out] p	Peer
  *
- * @return true if command was run. Command is deleted from the buffer.
+ * @return true if command was run and command deleted from the buffer.
  */
 static bool
 run_next(peer_t *p)
 {
     size_t cmdlen;
-    char *name;
 
-    /* Find a newline in the buffer. */
-    for (cmdlen = 0; cmdlen < p->buf_len; cmdlen++) {
-	if (p->buf[cmdlen] == '\n') {
+    while (true) {
+	/* Find the first newline in the buffer. */
+	for (cmdlen = p->pj_offset; cmdlen < p->buf_len; cmdlen++) {
+	    if (p->buf[cmdlen] == '\n') {
+		break;
+	    }
+	}
+	if (cmdlen >= p->buf_len) {
+	    /* No newline. */
+	    return false;
+	}
+
+	/*
+	 * Run the first command.
+	 * cmdlen is the number of characters in the command, not including the
+	 * newline.
+	 */
+	if (do_push(p, p->buf, cmdlen)) {
 	    break;
 	}
-    }
-    if (cmdlen >= p->buf_len) {
-	return false;
-    }
 
-    /*
-     * Run the first command.
-     * cmdlen is the number of characters in the command, not including the
-     * newline.
-     */
-    name = push_cb(p->buf, cmdlen,
-	    (p->capabilities & CBF_INTERACTIVE)? &interactive_cb : &peer_cb,
-	    (task_cbh)p);
-    Replace(p->name, NewString(name));
+	/* Partial JSON. */
+	p->pj_offset = cmdlen + 1;
+    }
+    p->pj_offset = 0;
 
     /* If there is more, shift it over. */
     cmdlen++; /* count the newline */
@@ -228,7 +302,7 @@ run_next(peer_t *p)
 }
 
 /**
- * Read the next command from a peer socket (Unix version).
+ * Read the next command from a peer socket.
  * @param[in] fd	File descriptor
  * @param[in] id	I/O identifier
  */
@@ -258,11 +332,10 @@ peer_input(iosrc_t fd _is_unused, ioid_t id)
 #if defined(_WIN32) /*[*/
 	if (GetLastError() != WSAECONNRESET) {
 	    /* Windows does this habitually. */
-	    popup_an_error("s3sock recv: %s",
-		    win32_strerror(GetLastError()));
+	    vtrace("s3sock recv: %s\n", win32_strerror(GetLastError()));
 	}
 #else /*][*/
-	popup_an_errno(errno, "s3sock read");
+	vtrace("s3sock recv: %s\n", strerror(errno));
 #endif /*]*/
 	close_peer(p);
 	return;
@@ -301,6 +374,31 @@ peer_input(iosrc_t fd _is_unused, ioid_t id)
     }
 }
 
+/**
+ * Send data on a socket and check the result.
+ *
+ * @param[in] s                Socket
+ * @param[in] data     Data to send
+ * @param[in] len      Length
+ * @paran[in] sender   Sending function
+ */
+static void
+check_send(socket_t s, const char *data, size_t len, const char *sender)
+{
+    ssize_t ns = send(s, data, (int)len, 0);
+    if (ns != (ssize_t)len) {
+	if (ns < 0) {
+#if !defined(_WIN32) /*[*/
+	    vtrace("%s send: %s\n", sender, strerror(errno));
+#else /*][*/
+	    vtrace("%s send: %s\n", sender, win32_strerror(GetLastError()));
+#endif/*]*/
+	} else {
+	    vtrace("%s: short send\n", sender);
+	}
+    }
+}
+
 /**
  * Callback for data returned to peer socket command.
  *
@@ -313,18 +411,43 @@ static void
 peer_data(task_cbh handle, const char *buf, size_t len, bool success)
 {
     peer_t *p = (peer_t *)handle;
-    char *s = lazyaf(DATA_PREFIX "%.*s\n", (int)len, buf);
-    ssize_t ns;
+    char *b;
+    char *newline;
     static bool recursing = false;
 
     if (recursing) {
 	return;
     }
     recursing = true;
-    ns = send(p->socket, s, strlen(s), 0);
-    if (ns < 0) {
-	popup_a_sockerr("s3sock send");
+
+    while (len > 0 && buf[len - 1] == '\n') {
+	len--;
     }
+    b = xs_buffer("%.*s", (int)len, buf);
+    while ((newline = strchr(b, '\n')) != NULL) {
+	*newline = ' ';
+    }
+
+    if (p->json_mode) {
+	json_t *result_array;
+
+	if (p->json_result == NULL) {
+            p->json_result = json_object();
+            result_array = json_array();
+            json_object_set(p->json_result, "result", NT, result_array);
+        } else {
+            assert(json_object_member(p->json_result, "result", NT,
+                        &result_array));
+        }
+	json_array_append(result_array, json_string(b, strlen(b)));
+    } else {
+	char *data = xs_buffer("%s%s\n", DATA_PREFIX, b);
+
+	check_send(p->socket, data, strlen(data), "peer_data");
+	Free(data);
+    }
+
+    Free(b);
     recursing = false;
 }
 
@@ -340,19 +463,17 @@ static void
 peer_reqinput(task_cbh handle, const char *buf, size_t len, bool echo)
 {
     peer_t *p = (peer_t *)handle;
-    char *s = lazyaf("%s%.*s\n", echo? INPUT_PREFIX: PWINPUT_PREFIX, (int)len,
-	    buf);
-    ssize_t ns;
+    char *s;
     static bool recursing = false;
 
     if (recursing) {
 	return;
     }
     recursing = true;
-    ns = send(p->socket, s, strlen(s), 0);
-    if (ns < 0) {
-	popup_a_sockerr("s3sock send");
-    }
+
+    s = xs_buffer("%s%.*s\n", echo? INPUT_PREFIX: PWINPUT_PREFIX, (int)len, buf);
+    check_send(p->socket, s, strlen(s), "peer_reqinput");
+    Free(s);
     recursing = false;
 }
 
@@ -370,13 +491,32 @@ peer_done(task_cbh handle, bool success, bool abort)
 {
     peer_t *p = (peer_t *)handle;
     char *prompt = task_cb_prompt(handle);
-    char *s = lazyaf("%s\n%s\n", prompt, success? PROMPT_OK: PROMPT_ERROR);
+    char *s;
     bool new_child = false;
 
-    /* Print the prompt. */
-    vtrace("Output for %s: '%s/%s'\n", p->name, prompt,
-	    success? PROMPT_OK: PROMPT_ERROR);
-    send(p->socket, s, strlen(s), 0);
+    if (p->json_mode) {
+	char *w;
+	/* Print the result. */
+
+        if (p->json_result == NULL) {
+            p->json_result = json_object();
+            json_object_set(p->json_result, "result", NT, json_array());
+        }
+        json_object_set(p->json_result, "success", NT, json_boolean(success));
+        json_object_set(p->json_result, "status", NT, json_string(prompt, NT));
+        s = xs_buffer("%s\n", w = json_write_o(p->json_result, JW_ONE_LINE));
+        Free(w);
+        json_free(p->json_result);
+	vtrace("JSON output for %s: '%s/%s'\n", p->name, prompt,
+		success? PROMPT_OK: PROMPT_ERROR);
+    } else {
+	/* Print the prompt. */
+	vtrace("Output for %s: '%s/%s'\n", p->name, prompt,
+		success? PROMPT_OK: PROMPT_ERROR);
+	s = xs_buffer("%s\n%s\n", prompt, success? PROMPT_OK: PROMPT_ERROR);
+    }
+    check_send(p->socket, s, strlen(s), "peer_done");
+    Free(s);
 
     if (abort || !p->enabled) {
 	close_peer(p);
@@ -513,9 +653,7 @@ peer_connection(iosrc_t fd _is_unused, ioid_t id)
     union {
 	struct sockaddr sa;
 	struct sockaddr_in sin;
-#if defined(X3270_IPV6) /*[*/
 	struct sockaddr_in6 sin6;
-#endif /*]*/
     } sa;
     socklen_t len = sizeof(sa);
     char hostbuf[128];
@@ -536,14 +674,11 @@ peer_connection(iosrc_t fd _is_unused, ioid_t id)
 	    vtrace("New script socket connection from %s:%u\n",
 		    inet_ntop(AF_INET, &sa.sin.sin_addr, hostbuf,
 			sizeof(hostbuf)), ntohs(sa.sin.sin_port));
-	}
-#if defined(X3270_IPV6) /*[*/
-	else if (sa.sa.sa_family == AF_INET6) {
+	} else if (sa.sa.sa_family == AF_INET6) {
 	    vtrace("New script socket connection from %s:%u\n",
 		    inet_ntop(AF_INET6, &sa.sin6.sin6_addr, hostbuf,
 			sizeof(hostbuf)), ntohs(sa.sin6.sin6_port));
 	}
-#endif /*]*/
 #if !defined(_WIN32) /*[*/
 	else if (sa.sa.sa_family == AF_UNIX) {
 	    vtrace("New Unix-domain script socket connection");
@@ -555,7 +690,11 @@ peer_connection(iosrc_t fd _is_unused, ioid_t id)
     }
 
     if (accept_fd == INVALID_SOCKET) {
-	popup_an_errno(errno, "script socket accept");
+#if !defined(_WIN32) /*[*/
+	vtrace("s3sock accept: %s\n", strerror(errno));
+#else /*][*/
+	vtrace("s3sock accept: %s\n", win32_strerror(GetLastError()));
+#endif /*]*/
 	return;
     }
 
@@ -738,9 +877,7 @@ peer_init(struct sockaddr *sa, socklen_t sa_len, peer_listen_mode mode)
 		    &sin->sin_addr, hostbuf, sizeof(hostbuf)),
 		ntohs(sin->sin_port));
 	vtrace("Listening for s3sock scripts on %s\n", listener->desc);
-    }
-#if defined(X3270_IPV6) /*[*/
-    else if (sa->sa_family == AF_INET6) {
+    } else if (sa->sa_family == AF_INET6) {
 	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
 
 	listener->desc = xs_buffer("[%s]:%u", inet_ntop(sa->sa_family,
@@ -748,7 +885,6 @@ peer_init(struct sockaddr *sa, socklen_t sa_len, peer_listen_mode mode)
 		ntohs(sin6->sin6_port));
 	vtrace("Listening for s3sock scripts on %s\n", listener->desc);
     }
-#endif /*]*/
 #if !defined(_WIN32) /*[*/
     else if (sa->sa_family == AF_UNIX) {
 	struct sockaddr_un *ssun = (struct sockaddr_un *)sa;
