@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2000-2021 Paul Mattes.
+ * Copyright (c) 2000-2023 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,6 +30,10 @@
  *
  *	pr3287 [options] [lu[,lu...]@]host[:port]
  *	Options are:
+ *	    -4:
+ *	        prefer IPv4 host addresses
+ *	    -6:
+ *	        prefer IPv6 host addresses
  *	    -accepthostname any|[DNS:]name|address
  *	        accept any certificate hostname, or a specific name, or an
  *	        IP address
@@ -94,6 +98,8 @@
  *		trace data stream to a file
  *          -tracedir dir
  *              directory to write trace file in (POSIX only)
+ *          -tracefile file
+ *              file to write trace file in (POSIX only)
  *          -trnpre file
  *          	file of transparent data to send before jobs
  *          -trnpost file
@@ -127,7 +133,6 @@
 #else /*][*/
 # include <winsock2.h>
 # include <ws2tcpip.h>
-# undef AF_INET6
 #endif /*]*/
 #include <time.h>
 #include <signal.h>
@@ -185,19 +190,27 @@ static char *proxy_portname = NULL;
 static unsigned short proxy_port = 0;
 
 void pr3287_exit(int);
-const char *build_options(void);
 
 /* Print a usage message and exit. */
 static void
-usage(void)
+usage(const char *errmsg)
 {
+    if (errmsg != NULL) {
+	fprintf(stderr, "%s\n", errmsg);
+    }
     fprintf(stderr,
-	    "usage: %s [options] [lu[,lu...]@]host[:port]\n",
+	    "Usage: %s [options] [lu[,lu...]@]host[:port]\n",
 	    programname);
     fprintf(stderr, "Use " OptHelp1 " for the list of options\n");
     pr3287_exit(1);
 }
 
+static void
+missing_value(const char *option)
+{
+    usage(xs_buffer("Missing value for '%s'\n", option));
+}
+
 /* Print command-line help. */
 static void
 cmdline_help(void)
@@ -208,6 +221,9 @@ cmdline_help(void)
 	    "Usage: %s [options] [lu[,lu...]@]host[:port]\n",
 	    programname);
     fprintf(stderr, "Options:\n");
+    fprintf(stderr,
+"  " OptPreferIpv4 "               prefer IPv4 host addresses\n"
+"  " OptPreferIpv6 "               prefer IPv6 host addresses\n");
     if (tls_options & TLS_OPT_ACCEPT_HOSTNAME) {
 	fprintf(stderr,
 "  " OptAcceptHostname " <name>\n"
@@ -296,8 +312,7 @@ cmdline_help(void)
 "  -printercp <codepage>\n"
 "                   code page for output (default is system ANSI code page)\n"
 #endif /*]*/
-"  -proxy \"<spec>\"\n"
-"                   connect to host via specified proxy\n"
+"  -proxy <spec>    connect to host via specified proxy\n"
 "  " OptReconnect "       keep trying to reconnect\n");
     fprintf(stderr,
 "  -skipcc          skip ASA carriage control characters in unformatted host\n"
@@ -306,9 +321,11 @@ cmdline_help(void)
 #if defined(_WIN32) /*[*/
 "  " OptTrace "           trace data stream to <wc3270appData>/x3trc.<pid>.txt\n"
 #else /*][*/
-"  " OptTrace "           trace data stream to /tmp/x3trc.<pid>\n"
+"  " OptTrace "           trace data stream to file (default /tmp/x3trc.<pid>)\n"
 #endif /*]*/
 "  -tracedir <dir>  directory to keep trace information in\n"
+"  " OptTraceFile " <file>\n"
+"                   specific file to write trace information to\n"
 "  -trnpre <file>   file of transparent data to send before each job\n"
 "  -trnpost <file>  file of transparent data to send after each job\n"
 "  -v               display version information and exit\n");
@@ -345,6 +362,7 @@ verrmsg(const char *fmt, va_list ap)
     } else {
 #endif /*]*/
 	fprintf(stderr, "%s: %s\n", programname, buf[ix]);
+	fflush(stderr);
 #if !defined(_WIN32) /*[*/
     }
 #endif /*]*/
@@ -532,6 +550,7 @@ init_options(void)
 #else /*][*/
     options.tracedir		= NULL;
 #endif /*]*/
+    options.tracefile		= NULL;
     options.tracing		= 0;
     options.trnpre		= NULL;
     options.trnpost		= NULL;
@@ -550,18 +569,13 @@ main(int argc, char *argv[])
     char *error;
     char *xtable = NULL;
     unsigned short p;
-    union {
-	struct sockaddr sa;
-	struct sockaddr_in sin;
-#if defined(AF_INET6) && defined(X3270_IPV6) /*[*/
-	struct sockaddr_in6 sin6;
-#endif /*]*/
-    } ha;
-    socklen_t ha_len = sizeof(ha);
     socket_t s = INVALID_SOCKET;
     int rc = 0;
     int report_success = 0;
     unsigned tls_options = sio_all_options_supported();
+    char hn[256];
+    char pn[256];
+    const char *bo;
 
     /* Learn our name. */
 #if defined(_WIN32) /*[*/
@@ -605,26 +619,27 @@ main(int argc, char *argv[])
 	    options.bdaemon = WILL_DAEMON;
 	} else
 #endif /*]*/
-	if ((tls_options & TLS_OPT_ACCEPT_HOSTNAME) &&
+	if (!strcmp(argv[i], OptPreferIpv4)) {
+	    options.prefer_ipv4 = true;
+	} else if (!strcmp(argv[i], OptPreferIpv6)) {
+	    options.prefer_ipv6 = true;
+	} else if ((tls_options & TLS_OPT_ACCEPT_HOSTNAME) &&
 		!strcmp(argv[i], OptAcceptHostname)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptAcceptHostname "\n");
-		usage();
+		missing_value(OptAcceptHostname);
 	    }
 	    options.tls.accept_hostname = argv[i + 1];
 	    i++;
 	} else if (!strcmp(argv[i], "-assoc")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -assoc\n");
-		usage();
+		missing_value("-assoc");
 	    }
 	    options.assoc = argv[i + 1];
 	    i++;
 #if !defined(_WIN32) /*[*/
 	} else if (!strcmp(argv[i], "-command")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -command\n");
-		usage();
+		missing_value("-command");
 	    }
 	    options.command = argv[i + 1];
 	    i++;
@@ -632,80 +647,70 @@ main(int argc, char *argv[])
 	} else if ((tls_options & TLS_OPT_CA_DIR) &&
 		!strcmp(argv[i], OptCaDir)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptCaDir "\n");
-		usage();
+		missing_value(OptCaDir);
 	    }
 	    options.tls.ca_dir = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_CA_FILE) &&
 		!strcmp(argv[i], OptCaFile)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptCaFile "\n");
-		usage();
+		missing_value(OptCaFile);
 	    }
 	    options.tls.ca_file = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_CERT_FILE) &&
 		!strcmp(argv[i], OptCertFile)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptCertFile "\n");
-		usage();
+		missing_value(OptCertFile);
 	    }
 	    options.tls.cert_file = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_CERT_FILE_TYPE) &&
 		!strcmp(argv[i], OptCertFileType)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptCertFileType "\n");
-		usage();
+		missing_value(OptCertFileType);
 	    }
 	    options.tls.cert_file_type = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_CHAIN_FILE) &&
 		!strcmp(argv[i], OptChainFile)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptChainFile "\n");
-		usage();
+		missing_value(OptChainFile);
 	    }
 	    options.tls.chain_file = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_KEY_FILE) &&
 		!strcmp(argv[i], OptKeyFile)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptKeyFile "\n");
-		usage();
+		missing_value(OptKeyFile);
 	    }
 	    options.tls.key_file = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_KEY_FILE_TYPE) &&
 		!strcmp(argv[i], OptKeyFileType)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptKeyFileType "\n");
-		usage();
+		missing_value(OptKeyFileType);
 	    }
 	    options.tls.key_file_type = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_KEY_PASSWD) &&
 		!strcmp(argv[i], OptKeyPasswd)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptKeyPasswd "\n");
-		usage();
+		missing_value(OptKeyPasswd);
 	    }
 	    options.tls.key_passwd = argv[i + 1];
 	    i++;
 	} else if ((tls_options & TLS_OPT_CLIENT_CERT) &&
 		!strcmp(argv[i], OptClientCert)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptClientCert "\n");
-		usage();
+		missing_value(OptClientCert);
 	    }
 	    options.tls.client_cert = argv[i + 1];
 	    i++;
 	} else if (!strcmp(argv[i], OptCharset) ||
 		   !strcmp(argv[i], OptCodePage)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for %s\n", argv[i]);
-		usage();
+		missing_value(argv[i]);
 	    }
 	    options.codepage = argv[i + 1];
 	    i++;
@@ -726,8 +731,7 @@ main(int argc, char *argv[])
 	    options.crthru = 1;
 	} else if (!strcmp(argv[i], "-eojtimeout")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -eojtimeout\n");
-		usage();
+		missing_value("-eojtimeout");
 	    }
 	    options.eoj_timeout = strtoul(argv[i + 1], NULL, 0);
 	    i++;
@@ -742,28 +746,24 @@ main(int argc, char *argv[])
 #if defined(_WIN32) /*[*/
 	} else if (!strcmp(argv[i], "-printer")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -printer\n");
-		usage();
+		missing_value("-printer");
 	    }
 	    options.printer = argv[i + 1];
 	    i++;
 	} else if (!strcmp(argv[i], "-printercp")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -printercp\n");
-		usage();
+		missing_value("-printercp");
 	    }
 	    options.printercp = (int)strtoul(argv[i + 1], NULL, 0);
 	    i++;
 #endif /*]*/
 	} else if (!strcmp(argv[i], "-mpp")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -mpp\n");
-		usage();
+		missing_value("-mpp");
 	    }
 	    options.mpp = (int)strtoul(argv[i + 1], NULL, 0);
 	    if (options.mpp < MIN_UNF_MPP || options.mpp > MAX_UNF_MPP) {
-		fprintf(stderr, "Invalid for -mpp\n");
-		usage();
+		usage("Invalid value for '-mpp'");
 	    }
 	    i++;
 	} else if ((tls_options & TLS_OPT_VERIFY_HOST_CERT) &&
@@ -772,9 +772,11 @@ main(int argc, char *argv[])
 	} else if (!strcmp(argv[i], OptReconnect)) {
 	    options.reconnect = 1;
 	} else if (!strcmp(argv[i], OptV) || !strcmp(argv[i], OptVersion)) {
-	    printf("%s\n%s\n", build, build_options());
+	    fprintf(stderr, "%s\n%s\n", build, bo = build_options());
+	    Free((char *)bo);
+	    fprintf(stderr, "TLS provider: %s\n", sio_provider());
 	    codepage_list();
-	    printf("\n\
+	    fprintf(stderr, "\n\
 Copyright 1989-%s, Paul Mattes, GTRC and others.\n\
 See the source code or documentation for licensing details.\n\
 Distributed WITHOUT ANY WARRANTY; without even the implied warranty of\n\
@@ -787,8 +789,7 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	    options.verbose = 1;
 	} else if (!strcmp(argv[i], "-syncport")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -syncport\n");
-		usage();
+		missing_value("-syncport");
 	    }
 	    options.syncport = (int)strtoul(argv[i + 1], NULL, 0);
 	    i++;
@@ -796,36 +797,37 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	    options.tracing = 1;
 	} else if (!strcmp(argv[i], "-tracedir")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -tracedir\n");
-		usage();
+		missing_value("-tracedir");
 	    }
 	    options.tracedir = argv[i + 1];
 	    i++;
+	} else if (!strcmp(argv[i], OptTraceFile)) {
+	    if (argc <= i + 1 || !argv[i + 1][0]) {
+		missing_value(OptTraceFile);
+	    }
+	    options.tracefile = argv[i + 1];
+	    i++;
 	} else if (!strcmp(argv[i], "-trnpre")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -trnpre\n");
-		usage();
+		missing_value("-trnpre");
 	    }
 	    options.trnpre = argv[i + 1];
 	    i++;
 	} else if (!strcmp(argv[i], "-trnpost")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -trnpost\n");
-		usage();
+		missing_value("-trnpost");
 	    }
 	    options.trnpost = argv[i + 1];
 	    i++;
 	} else if (!strcmp(argv[i], OptProxy)) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for " OptProxy "\n");
-		usage();
+		missing_value(OptProxy);
 	    }
 	    options.proxy_spec = argv[i + 1];
 	    i++;
 	} else if (!strcmp(argv[i], "-xtable")) {
 	    if (argc <= i + 1 || !argv[i + 1][0]) {
-		fprintf(stderr, "Missing value for -xtable\n");
-		usage();
+		missing_value("-xtable");
 	    }
 	    xtable = argv[i + 1];
 	    i++;
@@ -840,12 +842,12 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	    cmdline_help();
 	    exit(0);
 	} else {
-	    fprintf(stderr, "Unknown option '%s'\n", argv[i]);
-	    usage();
+	    fprintf(stderr, "Unknown or incomplete option: '%s'\n", argv[i]);
+	    usage(NULL);
 	}
     }
     if (argc != i + 1) {
-	usage();
+	usage("Too many command-line options");
     }
 
     /*
@@ -876,7 +878,7 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	    HOST_nFLAG(prefixes, PASSTHRU_HOST) ||
 	    HOST_nFLAG(prefixes, STD_DS_HOST) ||
 	    HOST_nFLAG(prefixes, BIND_LOCK_HOST)) {
-	usage();
+	usage(NULL);
     }
 
     if (options.tls_host && !sio_supported()) {
@@ -906,51 +908,56 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	char tracefile[4096];
 	time_t clk;
 	int i;
-	int u = 0;
-	int fd;
+
+	if (options.tracefile != NULL) {
+	    tracef = fopen(options.tracefile, "a");
+	} else {
+	    int u = 0;
+	    int fd;
 #if defined(_WIN32) /*[*/
-	size_t sl;
+	    size_t sl;
 #endif /*]*/
 
-	do {
-	    char dashu[32];
+	    do {
+		char dashu[32];
 
-	    if (u) {
-		snprintf(dashu, sizeof(dashu), "-%d", u);
-	    } else {
-		dashu[0] = '\0';
-	    }
-
-#if defined(_WIN32) /*[*/
-	    if (options.tracedir == NULL) {
-		options.tracedir = "";
-	    }
-	    sl = strlen(options.tracedir);
-	    snprintf(tracefile, sizeof(tracefile),
-		    "%s%sx3trc.%d%s.txt",
-		    options.tracedir,
-		    sl? ((options.tracedir[sl - 1] == '\\')?
-			"": "\\"): "",
-		    getpid(), dashu);
-#else /*][*/
-	    snprintf(tracefile, sizeof(tracefile),
-		    "%s/x3trc.%u%s",
-		    options.tracedir, (unsigned)getpid(), dashu);
-#endif /*]*/
-	    fd = open(tracefile, O_WRONLY | O_CREAT | O_EXCL, 0600);
-	    if (fd < 0) {
-		if (errno != EEXIST) {
-		    perror(tracefile);
-		    pr3287_exit(1);
+		if (u) {
+		    snprintf(dashu, sizeof(dashu), "-%d", u);
+		} else {
+		    dashu[0] = '\0';
 		}
-		u++;
-	    }
-	} while (fd < 0);
+
+#if defined(_WIN32) /*[*/
+		if (options.tracedir == NULL) {
+		    options.tracedir = "";
+		}
+		sl = strlen(options.tracedir);
+		snprintf(tracefile, sizeof(tracefile),
+			"%s%sx3trc.%d%s.txt",
+			options.tracedir,
+			sl? ((options.tracedir[sl - 1] == '\\')?
+			    "": "\\"): "",
+			getpid(), dashu);
+#else /*][*/
+		snprintf(tracefile, sizeof(tracefile),
+			"%s/x3trc.%u%s",
+			options.tracedir, (unsigned)getpid(), dashu);
+#endif /*]*/
+		fd = open(tracefile, O_WRONLY | O_CREAT | O_EXCL, 0600);
+		if (fd < 0) {
+		    if (errno != EEXIST) {
+			perror(tracefile);
+			pr3287_exit(1);
+		    }
+		    u++;
+		}
+	    } while (fd < 0);
     
 #if !defined(_WIN32) /*[*/
-	fcntl(fd, F_SETFD, 1);
+	    fcntl(fd, F_SETFD, 1);
 #endif /*]*/
-	tracef = fdopen(fd, "w");
+	    tracef = fdopen(fd, "w");
+	}
 	if (tracef == NULL) {
 	    perror(tracefile);
 	    pr3287_exit(1);
@@ -958,7 +965,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	SETLINEBUF(tracef);
 	clk = time((time_t *)0);
 	vtrace_nts("Trace started %s", ctime(&clk));
-	vtrace_nts(" Version: %s\n %s\n", build, build_options());
+	vtrace_nts(" Version: %s\n %s\n", build, bo = build_options());
+	Free((char *)bo);
 #if !defined(_WIN32) /*[*/
 	vtrace_nts(" Locale codeset: %s\n", locale_codeset);
 #else /*][*/
@@ -1064,23 +1072,36 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	vtrace("Connected to sync port %d.\n", options.syncport);
     }
 
+    /* Set up -4/-6 host lookup preference. */
+    set_46(options.prefer_ipv4, options.prefer_ipv6);
+
     /*
      * One-time initialization is now complete.
      * (Most) everything beyond this will now be retried, if the -reconnect
      * option is in effect.
      */
     for (;;) {
+	typedef union {
+	    struct sockaddr sa;
+	    struct sockaddr_in sin;
+	    struct sockaddr_in6 sin6;
+	} sockaddr_46_t;
+#       define NUM_HA 4
+	sockaddr_46_t ha[NUM_HA];
+	socklen_t ha_len[NUM_HA];
+	int ha_ix;
 	char *errtxt;
+	int n_ha;
 
 	/* Resolve the host name. */
 	if (proxy_type > 0) {
 	    unsigned long lport;
 	    char *ptr;
 	    struct servent *sp;
-	    int nr;
 
 	    if (resolve_host_and_port(proxy_host, proxy_portname, &proxy_port,
-			&ha.sa, sizeof(ha), &ha_len, &errtxt, 1, &nr) < 0) {
+			&ha[0].sa, sizeof(sockaddr_46_t), ha_len, &errtxt,
+			NUM_HA, &n_ha) < 0) {
 		popup_an_error("%s", errtxt);
 		rc = 1;
 		goto retry;
@@ -1095,28 +1116,46 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 		}
 		p = ntohs(sp->s_port);
 	    } else {
-		    p = (unsigned short)lport;
+		p = (unsigned short)lport;
 	    }
 	} else {
-	    int nr;
-
-	    if (resolve_host_and_port(host, port, &p, &ha.sa, sizeof(ha),
-			&ha_len, &errtxt, 1, &nr) < 0) {
+	    if (resolve_host_and_port(host, port, &p, &ha[0].sa,
+			sizeof(sockaddr_46_t), ha_len, &errtxt, NUM_HA,
+			&n_ha) < 0) {
 		popup_an_error("%s", errtxt);
 		rc = 1;
 		goto retry;
 	    }
 	}
 
-	/* Connect to the host. */
-	s = socket(ha.sa.sa_family, SOCK_STREAM, 0);
-	if (s == INVALID_SOCKET) {
-	    popup_a_sockerr("socket");
-	    pr3287_exit(1);
-	}
+	for (ha_ix = 0; ha_ix < n_ha; ha_ix++) {
+
+	    /* Connect to the host. */
+	    s = socket(ha[ha_ix].sa.sa_family, SOCK_STREAM, 0);
+	    if (s == INVALID_SOCKET) {
+		popup_a_sockerr("socket");
+		pr3287_exit(1);
+	    }
+
+	    if (numeric_host_and_port(&ha[ha_ix].sa, ha_len[ha_ix], hn,
+			sizeof(hn), pn, sizeof(pn), &errtxt)) {
+		vtrace("Trying %s, port %s...\n", hn, pn);
+	    }
+	    if (connect(s, &ha[ha_ix].sa, ha_len[ha_ix]) == 0) {
+		/* Success! */
+		if (ha[ha_ix].sa.sa_family == AF_INET) {
+		    p = htons(ha[ha_ix].sin.sin_port);
+		} else {
+		    p = htons(ha[ha_ix].sin6.sin6_port);
+		}
+		break;
+	    }
 
-	if (connect(s, &ha.sa, ha_len) < 0) {
 	    popup_a_sockerr("%s", (proxy_type > 0)? proxy_host: host);
+	    SOCK_CLOSE(s);
+	    s = INVALID_SOCKET;
+	}
+	if (s == INVALID_SOCKET) {
 	    rc = 1;
 	    goto retry;
 	}
@@ -1163,7 +1202,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 #endif /*]*/
 
 	/* Negotiate. */
-	if (!pr_net_negotiate(host, &ha.sa, ha_len, s, lu, options.assoc)) {
+	if (!pr_net_negotiate(host, &ha[ha_ix].sa, ha_len[ha_ix], s, lu,
+		    options.assoc)) {
 	    rc = 1;
 	    goto retry;
 	}
@@ -1233,24 +1273,19 @@ popup_a_vxerror(pae_t type, const char *fmt, va_list args)
 const char *
 build_options(void)
 {
-    const char *build = NULL;
-
-    if (build == NULL) {
-    	build = xs_buffer("Build options:%s"
+    return xs_buffer("Build options:%s"
 #if defined(_MSC_VER) /*[*/
-	    " via MSVC " xstr(_MSC_VER)
+	" via MSVC " xstr(_MSC_VER)
 #endif /*]*/
 #if defined(__GNUC__) /*[*/
-	    " via gcc " __VERSION__
+	" via gcc " __VERSION__
 #endif /*]*/
 #if defined(__LP64__) || defined(__LLP64__) /*[*/
-	    " 64-bit"
+	" 64-bit"
 #else /*][*/
-	    " 32-bit"
+	" 32-bit"
 #endif /*]*/
-	    , using_iconv()? " -with-iconv": "");
-    }
-    return build;
+	, using_iconv()? " -with-iconv": "");
 }
 
 /* Glue functions to allow proxy.c to link. */
