@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2007-2009, 2014-2016, 2019-2021 Paul Mattes.
+ * Copyright (c) 2007-2024 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -49,7 +49,7 @@
 # include "winvers.h"
 #endif /*]*/
 
-#if defined(X3270_IPV6) && (defined(_WIN32) || defined(HAVE_GETADDRINFO_A)) /*[*/
+#if defined(_WIN32) || defined(HAVE_GETADDRINFO_A) /*[*/
 # define ASYNC_RESOLVER 1
 #endif /*]*/
 
@@ -65,7 +65,7 @@ static struct gai {
     struct gaicb gaicb;		/* control block */
     struct gaicb *gaicbs;	/* control blocks (just one) */
     struct sigevent sigevent;	/* sigevent block */
-    struct addrinfo result;	/* result */
+    struct addrinfo hints;	/* hints */
 # else /*][*/
     int rc;			/* return code */
     struct addrinfo *result;	/* result */
@@ -74,7 +74,29 @@ static struct gai {
 } gai[GAI_SLOTS];
 #endif /*]*/
 
-#if defined(X3270_IPV6) /*[*/
+bool prefer_ipv4;
+bool prefer_ipv6;
+
+/* Set the IPv4/IPv6 lookup preferences. */
+void
+set_46(bool prefer4, bool prefer6)
+{
+    prefer_ipv4 = prefer4;
+    prefer_ipv6 = prefer6;
+}
+
+/* Map the -4 and -6 options onto the right getaddrinfo setting. */
+static int
+want_pf(void)
+{
+    if (prefer_ipv4 && !prefer_ipv6) {
+	return PF_INET;
+    } else if (!prefer_ipv4 && prefer_ipv6) {
+	return PF_INET6;
+    } else {
+	return PF_UNSPEC;
+    }
+}
 
 # if defined(_WIN32) /*[*/
 /* Wrap gai_strerror() in a function that translates the code page. */
@@ -92,7 +114,7 @@ my_gai_strerror(int rc)
  */
 static rhp_t
 resolve_host_and_port_v46(const char *host, char *portname,
-	unsigned short *pport, struct sockaddr *sa, size_t sa_len,
+	bool abs, unsigned short *pport, struct sockaddr *sa, size_t sa_len,
 	socklen_t *sa_rlen, char **errmsg, int max, int *nr)
 {
     struct addrinfo hints, *res0, *res;
@@ -116,7 +138,7 @@ resolve_host_and_port_v46(const char *host, char *portname,
 
     memset(&hints, '\0', sizeof(struct addrinfo));
     hints.ai_flags = 0;
-    hints.ai_family = PF_UNSPEC;
+    hints.ai_family = abs? PF_UNSPEC: want_pf();
     hints.ai_socktype = SOCK_STREAM;
     hints.ai_protocol = IPPROTO_TCP;
     rc = getaddrinfo(host, portname, &hints, &res0);
@@ -161,7 +183,6 @@ resolve_host_and_port_v46(const char *host, char *portname,
     freeaddrinfo(res0);
     return RHP_SUCCESS;
 }
-#endif /*]*/
 
 #if defined(ASYNC_RESOLVER) /*[*/
 
@@ -202,7 +223,7 @@ async_resolve(LPVOID parameter)
     gaip->done = true;
     memset(&hints, '\0', sizeof(struct addrinfo));
     hints.ai_flags = 0;
-    hints.ai_family = PF_UNSPEC;
+    hints.ai_family = want_pf();
     hints.ai_socktype = SOCK_STREAM;
     hints.ai_protocol = IPPROTO_TCP;
     gaip->rc = getaddrinfo(gaip->host, gaip->port, &hints, &gaip->result);
@@ -223,6 +244,26 @@ async_resolve(LPVOID parameter)
 
 # endif /*]*/
 
+/*
+ * Clean up a partially-complete slot.
+ */
+static void
+cleanup_partial_slot(int slot)
+{
+    struct gai *gaip = &gai[slot];
+
+    assert(gaip->busy == true);
+    assert(gaip->done == false);
+
+    gaip->busy = false;
+    gaip->pipe = -1;
+    Replace(gaip->host, NULL);
+    Replace(gaip->port, NULL);
+# if defined(_WIN32) /*[*/
+    gaip->event = INVALID_HANDLE_VALUE;
+# endif /*]*/
+}
+
 /*
  * Resolve a hostname and port using getaddrinfo_a, allowing IPv4 or IPv6.
  * Asynchronous version.
@@ -233,12 +274,24 @@ resolve_host_and_port_v46_a(const char *host, char *portname,
 	socklen_t *sa_rlen, char **errmsg, int max, int *nr, int *slot,
 	int pipe, iosrc_t event)
 {
+    static bool initted = false;
 # if !defined(_WIN32) /*[*/
     int rc;
 # else /*][*/
     HANDLE thread;
 # endif /*]*/
 
+    if (!initted) {
+	int i;
+	for (i = 0; i < GAI_SLOTS; i++) {
+	    gai[i].pipe = -1;
+# if defined(_WIN32) /*[*/
+	    gai[i].event = INVALID_HANDLE_VALUE;
+# endif /*]*/
+	}
+	initted = true;
+    }
+
     *nr = 0;
 
     /* getaddrinfo() does not appear to range-check the port. Do that here. */
@@ -278,15 +331,16 @@ resolve_host_and_port_v46_a(const char *host, char *portname,
     gai[*slot].port = portname? NewString(portname) : NULL;
 
 # if !defined(_WIN32) /*[*/
-    gai[*slot].result.ai_flags = 0;
-    gai[*slot].result.ai_family = PF_UNSPEC;
-    gai[*slot].result.ai_socktype = SOCK_STREAM;
-    gai[*slot].result.ai_protocol = IPPROTO_TCP;
+    gai[*slot].hints.ai_flags = AI_ADDRCONFIG;
+    gai[*slot].hints.ai_family = want_pf();
+    gai[*slot].hints.ai_socktype = SOCK_STREAM;
+    gai[*slot].hints.ai_protocol = IPPROTO_TCP;
 
     gai[*slot].gaicbs = &gai[*slot].gaicb;
     gai[*slot].gaicb.ar_name = host;
-    gai[*slot].gaicb.ar_service = portname;
-    gai[*slot].gaicb.ar_result = &gai[*slot].result;
+    gai[*slot].gaicb.ar_service = gai[*slot].port;
+    gai[*slot].gaicb.ar_request = &gai[*slot].hints;
+    gai[*slot].gaicb.ar_result = NULL;
 
     gai[*slot].sigevent.sigev_notify = SIGEV_THREAD;
     gai[*slot].sigevent.sigev_value.sival_int = *slot;
@@ -298,6 +352,7 @@ resolve_host_and_port_v46_a(const char *host, char *portname,
 	    *errmsg = lazyaf("%s/%s:\n%s", host, portname? portname: "(none)",
 		    my_gai_strerror(rc));
 	}
+	cleanup_partial_slot(*slot);
 	return RHP_CANNOT_RESOLVE;
     }
 # else /*][*/
@@ -307,6 +362,7 @@ resolve_host_and_port_v46_a(const char *host, char *portname,
 	    *errmsg = lazyaf("%s/%s:\n%s", host, portname? portname: "(none)",
 		    win32_strerror(GetLastError()));
 	}
+	cleanup_partial_slot(*slot);
 	return RHP_CANNOT_RESOLVE;
     }
     CloseHandle(thread);
@@ -343,10 +399,7 @@ collect_host_and_port(int slot, struct sockaddr *sa, size_t sa_len,
     case 0:			/* success */
 	/* Return the addresses. */
 	res = gaip->gaicb.ar_result;
-	for (i = 0; i < max && res != NULL; i++, res = res->ai_next) {
-	    if (res->ai_socktype != SOCK_STREAM) {
-		continue;
-	    }
+	for (i = 0; *nr < max && res != NULL; i++, res = res->ai_next) {
 	    memcpy(rsa, res->ai_addr, res->ai_addrlen);
 	    sa_rlen[*nr] = (socklen_t)res->ai_addrlen;
 	    if (i == 0) {
@@ -360,19 +413,18 @@ collect_host_and_port(int slot, struct sockaddr *sa, size_t sa_len,
 		    *pport =
 			ntohs(((struct sockaddr_in6 *)res->ai_addr)->sin6_port);
 		    break;
-		default:
-		    if (errmsg) {
-			*errmsg = lazyaf("unknown family %d", res->ai_family);
-		    }
-		    freeaddrinfo(gaip->gaicb.ar_result);
-		    return RHP_FATAL;
 		}
 	    }
 	    rsa = (char *)rsa + sa_len;
 	    (*nr)++;
 	}
-	freeaddrinfo(gaip->gaicb.ar_result);
+	if (gaip->gaicb.ar_result != NULL) {
+	    freeaddrinfo(gaip->gaicb.ar_result);
+	    gaip->gaicb.ar_result = NULL;
+	}
 	if (*nr) {
+	    Replace(gai->host, NULL);
+	    Replace(gai->port, NULL);
 	    return RHP_SUCCESS;
 	} else {
 	    if (errmsg) {
@@ -380,18 +432,32 @@ collect_host_and_port(int slot, struct sockaddr *sa, size_t sa_len,
 			gaip->port? gaip->port: "(none)",
 			"no suitable resolution");
 	    }
+	    Replace(gai->host, NULL);
+	    Replace(gai->port, NULL);
 	    return RHP_CANNOT_RESOLVE;
 	}
     case EAI_INPROGRESS:	/* still pending, should not happen */
     case EAI_CANCELED:		/* canceled, should not happen */
 	assert(rc != EAI_INPROGRESS && rc != EAI_CANCELED);
+	if (gaip->gaicb.ar_result != NULL) {
+	    freeaddrinfo(gaip->gaicb.ar_result);
+	    gaip->gaicb.ar_result = NULL;
+	}
+	Replace(gai->host, NULL);
+	Replace(gai->port, NULL);
 	return RHP_FATAL;
     default:			/* failure */
+	if (gaip->gaicb.ar_result != NULL) {
+	    freeaddrinfo(gaip->gaicb.ar_result);
+	    gaip->gaicb.ar_result = NULL;
+	}
 	if (errmsg) {
 	    *errmsg = lazyaf("%s/%s:\n%s", gaip->host,
 		    gaip->port? gaip->port: "(none)",
 		    my_gai_strerror(rc));
 	}
+	Replace(gai->host, NULL);
+	Replace(gai->port, NULL);
 	return RHP_CANNOT_RESOLVE;
     }
 
@@ -403,6 +469,8 @@ collect_host_and_port(int slot, struct sockaddr *sa, size_t sa_len,
 		    gaip->port? gaip->port: "(none)",
 		    my_gai_strerror(gaip->rc));
 	}
+	Replace(gai->host, NULL);
+	Replace(gai->port, NULL);
 	return RHP_CANNOT_RESOLVE;
     }
 
@@ -428,6 +496,8 @@ collect_host_and_port(int slot, struct sockaddr *sa, size_t sa_len,
 			    res->ai_family);
 		}
 		freeaddrinfo(gaip->result);
+		Replace(gai->host, NULL);
+		Replace(gai->port, NULL);
 		return RHP_FATAL;
 	    }
 	}
@@ -437,6 +507,8 @@ collect_host_and_port(int slot, struct sockaddr *sa, size_t sa_len,
     }
 
     freeaddrinfo(gaip->result);
+    Replace(gai->host, NULL);
+    Replace(gai->port, NULL);
     return RHP_SUCCESS;
 # endif /*]*/
 #else /*][*/
@@ -467,6 +539,7 @@ cleanup_host_and_port(int slot)
     switch ((rc = gai_error(&gaip->gaicb))) {
     case 0:			/* success */
 	freeaddrinfo(gaip->gaicb.ar_result);
+	gaip->gaicb.ar_result = NULL;
 	break;
     case EAI_INPROGRESS:	/* still pending, should not happen */
     case EAI_CANCELED:		/* canceled, should not happen */
@@ -478,6 +551,7 @@ cleanup_host_and_port(int slot)
 # else /*][*/
     if (gaip->rc == 0) {
 	freeaddrinfo(gaip->result);
+	gaip->result = NULL;
     }
 # endif /*]*/
 
@@ -486,77 +560,73 @@ cleanup_host_and_port(int slot)
 #endif /*]*/
 }
 
-#if !defined(X3270_IPV6) /*[*/
-/*
- * Resolve a hostname and port using gethostbyname and getservbyname, and
- * allowing only IPv4.
- * Synchronous only.
+/**
+ * Mock the behavior of the synchronous resolver.
+ *
+ * @param[in] m		Mock definition
+ * @param[in] host	Host name
+ * @param[in] portname	Port name
+ * @param[out] pport	Returned numeric port
+ * @param[out] sa	Returned array of addresses
+ * @param[in] sa_len	Number of elements in sa
+ * @param[out] sa_rlen	Returned size of elements in sa
+ * @param[out] errmsg	Returned error message
+ * @param[in] max	Maximum number of elements to return
+ * @param[out] nr	Number of elements returned
+ *
+ * @returns @ref rhp_t status
  */
 static rhp_t
-resolve_host_and_port_v4(const char *host, char *portname,
+mock_sync_resolver(const char *m, const char *host, char *portname,
 	unsigned short *pport, struct sockaddr *sa, size_t sa_len,
 	socklen_t *sa_rlen, char **errmsg, int max, int *nr)
 {
-    struct hostent *hp;
-    struct servent *sp;
-    unsigned short port;
-    unsigned long lport;
-    char *ptr;
-    struct sockaddr_in *sin = (struct sockaddr_in *)sa;
-    int i;
-    void *rsa = sa;
+    /*
+     * m is a string that looks like:
+     *  address/port[;address/port...]
+     * address is a numeric IPv4 or IPv6 address
+     * port is a port name or number
+     */
+    char *mdup = NewString(m);
+    char *outer_saveptr = NULL, *inner_saveptr = NULL;
+    char *outer_chunk, *inner_chunk;
+    char *outer_str, *inner_str;
+    struct addrinfo hints;
 
     *nr = 0;
+    memset(&hints, 0, sizeof(struct addrinfo));
+    hints.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
+    hints.ai_family = PF_UNSPEC;
+    hints.ai_socktype = SOCK_STREAM;
 
-    /* Get the port number. */
-    lport = strtoul(portname, &ptr, 0);
-    if (ptr == portname || *ptr != '\0' || lport == 0L || lport & ~0xffff) {
-	if (!(sp = getservbyname(portname, "tcp"))) {
-	    if (errmsg) {
-		*errmsg = lazyaf("Unknown port number or service: %s",
-			portname);
-	    }
-	    return RHP_FATAL;
+    outer_str = mdup;
+    while (*nr < max &&
+	    (outer_chunk = strtok_r(outer_str, ";", &outer_saveptr)) != NULL) {
+	int np = 0;
+	char *inner[2];
+	struct addrinfo *res = NULL;
+
+	outer_str = NULL;
+	inner_str = outer_chunk;
+	while ((inner_chunk = strtok_r(inner_str, "/", &inner_saveptr))
+		!= NULL) {
+	    inner_str = NULL;
+	    assert(np < 2);
+	    inner[np++] = inner_chunk;
 	}
-	port = sp->s_port;
-    } else {
-	port = htons((unsigned short)lport);
-    }
-    *pport = ntohs(port);
+	assert(np == 2);
 
-    /* Use gethostbyname() to resolve the hostname. */
-    hp = gethostbyname(host);
-    if (hp == (struct hostent *)0) {
-	/* Try inet_addr(). */
-	sin->sin_family = AF_INET;
-	sin->sin_addr.s_addr = inet_addr(host);
-	if (sin->sin_addr.s_addr == INADDR_NONE) {
-	    if (errmsg) {
-		*errmsg = lazyaf("Unknown host:\n%s", host);
-	    }
-	    return RHP_CANNOT_RESOLVE;
-	}
-	*sa_rlen = sizeof(struct sockaddr_in);
-	*nr = 1;
-	return RHP_SUCCESS;
-    }
-
-    /* Return the addresses. */
-    for (i = 0; i < max && hp->h_addr_list[i] != NULL; i++) {
-	struct sockaddr_in *rsin = rsa;
-
-	rsin->sin_family = hp->h_addrtype;
-	memcpy(&rsin->sin_addr, hp->h_addr_list[i], hp->h_length);
-	rsin->sin_port = port;
-	sa_rlen[*nr] = sizeof(struct sockaddr_in);
-
-	rsa = (char *)rsa + sa_len;
-	(*nr)++;
+	assert(getaddrinfo(inner[0], inner[1], &hints, &res) == 0);
+	memcpy(sa, res->ai_addr, res->ai_addrlen);
+	sa_rlen[*nr] = (socklen_t)res->ai_addrlen;
+	freeaddrinfo(res);
+	sa = (struct sockaddr *)((char *)sa + sa_len);
+	++(*nr);
     }
 
+    Free(mdup);
     return RHP_SUCCESS;
 }
-#endif /*]*/
 
 /**
  * Resolve a hostname and port.
@@ -579,13 +649,39 @@ resolve_host_and_port(const char *host, char *portname, unsigned short *pport,
 	struct sockaddr *sa, size_t sa_len, socklen_t *sa_rlen, char **errmsg,
 	int max, int *nr)
 {
-#if defined(X3270_IPV6) /*[*/
-    return resolve_host_and_port_v46(host, portname, pport, sa, sa_len,
+    const char *m = getenv("MOCK_SYNC_RESOLVER");
+
+    if (m != NULL && *m != '\0') {
+	return mock_sync_resolver(m, host, portname, pport, sa, sa_len,
+		sa_rlen, errmsg, max, nr);
+    }
+    return resolve_host_and_port_v46(host, portname, false, pport, sa, sa_len,
 	    sa_rlen, errmsg, max, nr);
-#else /*][*/
-    return resolve_host_and_port_v4(host, portname, pport, sa, sa_len,
+}
+
+/**
+ * Resolve a hostname and port.
+ * Synchronous version, without -4/-6 preferences.
+ *
+ * @param[in] host	Host name
+ * @param[in] portname	Port name
+ * @param[out] pport	Returned numeric port
+ * @param[out] sa	Returned array of addresses
+ * @param[in] sa_len	Number of elements in sa
+ * @param[out] sa_rlen	Returned size of elements in sa
+ * @param[out] errmsg	Returned error message
+ * @param[in] max	Maximum number of elements to return
+ * @param[out] nr	Number of elements returned
+ *
+ * @returns RHP_XXX status
+ */
+rhp_t
+resolve_host_and_port_abs(const char *host, char *portname,
+	unsigned short *pport, struct sockaddr *sa, size_t sa_len,
+	socklen_t *sa_rlen, char **errmsg, int max, int *nr)
+{
+    return resolve_host_and_port_v46(host, portname, true, pport, sa, sa_len,
 	    sa_rlen, errmsg, max, nr);
-#endif
 }
 
 /*
@@ -610,18 +706,14 @@ resolve_host_and_port_a(const char *host, char *portname, unsigned short *pport,
 	int max, int *nr, int *slot, int pipe, iosrc_t event)
 {
 #if defined(ASYNC_RESOLVER) /*[*/
-    return resolve_host_and_port_v46_a(host, portname, pport, sa, sa_len,
-	    sa_rlen, errmsg, max, nr, slot, pipe, event);
-#else /*][*/
-    *slot = -1;
-# if defined(X3270_IPV6)  /*][*/
-    return resolve_host_and_port_v46(host, portname, pport, sa, sa_len,
-	    sa_rlen, errmsg, max, nr);
-# else /*][*/
-    return resolve_host_and_port_v4(host, portname, pport, sa, sa_len,
-	    sa_rlen, errmsg, max, nr);
-# endif /*]*/
+    if (getenv("SYNC_RESOLVER") == NULL) {
+	return resolve_host_and_port_v46_a(host, portname, pport, sa, sa_len,
+		sa_rlen, errmsg, max, nr, slot, pipe, event);
+    }
 #endif /*]*/
+    *slot = -1;
+    return resolve_host_and_port_v46(host, portname, false, pport, sa, sa_len,
+	    sa_rlen, errmsg, max, nr);
 }
 
 #if defined(_WIN32) /*[*/
@@ -638,7 +730,6 @@ bool
 numeric_host_and_port(const struct sockaddr *sa, socklen_t salen, char *host,
 	size_t hostlen, char *serv, size_t servlen, char **errmsg)
 {
-#if defined(X3270_IPV6) /*[*/
     int rc;
 
     /* Use getnameinfo(). */
@@ -651,12 +742,4 @@ numeric_host_and_port(const struct sockaddr *sa, socklen_t salen, char *host,
 	return false;
     }
     return true;
-#else /*][*/
-    struct sockaddr_in *sin = (struct sockaddr_in *)sa;
-
-    /* Use inet_ntoa() and snprintf(). */
-    snprintf(host, hostlen, "%s", inet_ntoa(sin->sin_addr));
-    snprintf(serv, servlen, "%u", ntohs(sin->sin_port));
-    return true;
-#endif /*]*/
 }
