@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2017, 2019-2020 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * 
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -90,6 +90,8 @@ static ssl_sio_t *current_sio;
 # define INFO_CONST
 #endif /*]*/
 
+static int proto_map[] = { -1, SSL3_VERSION, TLS1_VERSION, TLS1_1_VERSION, TLS1_2_VERSION, TLS1_3_VERSION };
+
 static void client_info_callback(INFO_CONST SSL *s, int where, int ret);
 #if !defined(OPENSSL102) /*[*/
 static char *spc_verify_cert_hostname(X509 *cert, const char *hostname);
@@ -185,7 +187,7 @@ passwd_cb(char *buf, int size, int rwflag _is_unused, void *userdata)
     if (pass_len > size - 1) {
 	pass_len = size - 1;
     }
-    strncpy(buf, p, pass_len);
+    strncpy(buf, p, size - 1);
     buf[pass_len] = '\0';
     if (need_free) {
 	Free(p);
@@ -208,9 +210,8 @@ parse_file_type(const char *s)
 static char *
 get_ssl_error(char *buf)
 {
-    unsigned long e;
+    unsigned long e = ERR_get_error();
 
-    e = ERR_get_error();
     if (getenv("SSL_VERBOSE_ERRORS")) {
 	ERR_error_string(e, buf);
     } else {
@@ -478,6 +479,9 @@ sio_init(tls_config_t *config, const char *password, sio_t *sio_ret)
     char err_buf[120];
     int cert_file_type = SSL_FILETYPE_PEM;
     sio_init_ret_t err_ret = SI_FAILURE;
+    int min_protocol = -1;
+    int max_protocol = -1;
+    char *proto_error;
 
     sioc_error_reset();
 
@@ -498,6 +502,21 @@ sio_init(tls_config_t *config, const char *password, sio_t *sio_ret)
 	goto fail;
     }
     SSL_CTX_set_options(s->ctx, SSL_OP_ALL);
+    proto_error = sioc_parse_protocol_min_max(config->min_protocol, config->max_protocol, SIP_SSL3, -1, &min_protocol,
+	    &max_protocol);
+    if (proto_error != NULL) {
+	sioc_set_error("%s", proto_error);
+	Free(proto_error);
+	goto fail;
+    }
+    if (min_protocol >= 0 && SSL_CTX_set_min_proto_version(s->ctx, proto_map[min_protocol]) == 0) {
+	sioc_set_error("SSL_CTX_set_min_proto_version failed");
+	goto fail;
+    }
+    if (max_protocol >= 0 && SSL_CTX_set_max_proto_version(s->ctx, proto_map[max_protocol]) == 0) {
+	sioc_set_error("SSL_CTX_set_max_proto_version failed");
+	goto fail;
+    }
     SSL_CTX_set_info_callback(s->ctx, client_info_callback);
     SSL_CTX_set_default_passwd_cb_userdata(s->ctx, s);
     SSL_CTX_set_default_passwd_cb(s->ctx, passwd_cb);
@@ -753,6 +772,7 @@ display_cert(varbuf_t *v, X509 *cert, int level, const char *who)
 	    if (header) {
 		vb_appendf(v, "\n");
 	    }
+	    sk_GENERAL_NAME_pop_free(values, GENERAL_NAME_free);
 	}
     }
 }
@@ -854,6 +874,7 @@ display_server_subjects(varbuf_t *v, SSL *con)
 		OPENSSL_free(dns);
 	    }
 	}
+	sk_GENERAL_NAME_pop_free(values, GENERAL_NAME_free);
     }
     sioc_subject_print(v, &subjects);
 }
@@ -941,7 +962,7 @@ sio_negotiate(sio_t sio, socket_t sock, const char *hostname, bool *data)
 	    sioc_set_error("Host certificate verification failed:\n%s (%ld)%s",
 		    X509_verify_cert_error_string(vr), vr,
 		    (vr == X509_V_ERR_HOSTNAME_MISMATCH)?
-			"\nPossibly use " AnSubjectNames "() to list the host cert names":
+			"\nTry Y: to connect and " AnShow "(" KwTlsSubjectNames ") to list names":
 			"");
 	    return SIG_FAILURE;
 	}
@@ -950,10 +971,21 @@ sio_negotiate(sio_t sio, socket_t sock, const char *hostname, bool *data)
     }
 
     if (rv != 1) {
+	unsigned long e = SSL_get_error(s->con, rv);
 	char err_buf[120];
 
-	sioc_set_error("SSL_connect failed %d:\n%s", rv,
-		get_ssl_error(err_buf));
+	if (e == SSL_ERROR_SYSCALL) {
+	    if (errno == 0) {
+		sioc_set_error("SSL_connect failed:\nUnexpected EOF");
+	    } else {
+		sioc_set_error("SSL_connect failed:\n%s", strerror(errno));
+	    }
+	} else if (e == SSL_ERROR_ZERO_RETURN) {
+	    sioc_set_error("SSL_connect failed:\nUnexpected EOF");
+	} else {
+	    sioc_set_error("SSL_connect failed %d/%ld:\n%s", rv, e,
+		    get_ssl_error(err_buf));
+	}
 	return SIG_FAILURE;
     }
 
@@ -1089,9 +1121,6 @@ sio_close(sio_t sio)
 	return;
     }
     s = (ssl_sio_t *)sio;
-    if (s->con == NULL || s->sock == INVALID_SOCKET) {
-	return;
-    }
 
     if (s->ctx != NULL) {
 	SSL_CTX_free(s->ctx);
@@ -1114,9 +1143,11 @@ sio_close(sio_t sio)
 	s->server_subjects = NULL;
     }
 
-    SSL_shutdown(s->con);
-    SSL_free(s->con);
-    s->con = NULL;
+    if (s->con != NULL) {
+	SSL_shutdown(s->con);
+	SSL_free(s->con);
+	s->con = NULL;
+    }
 
     s->sock = INVALID_SOCKET;
 
@@ -1141,7 +1172,8 @@ sio_options_supported(void)
 {
     return TLS_OPT_CA_DIR | TLS_OPT_CA_FILE | TLS_OPT_CERT_FILE
 	| TLS_OPT_CERT_FILE_TYPE | TLS_OPT_CHAIN_FILE | TLS_OPT_KEY_FILE
-	| TLS_OPT_KEY_FILE_TYPE | TLS_OPT_KEY_PASSWD;
+	| TLS_OPT_KEY_FILE_TYPE | TLS_OPT_KEY_PASSWD | TLS_OPT_MIN_PROTOCOL
+	| TLS_OPT_MAX_PROTOCOL;
 }
 
 /*
