@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2021 Paul Mattes.
+ * Copyright (c) 1993-2023 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * All rights reserved.
  *
@@ -46,6 +46,7 @@
 #include "lazya.h"
 #include "popups.h"
 #include "product.h"
+#include "telnet.h"
 #include "trace.h"
 #include "unicodec.h"
 #include "varbuf.h"
@@ -177,6 +178,35 @@ qscatv(const char *s)
     return catv_common(s, true);
 }
 
+/**
+ * Definition resource splitter, for resources of the repeating form:
+ *	left: right\n
+ * Can be called iteratively to parse a list. Set offset to 0 to begin.
+ *
+ * @param[in] st		String to parse
+ * @param[in,out] offset	Offset within string
+ * @param[out] left		Returned left side
+ * @param[out] right		Returned right side
+ *
+ * @returns 1 for success, 0 for EOF, -1 for error.
+ */
+int
+s_split_dresource(const char *st, size_t *offset, char **left, char **right)
+{
+    char *st_copy = NewString(st + *offset);
+    char *st_copy0 = st_copy;
+    char *l, *r;
+    int ret = split_dresource(&st_copy, &l, &r);
+
+    if (ret > 0) {
+	*left = NewString(l);
+	*right = NewString(r);
+	*offset += st_copy - st_copy0;
+    }
+    Free(st_copy0);
+    return ret;
+}
+
 /*
  * Definition resource splitter, for resources of the repeating form:
  *	left: right\n
@@ -441,7 +471,7 @@ ex_getenv(const char *name, unsigned long flags, int *up)
 	    return xs_buffer("%u-%u", (unsigned)getpid(), *up);
 	}
     } else {
-	return getenv(name);
+	return NewString(getenv(name));
     }
 }
 
@@ -549,6 +579,7 @@ var_subst(const char *s, unsigned long flags)
 			o = strchr(ob, '\0');
 			strcpy(o, vv);
 			o += strlen(vv);
+			Free(vv);
 		    }
 		    Free(vn);
 		    if (state == VS_VNB) {
@@ -754,7 +785,7 @@ add_resource(const char *name, const char *value)
     }
     d = Malloc(sizeof(struct dresource));
     d->next = NULL;
-    d->name = name;
+    d->name = NewString(name);
     d->value = NewString(value);
     *drdb_next = d;
     drdb_next = &d->next;
@@ -868,13 +899,12 @@ strip_whitespace(const char *s)
  * Hierarchy (a>b>c) splitter.
  */
 bool
-split_hier(char *label, char **base, char ***parents)
+split_hier(const char *label, char **base, char ***parents)
 {
     int n_parents = 0;
     char *gt;
-    char *lp;
+    const char *lp;
 
-    label = NewString(label);
     for (lp = label; (gt = strchr(lp, '>')) != NULL; lp = gt + 1) {
 	if (gt == lp) {
 	    return false;
@@ -890,17 +920,28 @@ split_hier(char *label, char **base, char ***parents)
 	for (n_parents = 0, lp = label;
 	     (gt = strchr(lp, '>')) != NULL;
 	     lp = gt + 1) {
-	    (*parents)[n_parents++] = lp;
-	    *gt = '\0';
+	    (*parents)[n_parents++] = xs_buffer("%.*s", (int)(gt - lp), lp);
 	}
-	*base = lp;
+	*base = NewString(lp);
     } else {
 	(*parents) = NULL;
-	(*base) = label;
+	(*base) = NewString(label);
     }
     return true;
 }
 
+/* Free a parent list. */
+void
+free_parents(char **parents)
+{
+    int i;
+
+    for (i = 0; parents[i] != NULL; i++) {
+	Free(parents[i]);
+    }
+    Free(parents);
+}
+
 #if defined(_MSC_VER) /*[*/
 #define xstr(s)	str(s)
 #define str(s)	#s
@@ -916,7 +957,7 @@ build_options(void)
 	p = "";
     }
 
-    return lazyaf("%s%s",
+    return lazyaf("%s%s%s%s",
 #if defined(X3270_LOCAL_PROCESS) /*[*/
 	    "--enable-local-process"
 #else /*][*/
@@ -941,6 +982,7 @@ void
 dump_version(void)
 {
     fprintf(stderr, "%s\nBuild options: %s\n", build, build_options());
+    fprintf(stderr, "TLS provider: %s\n", net_sio_provider());
     codepage_list();
     fprintf(stderr, "\n"
 "Copyright 1989-%s, Paul Mattes, GTRC and others.\n"
@@ -981,7 +1023,7 @@ clean_termname(const char *tn)
     char *ret;
 
     if (tn == NULL) {
-	return (char *)tn;
+	return NULL;
     }
 
     while (*s && isspace((unsigned char)*s)) {
