@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1995-2009, 2013-2021 Paul Mattes.
+ * Copyright (c) 1995-2024 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -85,6 +85,7 @@
 #endif
 
 #include "base64.h"
+#include "names.h"
 #include "s3270_proto.h"
 #include "w3misc.h"
 
@@ -126,7 +127,7 @@ static void interactive_io(int port, const char *emulator_name,
 	const char *help_name, const char *localization);
 
 #if defined(HAVE_LIBREADLINE) /*[*/
-static char **attempted_completion();
+static char **attempted_completion(const char *text, int start, int end);
 static char *completion_entry(const char *, int);
 #endif /*]*/
 
@@ -556,6 +557,14 @@ retry:
 	    if (verbose) {
 		fprintf(stderr, "i+ in %s\n", buf);
 	    }
+	    if (buf[0] == '{') {
+		/* JSON output. */
+		printf("%s\n", buf);
+		fflush(stdout);
+		xs = 0;
+		done = 1;
+		break;
+	    }
 	    if (!strcmp(buf, PROMPT_OK)) {
 		fflush(stdout);
 		xs = 0;
@@ -1274,7 +1283,7 @@ interactive_io(int port, const char *emulator_name, const char *help_name,
 
     /* Announce our capabilities. */
     data_ret = NULL;
-    single_io(0, 0, s, infd, outfd, NO_STATUS, "Capabilities(interactive)",
+    single_io(0, 0, s, infd, outfd, NO_STATUS, AnCapabilities "(" KwInteractive ")",
 	    &data_ret, NULL, &itype);
 
     /* Set up the prompt. */
