@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2000-2020 Paul Mattes.
+ * Copyright (c) 2000-2024 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,6 +32,7 @@
  */
 
 #include "globals.h"
+
 #include <signal.h>
 #include "appres.h"
 #include "3270ds.h"
@@ -76,6 +77,11 @@
  */
 #undef COLS
 
+#if defined(CURSES_WIDE) /*[*/
+# include <wchar.h>
+# define NCURSES_WIDECHAR 1
+#endif /*]*/
+
 #if defined(HAVE_NCURSESW_NCURSES_H) /*[*/
 # include <ncursesw/ncurses.h>
 #elif defined(HAVE_NCURSES_NCURSES_H) /*][*/
@@ -103,6 +109,9 @@
 
 #define CM (60*10)	/* csec per minute */
 
+/* Delay for meta-escape mode. */
+#define ME_DELAY	25L
+
 #if !defined(HAVE_TIPARM) /*[*/
 #define tiparm		tparm
 #endif /*]*/
@@ -245,6 +254,7 @@ static ioid_t input_id = NULL_IOID;
 
 static int rmargin;
 
+bool screen_initted = false;
 bool escaped = true;
 bool initscr_done = false;
 int curs_set_state = -1;
@@ -299,7 +309,6 @@ static bool curses_alt = false;
 #if defined(HAVE_USE_DEFAULT_COLORS) /*[*/
 static bool default_colors = false;
 #endif /*]*/
-static bool screen_initted = false;
 
 static ioid_t disabled_done_id = NULL_IOID;
 
@@ -530,6 +539,11 @@ finish_screen_init(void)
     bool oversize = false;
     char *cl;
 
+    /* Dummy globals so Valgrind doesn't think the arguments to putenv() are leaked. */
+    static char *pe_columns = NULL;
+    static char *pe_lines = NULL;
+    static char *pe_escdelay = NULL;
+
     if (screen_initted) {
 	return;
     }
@@ -539,14 +553,18 @@ finish_screen_init(void)
     /* Clear the (original) screen first. */
 #if defined(C3270_80_132) /*[*/
     if (appres.c3270.defscreen != NULL) {
-	putenv(xs_buffer("COLUMNS=%d", defscreen_spec.cols));
-	putenv(xs_buffer("LINES=%d", defscreen_spec.rows));
+	putenv((pe_columns = xs_buffer("COLUMNS=%d", defscreen_spec.cols)));
+	putenv((pe_lines = xs_buffer("LINES=%d", defscreen_spec.rows)));
     }
 #endif /*]*/
     if ((cl = tigetstr("clear")) != NULL) {
 	putp(cl);
     }
 
+    if (getenv("ESCDELAY") == NULL) {
+	putenv((pe_escdelay = xs_buffer("ESCDELAY=%ld", ME_DELAY)));
+    }
+
 #if !defined(C3270_80_132) /*[*/
     /* Initialize curses. */
     if (initscr() == NULL) {
@@ -1632,8 +1650,10 @@ kybd_input(iosrc_t fd _is_unused, ioid_t id _is_unused)
 	}
 	ucs4 = 0;
 #if defined(CURSES_WIDE) /*[*/
+	vtrace("kybd_input: calling wget_wch()\n");
 	k = wget_wch(stdscr, &wch);
 #else /*][*/
+	vtrace("kybd_input: calling wgetch()\n");
 	k = wgetch(stdscr);
 #endif /*]*/
 	vtrace("kbd_input: k=%d "
@@ -1756,7 +1776,7 @@ kybd_input(iosrc_t fd _is_unused, ioid_t id _is_unused)
 	} else if (me_mode == TS_ON && ucs4 == 0x1b) {
 	    vtrace("Key '%s' (curses key 0x%x, char code 0x%x)\n",
 		    decode_key(k, ucs4, alt, dbuf), k, ucs4);
-	    eto = AddTimeOut(100L, escape_timeout);
+	    eto = AddTimeOut(ME_DELAY, escape_timeout);
 	    vtrace(" waiting to see if Escape is followed by another key\n");
 	    meta_escape = true;
 	    continue;
@@ -2240,14 +2260,7 @@ status_oerr(int error_type)
 void
 status_reset(void)
 {
-    if (kybdlock & KL_ENTER_INHIBIT) {
-	other_msg = "X Inhibit";
-    } else if (kybdlock & KL_DEFERRED_UNLOCK) {
-	other_msg = "X";
-    } else {
-	status_connect(PCONNECTED);
-    }
-    other_attr = status_colors(defcolor_offset + COLOR_WHITE) | A_BOLD;
+    status_connect(PCONNECTED);
 }
 
 void
@@ -2325,6 +2338,14 @@ status_connect(bool connected)
 	    other_msg = "X [TN3270E]";
 	} else if (kybdlock & KL_AWAITING_FIRST) {
 	    other_msg = "X [Field]";
+	} else if (kybdlock & KL_ENTER_INHIBIT) {
+	    other_msg = "X Inhibit";
+	} else if (kybdlock & KL_BID) {
+	    other_msg = "X Wait";
+	} else if (kybdlock & KL_FT) {
+	    other_msg = "X File Transfer";
+	} else if (kybdlock & KL_DEFERRED_UNLOCK) {
+	    other_msg = "X";
 	} else {
 	    other_msg = NULL;
 	}
@@ -3080,7 +3101,7 @@ display_ge(unsigned char ebc)
 }
 
 void
-screen_final()
+screen_final(void)
 {
     char *cl;
 
@@ -3091,9 +3112,10 @@ screen_final()
 
 /**
  * Check if an area of the screen is selected.
- * This is implemented by wc3270, but not c3270, so it is a stub here.
  *
  * @param[in] baddr	Buffer address.
+ *
+ * @return true if cell is selected
  */
 bool
 screen_selected(int baddr _is_unused)
@@ -3102,7 +3124,7 @@ screen_selected(int baddr _is_unused)
 }
 
 /**
- * Stub for scrollbar function.
+ * Set the scrollbar thumb.
  *
  * @param[in] top	Where the top of the scrollbar should be (percentage)
  * @param[in] shown	How much of the scrollbar to show (percentage)
@@ -3117,7 +3139,7 @@ screen_set_thumb(float top _is_unused, float shown _is_unused,
 }
 
 /**
- * Enable or disable the cursor when scrolling.
+ * Enable or disable the cursor.
  *
  * @param[in] on	Enable (true) or disable (false) the cursor display.
  */
