@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1995-2020, Paul Mattes.
+ * Copyright (c) 1995-2023 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -76,6 +76,7 @@
 #define ResConnectTimeout	"connectTimeout"
 #define ResConsole		"console"
 #define ResConsoleColorForHostColor "consoleColorForHostColor"
+#define ResContentionResolution	"contentionResolution"
 #define ResCrosshair		"crosshair"
 #define ResCrosshairColor	"crosshairColor"
 #define ResCursesColorFor	"cursesColorFor"
@@ -111,6 +112,7 @@
 #define ResFtLocalFile		"ftLocalFile"
 #define ResFtLrecl		"ftLrecl"
 #define ResFtMode		"ftMode"
+#define ResFtOtherOptions	"ftOtherOptions"
 #define ResFtPrimarySpace	"ftPrimarySpace"
 #define ResFtRecfm		"ftRecfm"
 #define ResFtRemap		"ftRemap"
@@ -131,12 +133,14 @@
 #define ResIdleCommand		"idleCommand"
 #define ResIdleCommandEnabled	"idleCommandEnabled"
 #define ResIdleTimeout		"idleTimeout"
+#define ResIndent		"indent"
 #define ResInlcr		"inlcr"
 #define ResInputColor		"inputColor"
 #define ResInputMethod		"inputMethod"
 #define ResInsertMode		"insertMode"
 #define ResIntr			"intr"
 #define ResInvertKeypadShift	"invertKeypadShift"
+#define ResJson			"json"
 #define ResKeyFile		"keyFile"
 #define ResKeyFileType		"keyFileType"
 #define ResKeymap		"keymap"
@@ -181,6 +185,8 @@
 #define ResOversize		"oversize"
 #define ResPort			"port"
 #define ResPreeditType		"preeditType"
+#define ResPreferIpv4		"preferIpv4"
+#define ResPreferIpv6		"preferIpv6"
 #define ResPrinterCodepage	"printer.codepage"
 #define ResPrinterCommand	"printer.command"
 #define ResPrinterLu		"printerLu"
@@ -192,6 +198,7 @@
 #define ResQrBgColor		"qrBgColor"
 #define ResReconnect		"reconnect"
 #define ResRectangleSelect	"rectangleSelect"
+#define ResRetry		"retry"
 #define ResReverseInputMode	"reverseInputMode"
 #define ResReverseVideo		"reverseVideo"
 #define ResRightToLeftMode	"rightToLeftMode"
@@ -203,6 +210,7 @@
 #define ResScreenTraceTarget	"screenTraceTarget"
 #define ResScreenTraceType	"screenTraceType"
 #define ResScripted		"scripted"
+#define ResScriptedAlways	"scriptedAlways"
 #define ResScriptPort		"scriptPort"
 #define ResScriptPortOnce	"scriptPortOnce"
 #define ResScrollBar		"scrollBar"
@@ -219,6 +227,8 @@
 #define ResSuppress		"suppress"
 #define ResTermName		"termName"
 #define ResTitle		"title"
+#define ResTlsMaxProtocol	"tlsMaxProtocol"
+#define ResTlsMinProtocol	"tlsMinProtocol"
 #define ResTrace		"trace"
 #define ResTraceDir		"traceDir"
 #define ResTraceFile		"traceFile"
@@ -226,6 +236,7 @@
 #define ResTraceMonitor		"traceMonitor"
 #define ResTypeahead		"typeahead"
 #define ResUnderscore		"underscore"
+#define ResUnderscoreBlankFill	"underscoreBlankFill"
 #define ResUnlockDelay		"unlockDelay"
 #define ResUnlockDelayMs	"unlockDelayMs"
 #define ResUseCursorColor	"useCursorColor"
@@ -238,6 +249,8 @@
 #define ResVisualSelectColor	"visualSelectColor"
 #define ResWaitCursor		"waitCursor"
 #define ResWerase		"werase"
+#define ResWrapperDoc		"wrapperDoc"
+#define ResWrongTerminalName	"wrongTerminalName"
 #define ResXQuartzHack		"xQuartzHack"
 
 /* Dotted resource names. */
@@ -267,6 +280,8 @@
 #define DotKeypadOn		"." ResKeypadOn
 #define DotKeyPasswd		"." ResKeyPasswd
 #define DotLoginMacro		"." ResLoginMacro
+#define DotTlsMaxProtocol	"." ResTlsMaxProtocol
+#define DotTlsMinProtocol	"." ResTlsMinProtocol
 #define DotMinVersion		"." ResMinVersion
 #define DotModel		"." ResModel
 #define DotMono			"." ResMono
@@ -276,6 +291,8 @@
 #define DotOversize		"." ResOversize
 #define DotPort			"." ResPort
 #define DotPreeditType		"." ResPreeditType
+#define DotPreferIpv4		"." ResPreferIpv4
+#define DotPreferIpv6		"." ResPreferIpv6
 #define DotPrinterLu		"." ResPrinterLu
 #define DotProxy		"." ResProxy
 #define DotReconnect		"." ResReconnect
@@ -331,6 +348,7 @@
 #define ClsConnectFileName	"ConnectFileName"
 #define ClsConnectTimeout	"ConnectTimeout"
 #define ClsConsole		"Console"
+#define ClsContentionResolution	"ContentionResolution"
 #define ClsCrosshair		"Crosshair"
 #define ClsCrosshairColor	"CrosshairColor"
 #define ClsCursorBlink		"CursorBlink"
@@ -357,11 +375,12 @@
 #define ClsFtLocalFile		"FtLocalFile"
 #define ClsFtLrecl		"FtLrecl"
 #define ClsFtMode		"FtMode"
+#define ClsFtOtherOptions	"FtOtherOptions"
 #define ClsFtPrimarySpace	"FtPrimarySpace"
 #define ClsFtRecfm		"FtRecfm"
 #define ClsFtRemap		"FtRemap"
 #define ClsFtSecondarySpace	"FtSecondarySpace"
-#define ClsFtWindowsCodePage	"ClsWindowsCodePage"
+#define ClsFtWindowsCodePage	"FtWindowsCodePage"
 #define ClsHighlightBold	"HighlightBold"
 #define ClsHostname		"Hostname"
 #define ClsHostsFile		"HostsFile"
@@ -417,11 +436,14 @@
 #define ClsOversize		"Oversize"
 #define ClsPort			"Port"
 #define ClsPreeditType		"PreeditType"
+#define ClsPreferIpv4		"PreferIpv4"
+#define ClsPreferIpv6		"PreferIpv6"
 #define ClsPrinterLu		"PrinterLu"
 #define ClsProxy		"Proxy"
 #define ClsQuit			"Quit"
 #define ClsReconnect		"Reconnect"
 #define ClsRectangleSelect	"RectangleSelect"
+#define ClsRetry		"Retry"
 #define ClsReverseInputMode	"ReverseInputMode"
 #define ClsRightToLeftMode	"RightToLeftMode"
 #define ClsRprnt		"Rprnt"
@@ -432,6 +454,7 @@
 #define ClsScreenTraceTarget	"ScreenTraceTarget"
 #define ClsScreenTraceType	"ScreenTraceType"
 #define ClsScripted		"Scripted"
+#define ClsScriptedAlways	"ScriptedAlways"
 #define ClsScriptPort		"ScriptPort"
 #define ClsScriptPortOnce	"ScriptPortOnce"
 #define ClsScrollBar		"ScrollBar"
@@ -445,12 +468,15 @@
 #define ClsSuppressHost		"SuppressHost"
 #define ClsSuppressFontMenu	"SuppressFontMenu"
 #define ClsTermName		"TermName"
+#define ClsTlsMaxProtocol	"TlsMaxProtocol"
+#define ClsTlsMinProtocol	"TlsMinProtocol"
 #define ClsTrace		"Trace"
 #define ClsTraceDir		"TraceDir"
 #define ClsTraceFile		"TraceFile"
 #define ClsTraceFileSize	"TraceFileSize"
 #define ClsTraceMonitor		"TraceMonitor"
 #define ClsTypeahead		"Typeahead"
+#define ClsUnderscoreBlankFill	"UnderscoreBlankFill"
 #define ClsUnlockDelay		"UnlockDelay"
 #define ClsUnlockDelayMs	"UnlockDelayMs"
 #define ClsUseCursorColor	"UseCursorColor"
@@ -463,6 +489,7 @@
 #define ClsVisualSelectColor	"VisualSelectColor"
 #define ClsWaitCursor		"WaitCursor"
 #define ClsWerase		"Werase"
+#define ClsWrongTerminalName	"WrongTerminalName"
 #define ClsXQuartzHack		"XQuartzHack"
 
 /* Options. */
@@ -489,7 +516,6 @@
 #define OptDefaultFgBg		"-defaultfgbg"
 #define OptDefScreen		"-defscreen"
 #define OptDevName		"-devname"
-#define OptTrace		"-trace"
 #define OptEmulatorFont		"-efont"
 #define OptHostsFile		"-hostsfile"
 #define OptHelp1		"--help"
@@ -499,7 +525,9 @@
 #define OptIconName		"-iconname"
 #define OptIconX		"-iconx"
 #define OptIconY		"-icony"
+#define OptIndent		"-indent"
 #define OptInputMethod		"-im"
+#define OptJson			"-json"
 #define OptKeyFile		"-keyfile"
 #define OptKeyFileType		"-keyfiletype"
 #define OptKeymap		"-keymap"
@@ -513,11 +541,14 @@
 #define OptMono			"-mono"
 #define OptNoPrompt		"-noprompt"
 #define OptNoScrollBar		"+sb"
+#define OptNoWrapperDoc		"-nowrapperdoc"
 #define OptNvtMode		"-nvt"
 #define OptOnce			"-once"
 #define OptOversize		"-oversize"
 #define OptPort			"-port"
 #define OptPreeditType		"-pt"
+#define OptPreferIpv4		"-4"
+#define OptPreferIpv6		"-6"
 #define OptPrinterLu		"-printerlu"
 #define OptProxy		"-proxy"
 #define OptReconnect		"-reconnect"
@@ -536,6 +567,9 @@
 #define OptNoVerifyHostCert	"-noverifycert"
 #define OptTermName		"-tn"
 #define OptTitle		"-title"
+#define OptTlsMaxProtocol	"-tlsmaxprotocol"
+#define OptTlsMinProtocol	"-tlsminprotocol"
+#define OptTrace		"-trace"
 #define OptTraceFile		"-tracefile"
 #define OptTraceFileSize	"-tracefilesize"
 #define OptUser			"-user"
@@ -543,6 +577,8 @@
 #define OptV			"-v"
 #define OptVerifyHostCert	"-verifycert"
 #define OptVersion		"--version"
+#define OptXml			"-xml"
+#define OptXrm			"-xrm"
 
 /* Miscellaneous values. */
 #define ResTrue			"true"
@@ -573,3 +609,6 @@
 #define ResPrintTextVerticalMargin "printTextVerticalMargin"
 #define ResPrintWindowCommand	"printWindowCommand"
 #define ResUser			"user"
+
+/* Handy macro. */
+#define TrueFalse(b)	((b)? ResTrue: ResFalse)
