@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2009, 2013-2020 Paul Mattes.
+ * Copyright (c) 1993-2023 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta,
  *  GA 30332.
@@ -56,11 +56,18 @@
 
 #include "globals.h"
 
+#include <errno.h>
+#include <assert.h>
+#include <fcntl.h>
+#include <signal.h>
 #include <sys/types.h>
+#include <sys/socket.h>
 #include <sys/wait.h>
+#include <netinet/in.h>
 
 #include "names.h"
 #include "s3270_proto.h"
+#include "utils.h"
 
 #if TCL_MAJOR_VERSION > 8 || TCL_MINOR_VERSION >= 6 /*[*/
 # define NEED_PTHREADS 1
@@ -68,6 +75,8 @@
 #endif /*]*/
 
 #define IBS	4096
+#define INTERNAL_ERROR "Internal error"
+
 /*
  * The following variable is a special hack that is needed in order for
  * Sun shared libraries to be used for Tcl.
@@ -78,18 +87,16 @@ extern int matherr();
 int *tclDummyMathPtr = (int *) matherr;
 #endif /*]*/
 
-static int s3270pipe[2];
+static int s3270_socket[2] = { -1, -1 };
 static bool verbose = false;
 static bool interactive = false;
 static pid_t s3270_pid;
-static bool s3270_exited = false;
-static char s3270_errmsg[1024];
 #if defined(NEED_PTHREADS) /*[*/
 static pthread_mutex_t cmd_mutex;
 #endif /*]*/
 
 static Tcl_ObjCmdProc x3270_cmd;
-static Tcl_ObjCmdProc Rows_cmd, Cols_cmd, Status_cmd;
+static Tcl_ObjCmdProc Rows_cmd, Cols_cmd, Status_cmd, Quit_cmd;
 static int tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[]);
 
 /*
@@ -116,12 +123,13 @@ main(int argc, char **argv)
     if (programname) {
 	char *path = getenv("PATH");
 	size_t path_len = programname - argv[0];
-	size_t buf_len = 5 + path_len + (path? strlen(path + 1): 0) + 1;
-	char *buf = Malloc(buf_len);
+	int ns;
+	char *buf;
 
 	/* Add our path to $PATH so we can find s3270 and x3270if. */
-	snprintf(buf, buf_len, "PATH=%.*s%s%s", (int)path_len, argv[0],
+	ns = asprintf(&buf, "PATH=%.*s%s%s", (int)path_len, argv[0],
 		path? ":": "", path? path: "");
+	assert(ns >= 0);
 	putenv(buf);
     }
 
@@ -200,6 +208,7 @@ Tcl_AppInit(Tcl_Interp *interp)
     Tcl_SetVar2Ex(interp, "argv", NULL, argv_obj, 0);
     sprintf(nbuf, "%d", i? i - 1 : 0);
     Tcl_SetVar(interp, "argc", nbuf, 0);
+    Replace(argv, NULL);
 
     /*
      * Call the init procedures for included packages.  Each call should
@@ -231,9 +240,8 @@ usage(const char *msg)
     if (msg != NULL) {
 	fprintf(stderr, "%s\n", msg);
     }
-    fprintf(stderr, "Usage:\n");
-    fprintf(stderr, "  tcl3270 [single-option]\n");
-    fprintf(stderr, "  tcl3270 [script [script-args]] [-- [tcl3270-options] [s3270-options] [host|session-file.tcl3270]]\n");
+    fprintf(stderr, "Usage: tcl3270 [single-option]\n");
+    fprintf(stderr, "       tcl3270 [script [script-args]] [-- [tcl3270-options] [s3270-options] [<host>|<session-file>.tcl3270]]\n");
     fprintf(stderr, "single-options:\n");
     fprintf(stderr, "  --help      display usage\n");
     fprintf(stderr, "  -v          display version\n");
@@ -242,15 +250,50 @@ usage(const char *msg)
     fprintf(stderr, "tcl3270-options:\n");
     fprintf(stderr, "  -d          debug s3270 I/O\n");
     fprintf(stderr, "s3270-options:\n");
-    system("s3270 --help 2>&1 | tail -n +3 - >&2");
-    exit(1);
+    (void) system("s3270 --help 2>&1 | tail -n +4 - >&2");
+    exit(99);
+}
+
+/*
+ * Poll s3270 for exit status. Returns a non-NULL error message if s3270
+ * exited, NULL otherwise.
+ * Does not block.
+ */
+static const char *
+poll_s3270_exit(void)
+{
+    static bool s3270_exited = false;
+    static char *s3270_exit_msg = NULL;
+    int st;
+
+    if (s3270_exited) {
+	return s3270_exit_msg;
+    }
+    if (waitpid(s3270_pid, &st, WNOHANG) > 0) {
+	int ns = -1;
+
+	s3270_exited = true;
+	if (WIFEXITED(st)) {
+	    ns = asprintf(&s3270_exit_msg,
+		    "s3270 backend exited with status %d", WEXITSTATUS(st));
+	} else if (WIFSIGNALED(st)) {
+	    ns = asprintf(&s3270_exit_msg,
+		    "s3270 backend killed by signal %d", WTERMSIG(st));
+	} else {
+	    ns = asprintf(&s3270_exit_msg,
+		    "Unknown s3270 backend exit status %d", st);
+	}
+	assert(ns >= 0);
+	return s3270_exit_msg;
+    }
+    return NULL;
 }
 
 /* Do a single command, and interpret the results. */
 static int
 run_s3270(const char *cmd, bool *success, char **status, char **ret)
 {
-    int st;
+    const char *exit_msg;
     int nw = 0;
     char buf[IBS];
     char rbuf[IBS];
@@ -273,28 +316,8 @@ run_s3270(const char *cmd, bool *success, char **status, char **ret)
 #endif /*]*/
 
     /* Check s3270. */
-    if (s3270_exited) {
-	*ret = NewString(s3270_errmsg);
-	rv = 0;
-	goto done;
-    }
-    if (waitpid(s3270_pid, &st, WNOHANG) > 0) {
-	s3270_exited = true;
-	if (WIFEXITED(st)) {
-	    if (WEXITSTATUS(st) == 0) {
-		exit(0);
-	    }
-	    snprintf(s3270_errmsg, sizeof(s3270_errmsg),
-		    "s3270 exited with status %d", WEXITSTATUS(st));
-	} else if (WIFSIGNALED(st)) {
-	    snprintf(s3270_errmsg, sizeof(s3270_errmsg),
-		    "s3270 killed by signal %d", WTERMSIG(st));
-	} else {
-	    snprintf(s3270_errmsg, sizeof(s3270_errmsg),
-		    "Unknown s3270 exit status %d", st);
-	}
-	*ret = NewString(s3270_errmsg);
-	rv = 0;
+    if ((exit_msg = poll_s3270_exit()) != NULL) {
+	*ret = NewString(exit_msg);
 	goto done;
     }
 
@@ -306,16 +329,18 @@ run_s3270(const char *cmd, bool *success, char **status, char **ret)
     cmd_nl = Malloc(strlen(cmd) + 2);
     sprintf(cmd_nl, "%s\n", cmd);
 
-    nw = write(s3270pipe[1], cmd_nl, strlen(cmd_nl));
+    nw = write(s3270_socket[0], cmd_nl, strlen(cmd_nl));
     if (nw < 0) {
-	perror("s3270 (back end): write");
+	int as = asprintf(ret, "s3270 backend write: %s", strerror(errno));
+
+	assert(as >= 0);
 	Free(cmd_nl);
 	goto done;
     }
     Free(cmd_nl);
 
     /* Get the answer. */
-    while (!complete && (nr = read(s3270pipe[0], rbuf, IBS)) > 0) {
+    while (!complete && (nr = read(s3270_socket[0], rbuf, IBS)) > 0) {
 	ssize_t i;
 	bool get_more = false;
 
@@ -368,22 +393,24 @@ run_s3270(const char *cmd, bool *success, char **status, char **ret)
 	    continue;
 	}
     }
-    if (nr < 0) {
-	perror("s3270 (back end) read");
-	if (status != NULL && *status != NULL) {
-	    Free(*status);
-	    *status = NULL;
-	}
+    if (nr <= 0) {
+	int as;
+
 	if (*ret != NULL) {
 	    Free(*ret);
 	    *ret = NULL;
 	}
-	goto done;
-    } else if (nr == 0) {
-	if (verbose) {
-	    fprintf(stderr, "s3270 EOF\n");
+	if (nr < 0) {
+	    as = asprintf(ret, "s3270 backend read: %s", strerror(errno));
+	    assert(as >= 0);
+	} else {
+	    *ret = NewString("s3270 backend exited");
 	}
-	exit(0);
+	if (status != NULL && *status != NULL) {
+	    Free(*status);
+	    *status = NULL;
+	}
+	goto done;
     }
 
     /* Make sure we return someting. */
@@ -404,19 +431,125 @@ done:
     return rv;
 }
 
+/* Wait for callbacks from s3270. */
+static void
+wait_for_callbacks(int listen_s, int sockets[], int n)
+{
+    time_t start;
+    int n_sockets = 0;
+
+    /* For five seconds, wait for s3270 to connect back to us twice or exit. */
+    start = time(NULL);
+    while (time(NULL) - start < 5 && n_sockets < n) {
+	fd_set rfds;
+	struct timeval tv;
+	int ns;
+	const char *exit_msg;
+
+	FD_ZERO(&rfds);
+	FD_SET(listen_s, &rfds);
+	tv.tv_sec = 0;
+	tv.tv_usec = 100000; /* 0.1 seconds */
+
+	ns = select(listen_s + 1, &rfds, NULL, NULL, &tv);
+	if (ns < 0) {
+	    perror("select");
+	    exit(99);
+	}
+
+	/* Check s3270. */
+	if ((exit_msg = poll_s3270_exit()) != NULL) {
+	    fprintf(stderr, "%s\n", exit_msg);
+	    exit(99);
+	}
+
+	if (ns > 0 && FD_ISSET(listen_s, &rfds)) {
+	    int s;
+	    struct sockaddr_in sin;
+	    socklen_t len;
+
+	    memset(&sin, 0, sizeof(sin));
+	    sin.sin_family = AF_INET;
+	    len = sizeof(sin);
+	    s = accept(listen_s, (struct sockaddr *)&sin, &len);
+	    if (s < 0) {
+		perror("accept");
+		exit(99);
+	    }
+	    if (verbose) {
+		fprintf(stderr, "Got connection %d from s3270\n",
+			n_sockets + 1);
+	    }
+	    sockets[n_sockets++] = s;
+	}
+    }
+
+    if (n_sockets < 2) {
+	fprintf(stderr, "s3270 backend did not start\n");
+	kill(s3270_pid, SIGKILL);
+	exit(99);
+    }
+}
+
+#if defined(NEED_PTHREADS) /*[*/
+/* Watch for s3270 exit. */
+static void *
+watch_s3270(void *arg)
+{
+    sigset_t set;
+    char buf;
+    int s = *(int *)arg;
+
+    /* No signals, please. */
+    sigfillset(&set);
+    pthread_sigmask(SIG_BLOCK, &set, NULL);
+
+    /* Wait until s3270 closes the second callback socket. */
+    (void) read(s, &buf, 1);
+
+    if (verbose) {
+	/*
+	 * Poll for up to 2 seconds for s3270 to finish exiting, so we can
+	 * display its exit status.
+	 */
+	time_t start = time(NULL);
+	const char *exit_msg = "s3270 backend exited";
+
+	while (time(NULL) - start < 2 &&
+		(exit_msg = poll_s3270_exit()) == NULL) {
+	    struct timespec ts;
+
+	    ts.tv_sec = 0;
+	    ts.tv_nsec = 100000000; /* 10ms */
+	    nanosleep(&ts, NULL);
+	}
+	fprintf(stderr, "%s\n", exit_msg);
+    }
+    exit(98);
+    return NULL;
+}
+#endif /*]*/
+
 /* Initialization procedure for tcl3270. */
 static int
 tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[])
 {
-    char **nargv = Calloc(argc + 7, sizeof(char *));
+    char **nargv = Calloc(argc + 9, sizeof(char *));
     int i_in, i_out = 0;
-    int to_s3270_pipe[2];
-    int from_s3270_pipe[2];
+    int listen_s;
+    struct sockaddr_in sin;
+    socklen_t len;
+    u_short port;
+    int devnull;
     bool success;
     char *ret;
     char *action;
     char *paren;
     int skip_ix = -1;
+    char cbuf[32];
+#if defined(NEED_PTHREADS) /*[*/
+    pthread_t s3270_watcher;
+#endif /*]*/
 
     /*
      * Handle special first arguments first, which completely violate the
@@ -426,6 +559,11 @@ tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[])
     if (argc > 1) {
 	if (!strcmp(argv[1], "-v") || !strcmp(argv[1], "--version")) {
 	    fprintf(stderr, "%s\n", build);
+	    fprintf(stderr,
+"Copyright 1989-%s, Paul Mattes, GTRC and others.\n"
+"See the source code or documentation for licensing details.\n"
+"Distributed WITHOUT ANY WARRANTY; without even the implied warranty of\n"
+"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n", cyear);
 	    exit(0);
 	}
 	if (!strcmp(argv[1], "--help") || !strcmp(argv[1], "-?")) {
@@ -458,13 +596,40 @@ tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[])
 	verbose = true;
     }
 
+    /* Set up a listening socket. */
+    listen_s = socket(AF_INET, SOCK_STREAM, 0);
+    if (listen_s < 0) {
+	perror("socket");
+	exit(99);
+    }
+    memset(&sin, 0, sizeof(sin));
+    sin.sin_family = AF_INET;
+    sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+    if (bind(listen_s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
+	perror("bind");
+	exit(99);
+    }
+    len = sizeof(sin);
+    if (getsockname(listen_s, (struct sockaddr *)&sin, &len) < 0) {
+	perror("getsockname");
+	exit(99);
+    }
+    port = htons(sin.sin_port);
+    snprintf(cbuf, sizeof(cbuf), "2x127.0.0.1:%u", port);
+    if (listen(listen_s, 1) < 0) {
+	perror("listen");
+	exit(99);
+    }
+
     /* Set up s3270's command-line arguments. */
     nargv[i_out++] = "s3270";
     nargv[i_out++] = "-utf8";
     nargv[i_out++] = "-minversion";
-    nargv[i_out++] = "4.0";
+    nargv[i_out++] = "4.1";
     nargv[i_out++] = "-alias";
     nargv[i_out++] = "tcl3270";
+    nargv[i_out++] = "-callback";
+    nargv[i_out++] = cbuf;
     if (skip_ix >= 0) {
 	for (i_in = skip_ix + 1; i_in < argc; i_in++) {
 	    nargv[i_out++] = (char *)argv[i_in];
@@ -472,12 +637,6 @@ tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[])
     }
     nargv[i_out++] = NULL;
 
-    /* Set up pipes. */
-    if (pipe(to_s3270_pipe) < 0 || pipe(from_s3270_pipe) < 0) {
-	perror("pipe");
-	return TCL_ERROR;
-    }
-
     /* Start s3270. */
     switch (s3270_pid = fork()) {
     case -1:
@@ -487,38 +646,53 @@ tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[])
 	/* Child. */
 
 	/* Redirect I/O. */
-	close(to_s3270_pipe[1]);
-	if (dup2(to_s3270_pipe[0], 0) < 0) {
-	    perror("dup2");
-	    exit(1);
+	close(listen_s);
+	devnull = open("/dev/null", O_RDWR);
+	if (devnull < 0) {
+	    perror("/dev/null");
+	    exit(99);
 	}
-	close(to_s3270_pipe[0]);
-	if (dup2(from_s3270_pipe[1], 1) < 0) {
+	if (dup2(devnull, 0) < 0) {
 	    perror("dup2");
-	    exit(1);
+	    exit(99);
 	}
-	close(from_s3270_pipe[1]);
+	if (dup2(devnull, 1) < 0) {
+	    perror("dup2");
+	    exit(99);
+	}
+	close(devnull);
 
 	/* Run s3270. */
 	if (execvp("s3270", nargv) < 0) {
-	    perror("s3270 (back end)");
-	    exit(1);
+	    perror("s3270 backend");
+	    exit(99);
 	}
 	break;
     default:
 	/* Parent. */
+	Free(nargv);
 	break;
     }
 
-    /* Redirect I/O. */
-    close(to_s3270_pipe[0]);
-    close(from_s3270_pipe[1]);
-    s3270pipe[0] = from_s3270_pipe[0];
-    s3270pipe[1] = to_s3270_pipe[1];
+    /* Wait for the connections back from s3270. */
+    if (verbose) {
+	fprintf(stderr, "Listening for connections from s3270 on %s\n",
+		cbuf + 2);
+    }
+
+    wait_for_callbacks(listen_s, s3270_socket, 2);
+    close(listen_s);
 
 #if defined(NEED_PTHREADS) /*[*/
     /* Set up the mutex. */
     pthread_mutex_init(&cmd_mutex, NULL);
+
+    /* Set up the watcher thread. */
+    if (pthread_create(&s3270_watcher, NULL, watch_s3270,
+		&s3270_socket[1]) < 0) {
+	perror("pthread_create");
+	exit(99);
+    }
 #endif /*]*/
 
     /* Run 'Query(Actions)' to learn what Tcl commands we need to add. */
@@ -534,11 +708,13 @@ tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[])
     action = ret;
     while ((paren = strchr(action, '(')) != NULL) {
 
-	/* Create the command. */
+	/* Create the command, except for Quit(), which we override. */
 	*paren = '\0';
-	if (Tcl_CreateObjCommand(interp, action, x3270_cmd, NULL, NULL)
-		== NULL) {
-	    return TCL_ERROR;
+	if (strcasecmp(action, AnQuit) && strcasecmp(action, AnExit)) {
+	    if (Tcl_CreateObjCommand(interp, action, x3270_cmd, NULL, NULL)
+		    == NULL) {
+		return TCL_ERROR;
+	    }
 	}
 
 	/* Skip to the next action. */
@@ -565,6 +741,12 @@ tcl3270_main(Tcl_Interp *interp, int argc, const char *argv[])
 	    == NULL) {
 	return TCL_ERROR;
     }
+    if (Tcl_CreateObjCommand(interp, AnQuit, Quit_cmd, NULL, NULL) == NULL) {
+	return TCL_ERROR;
+    }
+    if (Tcl_CreateObjCommand(interp, AnExit, Quit_cmd, NULL, NULL) == NULL) {
+	return TCL_ERROR;
+    }
 
     return TCL_OK;
 }
@@ -686,7 +868,16 @@ x3270_cmd(ClientData clientData, Tcl_Interp *interp, int objc,
     rv = run_s3270(cmd, &success, NULL, &ret);
     if (rv < 0) {
 	Free(cmd);
-	Tcl_SetResult(interp, "Internal error", TCL_STATIC);
+	if (ret == NULL) {
+	    Tcl_SetResult(interp, INTERNAL_ERROR, TCL_STATIC);
+	} else {
+	    char *result;
+	    int as = asprintf(&result, INTERNAL_ERROR ": %s", ret);
+
+	    assert(as >= 0);
+	    Free(ret);
+	    Tcl_SetResult(interp, result, TCL_VOLATILE);
+	}
 	return TCL_ERROR;
     }
 
@@ -728,7 +919,16 @@ Status_cmd(ClientData clientData, Tcl_Interp *interp, int objc,
     char *ret;
 
     if (run_s3270("", &success, &status, &ret) < 0) {
-	Tcl_SetResult(interp, "Internal error", TCL_STATIC);
+	if (ret == NULL) {
+	    Tcl_SetResult(interp, INTERNAL_ERROR, TCL_STATIC);
+	} else {
+	    char *result;
+	    int na = asprintf(&result, INTERNAL_ERROR ": %s", ret);
+
+	    assert(na >= 0);
+	    Free(ret);
+	    Tcl_SetResult(interp, result, TCL_VOLATILE);
+	}
 	return TCL_ERROR;
     }
     Tcl_SetResult(interp, status, TCL_VOLATILE);
@@ -737,6 +937,23 @@ Status_cmd(ClientData clientData, Tcl_Interp *interp, int objc,
     return TCL_OK;
 }
 
+/*
+ * Exit the emulator.
+ * This overrides the x3270 Quit() action, which will make s3270 exit
+ * asynchronously.
+ */
+static int
+Quit_cmd(ClientData clientData, Tcl_Interp *interp, int objc,
+	Tcl_Obj *CONST objv[])
+{
+    if (objc > 1) {
+	Tcl_SetResult(interp, AnQuit "() takes 0 arguments", TCL_STATIC);
+	return TCL_ERROR;
+    }
+    Tcl_Exit(0);
+    return TCL_OK;
+}
+
 /**
  * Isolate a field within the status line.
  *
@@ -778,7 +995,16 @@ Rows_cmd(ClientData clientData, Tcl_Interp *interp, int objc,
     char *f;
 
     if (run_s3270("", &success, &status, &ret) < 0) {
-	Tcl_SetResult(interp, "Internal error", TCL_STATIC);
+	if (ret == NULL) {
+	    Tcl_SetResult(interp, INTERNAL_ERROR, TCL_STATIC);
+	} else {
+	    char *result;
+	    int na = asprintf(&result, INTERNAL_ERROR ": %s", ret);
+
+	    assert(na >= 0);
+	    Free(ret);
+	    Tcl_SetResult(interp, result, TCL_VOLATILE);
+	}
 	return TCL_ERROR;
     }
     Free(ret);
@@ -800,7 +1026,16 @@ Cols_cmd(ClientData clientData, Tcl_Interp *interp, int objc,
     char *f;
 
     if (run_s3270("", &success, &status, &ret) < 0) {
-	Tcl_SetResult(interp, "Internal error", TCL_STATIC);
+	if (ret == NULL) {
+	    Tcl_SetResult(interp, INTERNAL_ERROR, TCL_STATIC);
+	} else {
+	    char *result;
+	    int na = asprintf(&result, INTERNAL_ERROR ": %s", ret);
+
+	    assert(na >= 0);
+	    Free(ret);
+	    Tcl_SetResult(interp, result, TCL_VOLATILE);
+	}
 	return TCL_ERROR;
     }
     Free(ret);
@@ -816,5 +1051,5 @@ void
 Error(const char *msg)
 {
     fprintf(stderr, "%s\n", msg);
-    exit(1);
+    exit(99);
 }
