@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2021 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -87,10 +87,10 @@
 toggle_widget_t toggle_widget[N_TOGGLES];
 
 static struct scheme {
-	char *label;
-	char **parents;
-	char *scheme;
-	struct scheme *next;
+    char *label;
+    char **parents;
+    char *scheme;
+    struct scheme *next;
 } *schemes, *last_scheme;
 static int scheme_count;
 static Widget  *scheme_widgets;
@@ -104,7 +104,6 @@ static Dimension fm_leftMargin;
 static Dimension fm_rightMargin;
 static bool snap_enabled = true;
 static bool keypad_sensitive = true;
-
 static struct codepage {
     char **parents;
     char *label;
@@ -113,6 +112,10 @@ static struct codepage {
 } *codepages, *last_codepage;
 static int codepage_count;
 static Widget  *codepage_widgets;
+static struct host_list {
+    char *name;
+    struct host_list *next;
+} *host_list;
 
 static void scheme_init(void);
 static void codepages_init(void);
@@ -189,6 +192,8 @@ static Widget   locked_icon;
 static Widget   unlocked_icon;
 static Widget   unverified_icon;
 static Widget   keypad_button;
+static Widget	retry_button;
+static Widget	reconnect_button;
 static Widget   linemode_button;
 static Widget   charmode_button;
 static Widget   models_option;
@@ -272,6 +277,7 @@ static bool	toggle_init(Widget, int, const char *, const char *, bool *);
 struct menu_hier {
     Widget menu_shell;		/* complexMenu widget */
     char *name;			/* my name (root name is NULL) */
+    char *menu_name;		/* menu name */
     struct menu_hier *parent;	/* parent menu */
     struct menu_hier *child;	/* child menu */
     struct menu_hier *sibling;	/* sibling menu */
@@ -349,6 +355,7 @@ add_menu_hier(struct menu_hier *root, char **parents, ArgList args,
 	     */
 	    XtSetArg(my_arglist[0], XtNrightBitmap, arrow);
 	    XtSetArg(my_arglist[1], XtNmenuName, menu_name);
+	    h->menu_name = menu_name;
 	    merged_args = XtMergeArgLists(my_arglist, 2, args, num_args);
 	    XtCreateManagedWidget(
 		    h->name, cmeBSBObjectClass, h->parent->menu_shell,
@@ -373,6 +380,7 @@ free_menu_hier(struct menu_hier *root)
     if (root->child) {
 	free_menu_hier(root->child);
     }
+    XtFree(root->menu_name);
     XtFree((char *)root);
 }
 
@@ -570,7 +578,7 @@ menubar_init(Widget container, Dimension overall_width, Dimension current_width)
 
     /* "Connect..." menu */
 
-    if (!appres.interactive.reconnect)
+    if (!appres.reconnect)
 	connect_menu_init(mb_old != menubar_buttons,
 		BUTTON_X((file_menu != NULL) + (options_menu != NULL)),
 		TOP_MARGIN);
@@ -670,7 +678,7 @@ menubar_connect(bool ignored _is_unused)
     }
 
     /* Set up the connect menu. */
-    if (!appres.interactive.reconnect && connect_menu != NULL) {
+    if (!appres.reconnect && connect_menu != NULL) {
 	if (PCONNECTED && connect_button != NULL) {
 	    XtUnmapWidget(connect_button);
 	} else {
@@ -1205,16 +1213,18 @@ file_menu_init(bool regen, Dimension x, Dimension y)
 
     /* Save/restore input. */
     spaced = false;
-    save_input_button = add_menu_itemv("saveInputOption", file_menu,
-	    do_save_input, NULL, &spaced,
-	    XtNsensitive, IN_3270,
-	    NULL);
-    any |= (save_input_button != NULL);
-    restore_input_button = add_menu_itemv("restoreInputOption", file_menu,
-	    do_restore_input, NULL, &spaced,
-	    XtNsensitive, IN_3270,
-	    NULL);
-    any |= (restore_input_button != NULL);
+    if (!appres.secure) {
+	save_input_button = add_menu_itemv("saveInputOption", file_menu,
+		do_save_input, NULL, &spaced,
+		XtNsensitive, IN_3270,
+		NULL);
+	any |= (save_input_button != NULL);
+	restore_input_button = add_menu_itemv("restoreInputOption", file_menu,
+		do_restore_input, NULL, &spaced,
+		XtNsensitive, IN_3270,
+		NULL);
+	any |= (restore_input_button != NULL);
+    }
 
     /* Re-enable keyboard. */
     spaced = false;
@@ -1330,6 +1340,7 @@ connect_menu_init(bool regen, Position x, Position y)
     int n_hosts = 0;
     bool any_hosts = false;
     struct host *h;
+    struct host_list *hl;
     bool need_line = false;
     int n_primary = 0;
     int n_recent = 0;
@@ -1362,6 +1373,12 @@ connect_menu_init(bool regen, Position x, Position y)
 
     /* Walk the host list from the file to produce the host menu */
 
+    while (host_list != NULL) {
+	hl = host_list->next;
+	free(host_list->name);
+	free(host_list);
+	host_list = hl;
+    }
     for (h = hosts; h; h = h->next) {
 	switch (h->entry_type) {
 	case ALIAS:
@@ -1390,8 +1407,11 @@ connect_menu_init(bool regen, Position x, Position y)
 		    add_menu_hier(root, h->parents, NULL, 0):
 		recent_menu,
 		NULL);
-	XtAddCallback(w, XtNcallback, host_connect_callback,
-		XtNewString(h->name));
+	hl = (struct host_list *)XtCalloc(1, sizeof(struct host_list));
+	hl->name = XtNewString(h->name);
+	hl->next = host_list;
+	host_list = hl;
+	XtAddCallback(w, XtNcallback, host_connect_callback, hl->name);
 	n_hosts++;
     }
     if (recent_menu) {
@@ -1634,6 +1654,7 @@ tls_icon_init(Position x, Position y)
 		show_about_status, NULL);
     } else {
 	XtVaSetValues(locked_icon, XtNx, x, NULL);
+	XtVaSetValues(unverified_icon, XtNx, x, NULL);
 	XtVaSetValues(unlocked_icon, XtNx, x, NULL);
     }
 }
@@ -1791,21 +1812,21 @@ scheme_init(void)
     char *label;
     char *scheme;
     struct scheme *s;
+    size_t offset = 0;
 
     cm = get_resource(ResSchemeList);
     if (cm == NULL) {
 	return;
     }
-    cm = XtNewString(cm);
 
     scheme_count = 0;
-    while (split_dresource(&cm, &label, &scheme) == 1) {
+    while (s_split_dresource(cm, &offset, &label, &scheme) == 1) {
 	s = (struct scheme *)XtMalloc(sizeof(struct scheme));
 	if (!split_hier(label, &s->label, &s->parents)) {
 	    XtFree((XtPointer)s);
 	    continue;
 	}
-	s->label = label;
+	XtFree(label);
 	s->scheme = scheme;
 	s->next = NULL;
 	if (last_scheme != NULL) {
@@ -1829,7 +1850,7 @@ do_newscheme(Widget w _is_unused, XtPointer userdata,
 static void
 codepages_init(void)
 {
-    char *cm;
+    char *cm, *cm0;
     char *label;
     char *codepage;
     struct codepage *s;
@@ -1838,7 +1859,7 @@ codepages_init(void)
     if (cm == NULL) {
 	return;
     }
-    cm = XtNewString(cm);
+    cm = cm0 = XtNewString(cm);
 
     codepage_count = 0;
     while (split_dresource(&cm, &label, &codepage) == 1) {
@@ -1847,7 +1868,7 @@ codepages_init(void)
 	    XtFree((XtPointer)s);
 	    continue;
 	}
-	s->codepage = codepage;
+	s->codepage = XtNewString(codepage);
 	s->next = NULL;
 	if (last_codepage != NULL) {
 	    last_codepage->next = s;
@@ -1857,11 +1878,11 @@ codepages_init(void)
 	last_codepage = s;
 	codepage_count++;
     }
+    XtFree(cm0);
 }
 
 static void
-do_newcodepage(Widget w _is_unused, XtPointer userdata,
-	XtPointer calldata _is_unused)
+do_newcodepage(Widget w, XtPointer userdata, XtPointer calldata _is_unused)
 {
     /* Change the code page. */
     screen_newcodepage((char *)userdata);
@@ -1928,6 +1949,20 @@ charmode_callback(Widget w _is_unused, XtPointer client_data _is_unused,
     net_charmode();
 }
 
+static void
+toggle_retry(Widget w _is_unused, XtPointer client_data _is_unused, 
+	XtPointer call_data _is_unused)
+{
+    push_macro(AnToggle "(" ResRetry ")");
+}
+
+static void
+toggle_reconnect(Widget w _is_unused, XtPointer client_data _is_unused, 
+	XtPointer call_data _is_unused)
+{
+    push_macro(AnToggle "(" ResReconnect ")");
+}
+
 /* Called to change models */
 static void
 change_model_callback(Widget w, XtPointer client_data,
@@ -2037,8 +2072,7 @@ create_font_menu(bool regen, bool even_if_unknown)
 		XtNrightMargin, fm_rightMargin,
 		XtNbackground, fm_background,
 		NULL);
-	XtAddCallback(font_widgets[ix], XtNcallback, do_newfont,
-		XtNewString(f->font));
+	XtAddCallback(font_widgets[ix], XtNcallback, do_newfont, f->font);
     }
     if (!xappres.no_other) {
 	other_font = XtVaCreateManagedWidget(
@@ -2068,8 +2102,8 @@ menubar_codepage(bool ignored _is_unused)
 	XtVaSetValues(codepage_widgets[i],
 		XtNleftBitmap,
 		(!strcmp(cpname, s->codepage) ||
-		 codepage_matches_alias(s->codepage, cpname)) ?
-		    diamond : no_diamond,
+		 codepage_matches_alias(s->codepage, cpname))?
+		    diamond: no_diamond,
 		NULL);
     }
 }
@@ -2200,6 +2234,8 @@ options_menu_init(bool regen, Position x, Position y)
 	toggle_init(t, MONOCASE, "monocaseOption", NULL, &spaced);
 	toggle_init(t, CURSOR_BLINK, "cursorBlinkOption", NULL, &spaced);
 	toggle_init(t, BLANK_FILL, "blankFillOption", NULL, &spaced);
+	toggle_init(t, UNDERSCORE_BLANK_FILL, "underscoreBlankFillOption",
+		NULL, &spaced);
 	toggle_init(t, SHOW_TIMING, "showTimingOption", NULL, &spaced);
 	toggle_init(t, SCROLL_BAR, "scrollBarOption", NULL, &spaced);
 	toggle_init(t, LINE_WRAP, "lineWrapOption", NULL, &spaced);
@@ -2212,6 +2248,18 @@ options_menu_init(bool regen, Position x, Position y)
 	toggle_init(t, TYPEAHEAD, "typeaheadOption", NULL, &spaced);
 	toggle_init(t, ALWAYS_INSERT, "alwaysInsertOption", NULL, &spaced);
 	toggle_init(t, SELECT_URL, "selectUrlOption", NULL, &spaced);
+	retry_button = add_menu_itemv("retryOption", t,
+		    toggle_retry, NULL,
+		    &spaced,
+		    XtNleftBitmap, appres.retry? dot: (Pixmap)NULL,
+		    XtNsensitive, True,
+		    NULL);
+	reconnect_button = add_menu_itemv("reconnectOption", t,
+		    toggle_reconnect, NULL,
+		    &spaced,
+		    XtNleftBitmap, appres.reconnect? dot: (Pixmap)NULL,
+		    XtNsensitive, True,
+		    NULL);
 	spaced = false;
 	toggle_init(t, ALT_CURSOR, "underlineCursorOption",
 		"blockCursorOption", &spaced);
@@ -2263,6 +2311,7 @@ options_menu_init(bool regen, Position x, Position y)
 	    toggle_widget[VISIBLE_CONTROL].w[0] != NULL ||
 	    toggle_widget[ALT_CURSOR].w[0] != NULL ||
 	    toggle_widget[ALWAYS_INSERT].w[0] != NULL ||
+	    toggle_widget[UNDERSCORE_BLANK_FILL].w[0] != NULL ||
 	    linemode_button != NULL ||
 	    charmode_button != NULL ||
 	    m3278_button != NULL ||
@@ -2419,6 +2468,7 @@ options_menu_init(bool regen, Position x, Position y)
     /* Create the "code page" pullright */
     if (codepage_count && !item_suppressed(options_menu, "codepageOption")) {
 	struct codepage *cs;
+	const char *cpname;
 
 	if (codepage_root != NULL) {
 	    free_menu_hier(codepage_root);
@@ -2429,14 +2479,16 @@ options_menu_init(bool regen, Position x, Position y)
 		"codepageMenu", complexMenuWidgetClass, menu_parent,
 		NULL);
 
+	cpname = get_codepage_name();
 	codepage_widgets = (Widget *)XtCalloc(codepage_count, sizeof(Widget));
 	for (ix = 0, cs = codepages; ix < codepage_count; ix++, cs = cs->next) {
 	    t = add_menu_hier(codepage_root, cs->parents, NULL, 0);
 	    codepage_widgets[ix] = XtVaCreateManagedWidget(
 		    cs->label, cmeBSBObjectClass, t,
 		    XtNleftBitmap,
-			(strcmp(get_codepage_name(), cs->codepage))? no_diamond:
-								   diamond,
+		    (!strcmp(cpname, cs->codepage) ||
+		     codepage_matches_alias(cs->codepage, cpname))?
+			diamond: no_diamond,
 		    NULL);
 	    XtAddCallback(codepage_widgets[ix], XtNcallback, do_newcodepage,
 		    cs->codepage);
@@ -2581,6 +2633,27 @@ screensave_option(Widget w _is_unused, XtPointer client_data _is_unused,
     stmenu_popup(STMP_AS_IS);
 }
 
+/* Extended toggle notification. */
+static void
+menubar_toggle_notify(const char *name, enum resource_type type, void **value, ia_t ia, unsigned flags)
+{
+    if (!strcasecmp(name, ResRetry)) {
+	if (retry_button != NULL) {
+	    XtVaSetValues(retry_button,
+		    XtNleftBitmap,
+		    *(bool *)value ? dot : None,
+		    NULL);
+	}
+    } else if (!strcasecmp(name, ResReconnect)) {
+	if (reconnect_button != NULL) {
+	    XtVaSetValues(reconnect_button,
+		    XtNleftBitmap,
+		    *(bool *)value ? dot : None,
+		    NULL);
+	}
+    }
+}
+
 /**
  * Menu module registration.
  */
@@ -2596,5 +2669,8 @@ menubar_register(void)
     register_schange(ST_CODEPAGE, menubar_codepage);
     register_schange(ST_KBD_DISABLE, menubar_keyboard_disable);
     register_schange(ST_SECURE, menubar_secure);
+
+    /* Register for extended toggle notifications. */
+    register_extended_toggle_notify(menubar_toggle_notify);
 }
 
