@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2009, 2013-2016, 2019-2020 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -51,6 +51,7 @@
 #include "names.h"
 #include "popups.h" /* must come before child_popups.h */
 #include "child_popups.h"
+#include "resources.h"
 #include "screen.h"
 #include "task.h"
 #include "trace.h"
@@ -75,6 +76,12 @@ static Dimension wm_width, wm_height;
 
 static ioid_t info_id = NULL_IOID;
 
+static struct {		/* Error pop-up delay: */
+    bool active;	/*  Is it active? */
+    char *text;		/*  Saved text */
+    pae_t type;		/*  Error type */
+} epd = { true, NULL, ET_OTHER };
+
 /*
  * General popup support
  */
@@ -206,11 +213,34 @@ enum placement *RightP = &RightD;
 static enum placement InsideRightD = InsideRight;
 enum placement *InsideRightP = &InsideRightD;
 
-typedef struct {
+typedef struct want {
+    struct want *next;
     Widget w;
     Position x, y;
     enum placement p;
+    XtIntervalId timeout_id;
 } want_t;
+static want_t *wants;
+
+/* Dequeue a pending move_again operation and free its context. */
+static void
+dequeue_and_free_want(want_t *wx)
+{
+    want_t *wp, *wp_prev = NULL;
+
+    for (wp = wants; wp != NULL; wp = wp->next) {
+	if (wp == wx) {
+	    if (wp_prev != NULL) {
+		wp_prev->next = wp->next;
+	    } else {
+		wants = wp->next;
+	    }
+	    break;
+	}
+	wp_prev = wp;
+    }
+    XtFree((XtPointer)wx);
+}
 
 static void
 popup_move_again(XtPointer closure, XtIntervalId *id _is_unused)
@@ -273,7 +303,9 @@ popup_move_again(XtPointer closure, XtIntervalId *id _is_unused)
 #endif /*]*/
 	XtVaSetValues(wx->w, XtNx, x, XtNy, y, NULL);
     }
-    XtFree((XtPointer)wx);
+
+    /* Dequeue and free. */
+    dequeue_and_free_want(wx);
 }
 
 /* Place a newly popped-up shell */
@@ -368,7 +400,10 @@ place_popup(Widget w, XtPointer client_data, XtPointer call_data _is_unused)
 	    wx->x = x;
 	    wx->y = y;
 	    wx->p = p;
-	    XtAppAddTimeOut(appcontext, 250, popup_move_again, (XtPointer)wx);
+	    wx->timeout_id = XtAppAddTimeOut(appcontext, 250, popup_move_again,
+		    (XtPointer)wx);
+	    wx->next = wants;
+	    wants = wx;
 	    break;
 	default:
 	case WMT_SIMPLE:
@@ -391,7 +426,10 @@ place_popup(Widget w, XtPointer client_data, XtPointer call_data _is_unused)
 	    wx->x = x;
 	    wx->y = y;
 	    wx->p = p;
-	    XtAppAddTimeOut(appcontext, 250, popup_move_again, (XtPointer)wx);
+	    wx->timeout_id = XtAppAddTimeOut(appcontext, 250, popup_move_again,
+		    (XtPointer)wx);
+	    wx->next = wants;
+	    wants = wx;
 	    break;
 	default:
 	case WMT_SIMPLE:
@@ -423,7 +461,10 @@ place_popup(Widget w, XtPointer client_data, XtPointer call_data _is_unused)
 	    wx->x = x;
 	    wx->y = y;
 	    wx->p = p;
-	    XtAppAddTimeOut(appcontext, 250, popup_move_again, (XtPointer)wx);
+	    wx->timeout_id = XtAppAddTimeOut(appcontext, 250, popup_move_again,
+		    (XtPointer)wx);
+	    wx->next = wants;
+	    wants = wx;
 	    break;
 	default:
 	case WMT_SIMPLE:
@@ -453,7 +494,10 @@ place_popup(Widget w, XtPointer client_data, XtPointer call_data _is_unused)
 	    wx->x = x;
 	    wx->y = y;
 	    wx->p = p;
-	    XtAppAddTimeOut(appcontext, 250, popup_move_again, (XtPointer)wx);
+	    wx->timeout_id = XtAppAddTimeOut(appcontext, 250, popup_move_again,
+		    (XtPointer)wx);
+	    wx->next = wants;
+	    wants = wx;
 	    break;
 	default:
 	case WMT_SIMPLE:
@@ -469,6 +513,22 @@ place_popup(Widget w, XtPointer client_data, XtPointer call_data _is_unused)
     }
 }
 
+/* Cancel a pending pop-up placement. */
+void
+unplace_popup(Widget w)
+{
+    want_t *wx;
+
+    /* Cancel any pending move_again activity. */
+    for (wx = wants; wx != NULL; wx = wx->next) {
+	if (wx->w == w) {
+	    XtRemoveTimeOut(wx->timeout_id);
+	    dequeue_and_free_want(wx);
+	    return;
+	}
+    }
+}
+
 #if defined(POPUP_DEBUG) /*[*/
 static void
 dump_windows(char *what, Widget w)
@@ -850,8 +910,8 @@ struct rop {
 };
 
 static struct rop error_popup = {
-    "errorPopup", XtGrabExclusive, true, false,
-    "first line\nsecond line\nthird line",
+    "errorPopup", XtGrabExclusive, true, true,
+    "first line\nsecond line\nthird line\nfourth line",
     NULL, NULL, NULL, NULL,
     false, false, NULL
 };
@@ -1057,9 +1117,11 @@ popup_rop(struct rop *rop, abort_callback_t *a, const char *fmt, va_list args)
 }
 
 static void
-error_exit(void)
+stop_trying(void)
 {
-    x3270_exit(0);
+    push_macro(AnSet "(" ResReconnect "=" ResFalse ","
+	    ResRetry "=" ResFalse ")");
+    popdown_an_error();
 }
 
 /* Pop up an error dialog. */
@@ -1068,11 +1130,27 @@ popup_a_vxerror(pae_t type, const char *fmt, va_list args)
 {
     char *s = NULL;
 
-    if (type == ET_CONNECT) {
-	s = xs_buffer("Connection failed:\n%s", fmt);
+    if (task_redirect()) {
+	char *buf = xs_vbuffer(fmt, args);
+
+	task_error(buf);
+	Free(buf);
+	return;
     }
 
-    popup_rop(&error_popup, appres.interactive.reconnect? error_exit: NULL,
+    if (epd.active) {
+	epd.type = type;
+	Replace(epd.text, xs_vbuffer(fmt, args));
+	return;
+    }
+
+    if (type == ET_CONNECT) {
+	s = xs_buffer("Connection failed%s:\n%s",
+		host_retry_mode? ", retrying": "", fmt);
+    }
+
+    popup_rop(&error_popup,
+	    (host_retry_mode && !appres.secure)? stop_trying: NULL,
 	    (s != NULL)? s: fmt, args);
     if (s != NULL) {
 	Free(s);
@@ -1083,11 +1161,26 @@ popup_a_vxerror(pae_t type, const char *fmt, va_list args)
 void
 popdown_an_error(void)
 {
+    if (epd.active && epd.text != NULL) {
+	Replace(epd.text, NULL);
+	return;
+    }
     if (error_popup.visible) {
 	XtPopdown(error_popup.shell);
     }
 }
 
+/* Error popup delay completion. */
+void
+error_popup_resume(void)
+{
+    epd.active = false;
+    if (epd.text != NULL) {
+	popup_an_xerror(epd.type, "%s", epd.text);
+	Replace(epd.text, NULL);
+    }
+}
+
 /* Pop up an info dialog. */
 void
 popup_an_info(const char *fmt, ...)
@@ -1221,7 +1314,7 @@ child_popup_init(void)
 bool
 error_popup_visible(void)
 {
-    return error_popup.visible;
+    return (epd.active && epd.text != NULL) || error_popup.visible;
 }
 
 /*
