@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2017, 2019-2020 Paul Mattes.
+ * Copyright (c) 1993-2023 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta,
  *  GA 30332.
@@ -162,6 +162,10 @@ XtResource resources[] = {
       offset(tls.key_file_type), XtRString, 0 },
     { ResKeyPasswd, ClsKeyPasswd, XtRString, sizeof(char *),
       offset(tls.key_passwd), XtRString, 0 },
+    { ResTlsMaxProtocol, ClsTlsMaxProtocol, XtRString, sizeof(char *),
+      offset(tls.max_protocol), XtRString, 0 },
+    { ResTlsMinProtocol, ClsTlsMinProtocol, XtRString, sizeof(char *),
+      offset(tls.min_protocol), XtRString, 0 },
 
     { ResFtAllocation, ClsFtAllocation, XtRString, sizeof(char *),
       offset(ft.allocation), XtRString, 0 },
@@ -187,6 +191,8 @@ XtResource resources[] = {
       offset(ft.lrecl), XtRString, "0" },
     { ResFtMode, ClsFtMode, XtRString, sizeof(char *),
       offset(ft.mode), XtRString, 0 },
+    { ResFtOtherOptions, ClsFtOtherOptions, XtRString, sizeof(char *),
+      offset(ft.other_options), XtRString, 0 },
     { ResFtPrimarySpace, ClsFtPrimarySpace, XtRInt, sizeof(int),
       offset(ft.primary_space), XtRString, "0" },
     { ResFtRecfm, ClsFtRecfm, XtRString, sizeof(char *),
@@ -295,10 +301,18 @@ XtResource xresources[] = {
       boffset(once), XtRString, ResFalse },
     { ResScripted, ClsScripted, XtRBoolean, sizeof(Boolean),
       boffset(scripted), XtRString, ResFalse },
+    { ResScriptedAlways, ClsScriptedAlways, XtRBoolean, sizeof(Boolean),
+      boffset(scripted_always), XtRString, ResFalse },
+    { ResPreferIpv4, ClsPreferIpv4, XtRBoolean, sizeof(Boolean),
+      boffset(prefer_ipv4), XtRString, ResFalse },
+    { ResPreferIpv6, ClsPreferIpv6, XtRBoolean, sizeof(Boolean),
+      boffset(prefer_ipv6), XtRString, ResFalse },
     { ResModifiedSel, ClsModifiedSel, XtRBoolean, sizeof(Boolean),
       boffset(modified_sel), XtRString, ResFalse },
     { ResUnlockDelay, ClsUnlockDelay, XtRBoolean, sizeof(Boolean),
       boffset(unlock_delay), XtRString, ResFalse },
+    { ResContentionResolution, ClsContentionResolution, XtRBoolean, sizeof(Boolean),
+      boffset(contention_resolution), XtRString, ResTrue },
     { ResBindLimit, ClsBindLimit, XtRBoolean, sizeof(Boolean),
       boffset(bind_limit), XtRString, ResTrue },
     { ResBindUnlock, ClsBindUnlock, XtRBoolean, sizeof(Boolean),
@@ -308,7 +322,9 @@ XtResource xresources[] = {
     { ResSocket, ClsSocket, XtRBoolean, sizeof(Boolean),
       boffset(socket), XtRString, ResFalse },
     { ResReconnect, ClsReconnect, XtRBoolean, sizeof(Boolean),
-      boffset(interactive.reconnect), XtRString, ResFalse },
+      boffset(reconnect), XtRString, ResFalse },
+    { ResRetry, ClsRetry, XtRBoolean, sizeof(Boolean),
+      boffset(retry), XtRString, ResFalse },
     { ResDoConfirms, ClsDoConfirms, XtRBoolean, sizeof(Boolean),
       boffset(interactive.do_confirms), XtRString, ResTrue },
     { ResNumericLock, ClsNumericLock, XtRBoolean, sizeof(Boolean),
@@ -335,6 +351,8 @@ XtResource xresources[] = {
       boffset(script_port_once), XtRString, ResFalse },
     { ResUtf8, ClsUtf8, XtRBoolean, sizeof(Boolean),
       boffset(utf8), XtRString, ResFalse },
+    { ResWrongTerminalName, ClsWrongTerminalName, XtRBoolean, sizeof(Boolean),
+      boffset(wrong_terminal_name), XtRString, ResFalse },
 
     { ResMonoCase, ClsMonoCase, XtRBoolean, sizeof(Boolean),
       btoffset(MONOCASE), XtRString, ResFalse },
@@ -380,6 +398,8 @@ XtResource xresources[] = {
       btoffset(INSERT_MODE), XtRString, ResFalse },
     { ResSelectUrl, ClsSelectUrl, XtRBoolean, sizeof(Boolean),
       btoffset(SELECT_URL), XtRString, ResTrue },
+    { ResUnderscoreBlankFill, ClsUnderscoreBlankFill, XtRBoolean,
+	sizeof(Boolean), btoffset(UNDERSCORE_BLANK_FILL), XtRString, ResTrue },
 
     { ResIcrnl, ClsIcrnl, XtRBoolean, sizeof(Boolean),
       boffset(linemode.icrnl), XtRString, ResTrue },
