@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2009, 2013-2015, 2018-2021 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -188,6 +188,7 @@ static void do_dbcs(void);
 static void do_scrolled(void);
 static void do_minus(void);
 static void do_disabled(void);
+static void do_file_transfer(void);
 
 static bool oia_undera = true;
 static bool oia_boxsolid = false;
@@ -219,6 +220,7 @@ static enum msg {
     SCROLLED,		/* X Scrolled */
     MINUS,		/* X -f */
     KBD_DISABLED,	/* X Disabled */
+    FILE_TRANSFER,	/* X File Transfer */
     N_MSGS
 } oia_msg = DISCONNECTED, scroll_saved_msg, disabled_saved_msg = BLANK;
 static char oia_lu[LUCNT+1];
@@ -245,7 +247,8 @@ static void (*msg_proc[N_MSGS])(void) = {
     do_dbcs,
     do_scrolled,
     do_minus,
-    do_disabled
+    do_disabled,
+    do_file_transfer,
 };
 static int msg_color[N_MSGS] = {
     FA_INT_NORM_NSEL,	/* disconnected */
@@ -269,6 +272,7 @@ static int msg_color[N_MSGS] = {
     FA_INT_NORM_NSEL,	/* scrolled */
     FA_INT_HIGH_SEL,	/* minus */
     FA_INT_HIGH_SEL,	/* disabled */
+    FA_INT_NORM_NSEL,	/* file transfer */
 };
 static int msg_color3279[N_MSGS] = {
     HOST_COLOR_WHITE,	/* disconnected */
@@ -292,14 +296,15 @@ static int msg_color3279[N_MSGS] = {
     HOST_COLOR_WHITE,	/* scrolled */
     HOST_COLOR_RED,	/* minus */
     HOST_COLOR_RED,	/* disabled */
+    HOST_COLOR_WHITE,	/* file transfer */
 };
 static bool oia_insert = false;
 static bool oia_reverse = false;
 static bool oia_kmap = false;
 static bool oia_script = false;
 static bool oia_printer = false;
-static char *oia_cursor = (char *) 0;
-static char *oia_timing = (char *) 0;
+static char *oia_cursor = NULL;
+static char *oia_timing = NULL;
 
 static unsigned char disc_msg[] = {
     CG_lock, CG_space, CG_commhi, CG_badcommhi, CG_commhi, CG_commjag,
@@ -339,15 +344,13 @@ static unsigned char proxy_msg[] = {
 static int proxy_len = sizeof(proxy_msg);
 
 static unsigned char telnet_msg[] = {
-    CG_lock, CG_space, CG_commhi, CG_badcommhi, CG_commhi, CG_commjag,
-    CG_commlo, CG_space, CG_bracketleft, CG_T, CG_E, CG_L, CG_N, CG_E, CG_T,
+    CG_lock, CG_space, CG_bracketleft, CG_T, CG_E, CG_L, CG_N, CG_E, CG_T,
     CG_bracketright
 };
 static int telnet_len = sizeof(telnet_msg);
 
 static unsigned char tn3270e_msg[] = {
-    CG_lock, CG_space, CG_commhi, CG_badcommhi, CG_commhi, CG_commjag,
-    CG_commlo, CG_space, CG_bracketleft, CG_T, CG_N, CG_3, CG_2, CG_7, CG_0,
+    CG_lock, CG_space, CG_bracketleft, CG_T, CG_N, CG_3, CG_2, CG_7, CG_0,
     CG_E, CG_bracketright
 };
 static int tn3270e_len = sizeof(tn3270e_msg);
@@ -377,6 +380,7 @@ static unsigned char *a_dbcs;
 static unsigned char *a_scrolled;
 static unsigned char *a_minus;
 static unsigned char *a_disabled;
+static unsigned char *a_file_transfer;
 
 static ioid_t revert_timer_id = NULL_IOID;
 
@@ -439,6 +443,7 @@ status_init(void)
     a_scrolled = make_amsg("statusScrolled");
     a_minus = make_amsg("statusMinus");
     a_disabled = make_amsg("statusDisabled");
+    a_file_transfer = make_amsg("statusFileTransfer");
 
     oia_shift = toggled(APL_MODE)? AplMode: 0;
 }
@@ -642,6 +647,18 @@ status_connect(bool connected)
 	} else if (kybdlock & KL_AWAITING_FIRST) {
 	    cancel_disabled_revert();
 	    do_msg(AWAITING_FIRST);
+	} else if (kybdlock & KL_ENTER_INHIBIT) {
+	    cancel_disabled_revert();
+	    do_msg(INHIBIT);
+        } else if (kybdlock & KL_BID) {
+	    cancel_disabled_revert();
+	    do_msg(TWAIT);
+        } else if (kybdlock & KL_FT) {
+	    cancel_disabled_revert();
+	    do_msg(FILE_TRANSFER);
+        } else if (kybdlock & KL_DEFERRED_UNLOCK) {
+	    cancel_disabled_revert();
+	    do_msg(UNLOCK_DELAY);
 	} else {
 	    cancel_disabled_revert();
 	    do_msg(BLANK);
@@ -704,7 +721,7 @@ revert_early(void)
 
 /* Keyboard disable flash. */
 void
-status_keyboard_disable_flash()
+status_keyboard_disable_flash(void)
 {
     if (keyboard_disabled()) {
 	if (oia_msg == KBD_DISABLED) {
@@ -828,13 +845,7 @@ void
 status_reset(void)
 {
     cancel_disabled_revert();
-    if (kybdlock & KL_ENTER_INHIBIT) {
-	do_msg(INHIBIT);
-    } else if (kybdlock & KL_DEFERRED_UNLOCK) {
-	do_msg(UNLOCK_DELAY);
-    } else {
-	status_connect(PCONNECTED);
-    }
+    status_connect(PCONNECTED);
 }
 
 /* Toggle insert mode */
@@ -943,29 +954,32 @@ status_timing(struct timeval *t0, struct timeval *t1)
 void
 status_untiming(void)
 {
-    do_timing(oia_timing = (char *) 0);
+    do_timing(oia_timing = NULL);
 }
 
 /* Update cursor position */
 void
 status_cursor_pos(int ca)
 {
-    static char buf[CCNT+1];
+    char *buf;
 
     if (xappres.xquartz_hack) {
-	snprintf(buf, sizeof(buf), "%02d/%02d", (ca/COLS + 1) % 100,
-		(ca%COLS + 1) % 100);
+	buf = xs_buffer("%02d/%02d", ((ca / COLS) + 1) % 100,
+		((ca % COLS) + 1) % 100);
     } else {
-	snprintf(buf, sizeof(buf), "%03d/%03d", ca/COLS + 1, ca%COLS + 1);
+	buf = xs_buffer("%03d/%03d", ((ca / COLS) + 1) % 1000,
+		((ca % COLS) + 1) % 1000);
     }
-    do_cursor(oia_cursor = buf);
+    Replace(oia_cursor, buf);
+    do_cursor(oia_cursor);
 }
 
 /* Erase cursor position */
 void
 status_uncursor_pos(void)
 {
-    do_cursor(oia_cursor = (char *) 0);
+    Replace(oia_cursor, NULL);
+    do_cursor(oia_cursor);
 }
 
 /* Internal routines */
@@ -1505,6 +1519,21 @@ do_disabled(void)
     }
 }
 
+static void
+do_file_transfer(void)
+{
+    static unsigned char file_transfer[] = {
+	CG_lock, CG_space, CG_F, CG_i, CG_l, CG_e, CG_space,
+	CG_T, CG_r, CG_a, CG_n, CG_s, CG_f, CG_e, CG_r
+    };
+
+    if (*standard_font) {
+	status_msg_set(a_file_transfer, strlen((char *)a_file_transfer));
+    } else {
+	status_msg_set(file_transfer, sizeof(file_transfer));
+    }
+}
+
 /* Insert, reverse, kmap, script, shift, compose */
 
 static void
