@@ -1,5 +1,6 @@
-/*
- * Copyright (c) 1993-2021 Paul Mattes.
+/**
+ * @copyright
+ * Copyright (c) 1993-2024 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta, GA
  *  30332.
@@ -29,10 +30,10 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-/*
- *	x3270.c
- *		A 3270 Terminal Emulator for X11
- *		Main proceudre.
+/**
+ *	@file x3270.c
+ *		@par A 3270 Terminal Emulator for X11
+ *		Main procedure.
  */
 
 #include "globals.h"
@@ -68,16 +69,19 @@
 #include "opts.h"
 #include "popups.h"
 #include "pr3287_session.h"
+#include "prefer.h"
 #include "print_screen.h"
 #include "print_window.h"
 #include "product.h"
 #include "proxy_toggle.h"
 #include "query.h"
+#include "resolver.h"
 #include "resourcesc.h"
 #include "save_restore.h"
 #include "screen.h"
 #include "selectc.h"
 #include "sio.h"
+#include "sio_glue.h"
 #include "status.h"
 #include "task.h"
 #include "telnet.h"
@@ -88,6 +92,7 @@
 #include "vstatus.h"
 #include "xactions.h"
 #include "xappres.h"
+#include "xglobals.h"
 #include "xio.h"
 #include "xkybd.h"
 #include "xmenubar.h"
@@ -108,7 +113,8 @@ Widget          toplevel;
 XtAppContext    appcontext;
 Atom            a_delete_me, a_save_yourself, a_3270, a_registry, a_encoding,
 		a_state, a_net_wm_state, a_net_wm_state_maximized_horz,
-		a_net_wm_state_maximized_vert, a_atom, a_spacing;
+		a_net_wm_state_maximized_vert, a_net_wm_name, a_atom, a_spacing,
+		a_pixel_size, a_font;
 char		full_model_name[13] = "IBM-";
 char	       *model_name = &full_model_name[4];
 Pixmap          gray;
@@ -155,6 +161,9 @@ XrmOptionDescRec base_options[]= {
     { OptConnectTimeout,DotConnectTimeout,XrmoptionSepArg,	NULL },
     { OptDevName,	DotDevName,	XrmoptionSepArg,	NULL },
     { OptTrace,		DotTrace,	XrmoptionNoArg,		ResTrue },
+#if defined(LOCAL_PROCESS) /*[*/
+    { OptLocalProcess,	NULL,		XrmoptionSkipLine,	NULL },
+#endif /*]*/
     { OptEmulatorFont,	DotEmulatorFont,XrmoptionSepArg,	NULL },
     { OptHostsFile,	DotHostsFile,	XrmoptionSepArg,	NULL },
     { OptHttpd,		DotHttpd,	XrmoptionSepArg,	NULL },
@@ -167,6 +176,8 @@ XrmOptionDescRec base_options[]= {
     { OptKeypadOn,	DotKeypadOn,	XrmoptionNoArg,		ResTrue },
     { OptKeyPasswd,	DotKeyPasswd,	XrmoptionSepArg,	NULL },
     { OptLoginMacro,	DotLoginMacro,	XrmoptionSepArg,	NULL },
+    { OptTlsMaxProtocol,DotTlsMaxProtocol,XrmoptionSepArg,	NULL },
+    { OptTlsMinProtocol,DotTlsMinProtocol,XrmoptionSepArg,	NULL },
     { OptMinVersion,	DotMinVersion,	XrmoptionSepArg,	NULL },
     { OptModel,		DotModel,	XrmoptionSepArg,	NULL },
     { OptMono,		DotMono,	XrmoptionNoArg,		ResTrue },
@@ -176,6 +187,8 @@ XrmOptionDescRec base_options[]= {
     { OptOnce,		DotOnce,	XrmoptionNoArg,		ResTrue },
     { OptOversize,	DotOversize,	XrmoptionSepArg,	NULL },
     { OptPort,		DotPort,	XrmoptionSepArg,	NULL },
+    { OptPreferIpv4,	DotPreferIpv4,	XrmoptionNoArg,		ResTrue },
+    { OptPreferIpv6,	DotPreferIpv6,	XrmoptionNoArg,		ResTrue },
     { OptPrinterLu,	DotPrinterLu,	XrmoptionSepArg,	NULL },
     { OptProxy,		DotProxy,	XrmoptionSepArg,	NULL },
     { OptReconnect,	DotReconnect,	XrmoptionNoArg,		ResTrue },
@@ -195,18 +208,19 @@ XrmOptionDescRec base_options[]= {
     { OptUser,		DotUser,	XrmoptionSepArg,	NULL },
     { OptUtf8,		DotUtf8,	XrmoptionNoArg,		ResTrue },
     { OptVerifyHostCert,DotVerifyHostCert,XrmoptionNoArg,	ResTrue },
-    { "-xrm",		NULL,		XrmoptionResArg,	NULL }
+    { OptXrm,		NULL,		XrmoptionResArg,	NULL }
 };
 int num_base_options = XtNumber(base_options);
 
 XrmOptionDescRec *options;
-int num_options;
+int num_options; /**< number of options */
 
+/** Option help. */
 static struct option_help {
-    char *opt;
-    char *args;
-    char *help;
-    unsigned tls_flag;
+    char *opt;		/**< option name */
+    char *args;		/**< arguments or NULL */
+    char *help;		/**< help text */
+    unsigned tls_flag;	/**< flags for conditional TLS options */
 } option_help[] = {
     { OptAcceptHostname, "[DNS:]<name>",
 	"Host name to accept from server certificate",
@@ -230,6 +244,9 @@ static struct option_help {
     { OptColorScheme, "<name>", "Use color scheme <name>" },
     { OptConnectTimeout, "<seconds>", "Timeout for host connect requests" },
     { OptDevName, "<name>", "Device name (workstation ID)" },
+#if defined(LOCAL_PROCESS) /*[*/
+    { OptLocalProcess, "<command> [arg...]", "Run process instead of connecting to host" },
+#endif /*]*/
     { OptEmulatorFont, "<font>", "Font for emulator window" },
     { OptHttpd, "[<addr>:]<port>", "TCP port to listen on for http requests" },
     { OptHostsFile, "<filename>", "Pathname of ibm_hosts file" },
@@ -255,6 +272,8 @@ static struct option_help {
     { OptOnce, NULL, "Exit as soon as the host disconnects" },
     { OptOversize,  "<cols>x<rows>", "Larger screen dimensions" },
     { OptPort, "<port>", "Default TELNET port" },
+    { OptPreferIpv4, NULL, "Prefer IPv4 host addresses" },
+    { OptPreferIpv6, NULL, "Prefer IPv6 host addresses" },
     { OptPrinterLu,  "<luname>",
 	"Automatically start a pr3287 printer session to <luname>" },
     { OptProxy, "<type>:<host>[:<port>]", "Secify proxy type and server" },
@@ -269,6 +288,8 @@ static struct option_help {
     { OptScriptPortOnce, NULL, "Accept one script connection, then exit" },
     { OptSecure, NULL, "Set secure mode" },
     { OptTermName, "<name>", "Send <name> as TELNET terminal name" },
+    { OptTlsMaxProtocol, "<protocol>", "TLS maximum protocol version" },
+    { OptTlsMinProtocol, "<protocol>", "TLS minimum protocol version" },
     { OptTrace, NULL, "Enable tracing" },
     { OptTraceFile, "<file>", "Write traces to <file>" },
     { OptTraceFileSize, "<n>[KM]", "Limit trace file to <n> bytes" },
@@ -280,7 +301,7 @@ static struct option_help {
     { OptVerifyHostCert, NULL, "Verify TLS host certificate (enabled by default)",
 	TLS_OPT_VERIFY_HOST_CERT },
     { OptVersion, NULL, "Display build options and character sets" },
-    { "-xrm", "'x3270.<resource>: <value>'", "Set <resource> to <vale>" }
+    { OptXrm, "'x3270.<resource>: <value>'", "Set <resource> to <vale>" }
 };
 
 /* Fallback resources. */
@@ -354,6 +375,8 @@ usage(const char *msg)
 
     fprintf(stderr, "Usage: %s [options] [[prefix:][LUname@]hostname[:port]]\n",
 	    programname);
+    fprintf(stderr, "       %s [options] [<session-file>.x3270]\n",
+	    programname);
     fprintf(stderr, "Use " OptHelp1 " for the list of options\n");
     exit(1);
 }
@@ -366,6 +389,8 @@ cmdline_help(void)
 
     fprintf(stderr, "Usage: %s [options] [[prefix:][LUname@]hostname[:port]]\n",
 	    programname);
+    fprintf(stderr, "       %s [options] [<session-file>.x3270]\n",
+	    programname);
     fprintf(stderr, "Options:\n");
     for (i = 0; i < XtNumber(option_help); i++) {
 	if (option_help[i].tls_flag == 0
@@ -385,7 +410,7 @@ static void
 no_minus(char *arg)
 {
     if (arg[0] == '-') {
-	usage(xs_buffer("Unknown or incomplete option: %s", arg));
+	usage(xs_buffer("Unknown or incomplete option: '%s'", arg));
     }
 }
 
@@ -507,6 +532,7 @@ main(int argc, char *argv[])
     screentrace_register();
     x3270_register();
     xio_register();
+    sio_glue_register();
     hio_register();
     proxy_register();
     model_register();
@@ -514,6 +540,7 @@ main(int argc, char *argv[])
     xkybd_register();
     login_macro_register();
     vstatus_register();
+    prefer_register();
 
     /* Translate and validate -set and -clear toggle options. */
 #if defined(DEBUG_SET_CLEAR) /*[*/
@@ -556,11 +583,7 @@ main(int argc, char *argv[])
     /* Initialize. */
     toplevel = XtVaAppInitialize(
 	    &appcontext,
-#if defined(USE_APP_DEFAULTS) /*[*/
 	    "X3270",
-#else /*][*/
-	    "X3270xad",	/* explicitly _not_ X3270 */
-#endif /*]*/
 	    options, num_options,
 	    &argc, argv,
 	    fallbacks,
@@ -609,7 +632,10 @@ main(int argc, char *argv[])
 	cl_hostname = xs_buffer("%s:%s", argv[1], argv[2]);
 	break;
     default:
-	usage(NULL);
+	for (i = 0; i < argc; i++) {
+	    no_minus(argv[i]);
+	}
+	usage("Too many command-line options");
 	break;
     }
 
@@ -718,19 +744,24 @@ main(int argc, char *argv[])
 	appres.disconnect_clear = true;
     }
 
-    a_delete_me = XInternAtom(display, "WM_DELETE_WINDOW", False);
-    a_save_yourself = XInternAtom(display, "WM_SAVE_YOURSELF", False);
+    /* Set up the resolver. */
+    set_46(appres.prefer_ipv4, appres.prefer_ipv6);
+
+    /* Set up atoms. */
     a_3270 = XInternAtom(display, "3270", False);
-    a_registry = XInternAtom(display, "CHARSET_REGISTRY", False);
+    a_atom = XInternAtom(display, "ATOM", False);
+    a_delete_me = XInternAtom(display, "WM_DELETE_WINDOW", False);
     a_encoding = XInternAtom(display, "CHARSET_ENCODING", False);
+    a_font = XInternAtom(display, "FONT", False);
+    a_net_wm_name = XInternAtom(display, "_NET_WM_NAME", False);
+    a_net_wm_state_maximized_horz = XInternAtom(display, "_NET_WM_STATE_MAXIMIZED_HORZ", False);
+    a_net_wm_state_maximized_vert = XInternAtom(display, "_NET_WM_STATE_MAXIMIZED_VERT", False);
+    a_net_wm_state = XInternAtom(display, "_NET_WM_STATE", False);
+    a_pixel_size = XInternAtom(display, "PIXEL_SIZE", False);
+    a_registry = XInternAtom(display, "CHARSET_REGISTRY", False);
+    a_save_yourself = XInternAtom(display, "WM_SAVE_YOURSELF", False);
     a_spacing = XInternAtom(display, "SPACING", False);
     a_state = XInternAtom(display, "WM_STATE", False);
-    a_net_wm_state = XInternAtom(display, "_NET_WM_STATE", False);
-    a_net_wm_state_maximized_horz = XInternAtom(display,
-	    "_NET_WM_STATE_MAXIMIZED_HORZ", False);
-    a_net_wm_state_maximized_vert = XInternAtom(display,
-	    "_NET_WM_STATE_MAXIMIZED_VERT", False);
-    a_atom = XInternAtom(display, "ATOM", False);
 
     /* Add the Xt-only actions. */
     xaction_init();
@@ -798,24 +829,10 @@ main(int argc, char *argv[])
     set_rows_cols(model_number, ovc, ovr);
     net_set_default_termtype();
 
-    hostfile_init();
-
     /* Initialize the icon. */
     icon_init();
 
-    /*
-     * If no hostname is specified on the command line, ignore certain
-     * options.
-     */
-    if (argc <= 1) {
-#if defined(LOCAL_PROCESS) /*[*/
-	if (cl_hostname == NULL)
-#endif /*]*/
-	{
-	    appres.once = false;
-	}
-	appres.interactive.reconnect = false;
-    }
+    hostfile_init();
 
     if (xappres.char_class != NULL) {
 	reclass(xappres.char_class);
@@ -973,7 +990,7 @@ relabel(bool ignored _is_unused)
 {
     char *title;
     char icon_label[8];
-    bool reconnect = appres.interactive.reconnect && reconnect_host != NULL;
+    bool reconnect = host_retry_mode && reconnect_host != NULL;
 
     if (user_title != NULL && user_icon_name != NULL) {
 	return;
@@ -984,7 +1001,7 @@ relabel(bool ignored _is_unused)
 	sprintf(title, "x3270-%d%s %s", model_num, (IN_NVT ? "A" : ""),
 		reconnect_host);
 	if (user_title == NULL) {
-	    XtVaSetValues(toplevel, XtNtitle, title, NULL);
+	    screen_set_title(title);
 	}
 	if (user_icon_name == NULL) {
 	    XtVaSetValues(toplevel, XtNiconName, reconnect_host, NULL);
@@ -994,7 +1011,7 @@ relabel(bool ignored _is_unused)
 	sprintf(title, "x3270-%d", model_num);
 	sprintf(icon_label, "x3270-%d", model_num);
 	if (user_title == NULL) {
-	    XtVaSetValues(toplevel, XtNtitle, title, NULL);
+	    screen_set_title(title);
 	}
 	if (user_icon_name == NULL) {
 	    XtVaSetValues(toplevel, XtNiconName, icon_label, NULL);
@@ -1022,6 +1039,11 @@ label_init(void)
     if (user_icon_name != NULL) {
 	set_aicon_label(user_icon_name);
     }
+
+    /* Set the _NET_WM_NAME property, because Xt doesn't. */
+    if (user_title != NULL) {
+	screen_set_title(user_title);
+    }
 }
 
 /*
@@ -1119,16 +1141,6 @@ dump_argv(const char *when, int argc, char **argv)
 }
 #endif /*]*/
 
-/* Comparison function for toggle name qsort. */
-static int
-name_cmp(const void *p1, const void *p2)
-{
-    const char *s1 = *(char *const *)p1;
-    const char *s2 = *(char *const *)p2;
-
-    return strcmp(s1, s2);
-}
-
 /* Double backslashes in a value so it gets past -xrm. */
 static char *
 requote(const char *s)
@@ -1209,7 +1221,7 @@ parse_set_clear(int *argcp, char **argv)
 			exit(1);
 		    }
 		}
-		argv_out[argc_out++] = "-xrm";
+		argv_out[argc_out++] = OptXrm;
 		argv_out[argc_out++] = xs_buffer("x3270.%s: %s",
 			toggle_names[j].name, value? ResTrue: ResFalse);
 		found = true;
@@ -1217,47 +1229,17 @@ parse_set_clear(int *argcp, char **argv)
 	    }
 	}
 	if (!found) {
-	    const char *proper_name;
+	    const char *proper_name = argv[i];
 	    int xt = init_extended_toggle(argv[i], nlen, bool_only,
 		    eq? eq + 1: (is_set? ResTrue: ResFalse), &proper_name);
 
-	    if (xt < 0) {
-		fprintf(stderr, "Error: " OptSet " %s: invalid value\n",
-			 argv[i]);
-		exit(1);
+	    if (xt == 0 && eq) {
+		proper_name = lazyaf("%.*s", (int)(eq - argv[i]), argv[i]);
 	    }
-	    if (xt == 1) {
-		argv_out[argc_out++] = "-xrm";
-		argv_out[argc_out++] = xs_buffer("x3270.%s: %s", proper_name,
-			eq? requote(eq + 1): (is_set? ResTrue: ResFalse));
-		found = true;
-	    }
-	}
-	if (!found) {
-	    const char **tn;
-	    int ntn = 0;
-	    int nx;
-	    char **nxnames;
-
-	    nxnames = extended_toggle_names(&nx, bool_only);
-	    tn = (const char **)Calloc(N_TOGGLES + nx, sizeof(char **));
-	    for (j = 0; toggle_names[j].name != NULL; j++) {
-		if (toggle_supported(toggle_names[j].index) &&
-			!toggle_names[j].is_alias) {
-		    tn[ntn++] = toggle_names[j].name;
-		}
-	    }
-	    memcpy(tn + ntn, nxnames, nx * sizeof(char **));
-	    ntn += nx;
-	    qsort(tn, ntn, sizeof(const char *), name_cmp);
-	    fprintf(stderr, "Unknown %stoggle name '%.*s'. Toggle names are:\n",
-		    bool_only? "Boolean ": "", (int)nlen, argv[i]);
-	    for (j = 0; j < ntn; j++) {
-		fprintf(stderr, " %s", tn[j]);
-	    }
-	    fprintf(stderr, "\n");
-	    Free(tn);
-	    exit(1);
+	    argv_out[argc_out++] = OptXrm;
+	    argv_out[argc_out++] = lazyaf("x3270.%s: %s", proper_name,
+			(eq != NULL)? requote(eq + 1):
+			    (is_set? ResTrue: ResFalse));
 	}
     }
 
@@ -1306,32 +1288,38 @@ copy_xres_to_res_bool(void)
     int i;
 #   define copy_bool(field)	appres.field = xappres.bools.field
 
-    copy_bool(once);
-    copy_bool(scripted);
-    copy_bool(modified_sel);
-    copy_bool(unlock_delay);
     copy_bool(bind_limit);
     copy_bool(bind_unlock);
-    copy_bool(new_environ);
-    copy_bool(socket);
-    copy_bool(numeric_lock);
-    copy_bool(secure);
-    copy_bool(oerr_lock);
+    copy_bool(bsd_tm);
+    copy_bool(contention_resolution);
     copy_bool(debug_tracing);
     copy_bool(disconnect_clear);
     copy_bool(highlight_bold);
-    copy_bool(bsd_tm);
-    copy_bool(trace_monitor);
     copy_bool(idle_command_enabled);
+    copy_bool(modified_sel);
+    copy_bool(numeric_lock);
+    copy_bool(new_environ);
     copy_bool(nvt_mode);
+    copy_bool(oerr_lock);
+    copy_bool(once);
+    copy_bool(prefer_ipv4);
+    copy_bool(prefer_ipv6);
+    copy_bool(reconnect);
+    copy_bool(retry);
     copy_bool(script_port_once);
+    copy_bool(scripted);
+    copy_bool(scripted_always);
+    copy_bool(secure);
+    copy_bool(socket);
+    copy_bool(trace_monitor);
+    copy_bool(unlock_delay);
     copy_bool(utf8);
+    copy_bool(wrong_terminal_name);
 
+    copy_bool(interactive.do_confirms);
     copy_bool(interactive.mono);
     copy_bool(interactive.menubar);
     copy_bool(interactive.visual_bell);
-    copy_bool(interactive.reconnect);
-    copy_bool(interactive.do_confirms);
 
     for (i = 0; i < N_TOGGLES; i++) {
 	copy_bool(toggle[i]);
@@ -1346,13 +1334,21 @@ copy_xres_to_res_bool(void)
 }
 
 /* Child exit callbacks. */
+
+/** Child exit state */
 typedef struct child_exit {
-    struct child_exit *next;
-    pid_t pid;
-    childfn_t proc;
+    struct child_exit *next;	/**< Linkage. */
+    pid_t pid;			/**< Child process ID. */
+    childfn_t proc;		/**< Function to call on exit.  */
 } child_exit_t;
 static child_exit_t *child_exits = NULL;
 
+/**
+ * Add a function to be called when a child exits.
+ * @param[in] pid	Child process ID.
+ * @param[in] fn	Function to call on exit.
+ * @return @ref ioid_t to identify this callback
+ */
 ioid_t
 AddChild(pid_t pid, childfn_t fn)
 {
@@ -1368,6 +1364,10 @@ AddChild(pid_t pid, childfn_t fn)
     return (ioid_t)cx;
 }
 
+
+/**
+ * Poll for exited children and call registered callbacks.
+ */
 static void
 poll_children(void)
 {
@@ -1395,8 +1395,39 @@ poll_children(void)
     }
 }
 
-/* Horrific glue for task.c. */
+/* Glue for redundant functions normally supplied by glue.c. */
+
+typedef struct dummy_reg {
+   struct dummy_reg *next;
+   void *value;
+} dummy_reg_t;
+static dummy_reg_t *dummy_reg;
+
+/* Put a little piece of allocated memory somewhere Valgrind can find it. */
+static void
+valkeep(void *p)
+{
+    dummy_reg_t *d = Malloc(sizeof(dummy_reg_t));
+
+    d->next = dummy_reg;
+    d->value = p;
+    dummy_reg = d;
+}
+
+void
+register_opts(opt_t *opts, unsigned num_opts)
+{
+    valkeep((void *)opts);
+}
+
+void
+register_resources(res_t *res, unsigned num_res)
+{
+    valkeep((void *)res);
+}
+
 void
 register_xresources(xres_t *res, unsigned num_xres)
 {
+    valkeep((void *)res);
 }
