@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2016, 2020 Paul Mattes.
+ * Copyright (c) 1993-2024 Paul Mattes.
  * Copyright (c) 1990, Jeff Sparkes.
  * Copyright (c) 1989, Georgia Tech Research Corporation (GTRC), Atlanta, GA
  *  30332.
@@ -137,16 +137,19 @@ static bool know_mods = false;
 static XtActionsRec xactions[] = {
     { "Cut",		Cut_xaction },
     { "Default",	Default_xaction },
+    { AnStepEfont,	StepEfont_xaction },
     { "HandleMenu",	HandleMenu_xaction },
     { "insert-selection",	insert_selection_xaction },
     { "KybdSelect",	KybdSelect_xaction },
     { "MouseSelect",	MouseSelect_xaction },
-    { "MoveCursor",	MoveCursor_xaction },
+    { AnMoveCursor,	MoveCursor_xaction },
+    { AnMoveCursor1,	MoveCursor1_xaction },
     { "move-select",	move_select_xaction },
     { PA_END,		PA_End_xaction },
     { PA_KEYMAP_TRACE,	PA_KeymapTrace_xaction },
     { PA_PFX "ConfigureNotify", PA_ConfigureNotify_xaction },
     { PA_PFX "confirm",	PA_confirm_xaction },
+    { PA_PFX "dialog-copy", PA_dialog_copy_xaction },
     { PA_PFX "dialog-focus", PA_dialog_focus_xaction },
     { PA_PFX "dialog-next",	PA_dialog_next_xaction },
     { PA_PFX "EnterLeave",	PA_EnterLeave_xaction },
@@ -365,7 +368,7 @@ xaction_init2(void)
 
     /* Fill in the table. */
     FOREACH_LLIST(&actions_list, e, action_elt_t *) {
-	if (strcmp(e->t.name, "MoveCursor") && (e->t.flags & ACTION_KE)) {
+	if (strcmp(e->t.name, AnMoveCursor) && strcmp(e->t.name, AnMoveCursor1) && (e->t.flags & ACTION_KE)) {
 	    wrapper_actions[nwrappers].string = (String)e->t.name;
 	    wrapper_actions[nwrappers].proc = xt_mapped_actions[nwrappers];
 	    nwrappers++;
@@ -654,6 +657,7 @@ trace_event(XEvent *event)
     XConfigureEvent *cevent;
     XClientMessageEvent *cmevent;
     XExposeEvent *exevent;
+    XVisibilityEvent *vevent;
     const char *press = "Press";
     const char *direction = "Down";
     char dummystr[KSBUF+1];
@@ -661,6 +665,7 @@ trace_event(XEvent *event)
     int ambiguous = 0;
     int state;
     const char *symname = "";
+    const char *viz[3] = { "Unobscured", "PartiallyObscured", "FullyObscured" };
 
     if (event == NULL) {
 	vtrace(" %s", ia_name[(int)ia_cause]);
@@ -787,6 +792,15 @@ trace_event(XEvent *event)
 	vtrace("ConfigureNotify [%dx%d+%d+%d]",
 		cevent->width, cevent->height, cevent->x, cevent->y);
 	break;
+    case VisibilityNotify:
+	vevent = (XVisibilityEvent *)event;
+	if (vevent->state >= VisibilityUnobscured &&
+		vevent->state <= VisibilityFullyObscured) {
+	    vtrace("VisibilityNotify [%s]", viz[vevent->state]);
+	} else {
+	    vtrace("VisibilityNotify [%d]", vevent->state);
+	}
+	break;
     default:
 	vtrace("Event %d", event->type);
 	break;
